/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.bloom;
import soss.core.BuilderIF;
import ostore.util.Carp;
import ostore.util.NodeId;
import seda.sandStorm.main.SandstormConfig;

/**
 * An interface for building config files.
 *
 * <p>Must have a constructor of the form 
 * <code>public Builder (String [] args)</code>.
 *
 * @author     Sean C. Rhea
 * @version    $Id: DualBuilder.java,v 1.5 2002/07/22 20:54:52 srhea Exp $
 */
public class DualBuilder implements BuilderIF {

    public DualBuilder (String [] args) throws Exception {
	
	if (args.length != 11) {
	    System.err.print ("usage: DualBuilder <fed> <filter size> " +
		    "<filter depth> <hashes> <src filtering> " +
		    "<dest filtering> <use bloom> <static/dynamic>" +
		    " <bloom limit> <use tapestry> <seed>\ngot:");
	    for (int i = 0; i < args.length; ++i)
		System.err.print (" " + args [i]);
	    System.err.println ("");
	    System.exit (1);
	}

	_constructor_args = args;
    }

    public SandstormConfig build_node (String [] args) throws Exception {

	if (args.length < 3) {
	    Carp.die ("usage: DualBuilder.build_node <node id> " +
		    "<file list> <reads file> <neighbors>");
	}
	
	String num_neighbors = new Integer (args.length - 3).toString ();
	String node_id = args [0];
	String file_list = (args [1].equals ("NULL") ? null : args [1]);
	String surgedir = (args [2].equals ("NULL") ? null : args [2]);
	String federation = _constructor_args [0];
	String filter_size = _constructor_args [1];
	String filter_depth = _constructor_args [2];
	String hashes = _constructor_args [3];
	String src_filtering = _constructor_args [4];
	String dest_filtering = _constructor_args [5];
	boolean use_bloom = _constructor_args [6].equals ("true");
	boolean dynamic = _constructor_args [7].equals ("dynamic");
	String bloom_limit = _constructor_args [8];
	boolean use_tapestry = _constructor_args [9].equals ("true");
	String seed = _constructor_args [10];

	String [] default_args = {
	"node_id=" + node_id,
	};
	SandstormConfig result = new SandstormConfig (default_args);

	// StaticTClient
	{
	    String [] init_args = {
		"node_id=" + node_id,
		"federation=" + federation
	    };
	    result.addStage ("StaticTClient", 
		    "ostore.tapestry.impl.StaticTClient", init_args);
	}

	if (use_tapestry) {
	    // Router
	    String [] init_args = {
		"node_id=" + node_id,
		"using_bloom=" + use_bloom, 
		"fake_keys=true"
	    };
	    result.addStage ("Router", 
		    "ostore.tapestry.impl.Router", init_args);
	}
	else {
	    // Home Node
	    String [] init_args = {
		"node_id=" + node_id,
		"using_bloom=" + use_bloom, 
		"seed=" + seed, 
		"federation=" + federation
	    };
	    result.addStage ("HomeNode", 
		    "ostore.bloom.HomeNode", init_args);
	}

	if (use_bloom) {
	    // BloomStage
	    {
		String [] base_args = {
		    "node_id=" + node_id,
		    "federation=" + federation,
		    "src_filtering=" + src_filtering,
		    "dest_filtering=" + dest_filtering,
		    "bloom_limit=" + bloom_limit, 
		    "filter_size=" + filter_size, 
		    "filter_depth=" + filter_depth,
		    "bloom_hashes=" + hashes,
		    "num_neighbors=" + num_neighbors,
		};

		String [] init_args = new String [base_args.length +
		    Integer.parseInt (num_neighbors)];

		for (int i = 0; i < base_args.length; ++i) 
		    init_args [i] = base_args [i];

		int neighbor = 0;
		for (int i = base_args.length; i < init_args.length; ++i) {
		    init_args [i] = "neighbor-" + neighbor + "=" +
			args [3 + neighbor];
		    neighbor += 1;
		}

		result.addStage ("BloomStage", 
			"ostore.bloom.BloomStage", init_args);
	    }
	}

	if (dynamic) {
	    // SimulatedReplica
	    if (file_list != null) {
		String [] init_args = {
		    "node_id=" + node_id,
		    "file_list=" + file_list,
		    "max_cache_size=" + (420*1024) // 420kbyes
		};
		result.addStage ("SimulatedReplica", 
			"ostore.bloom.SimulatedReplica", init_args);
	    }

	    // SurgeClient
	    if (surgedir != null) {
		String [] init_args = {
		    "node_id=" + node_id,
		    "surgedir=" + surgedir
		};
		result.addStage ("SurgeClient", 
			"ostore.bloom.SurgeClient", init_args);
	    }
	}
	else {
	    // SimulatedReplica
	    if (file_list != null) {
		String [] init_args = {
		    "node_id=" + node_id,
		    "file_list=" + file_list
		};
		result.addStage ("SimulatedReplica", 
			"ostore.bloom.SimulatedReplica", init_args);
	    }

	    // StaticClient
	    if (surgedir != null) {
		int port = (new NodeId (node_id)).port ();
		String [] init_args = {
		    "node_id=" + node_id,
		    "read-file=" + surgedir + "/read-file." + port
		};
		result.addStage ("StaticClient", 
			"ostore.bloom.StaticClient", init_args);
	    }
	}

	return result;
    }

    String [] _constructor_args;
}

