/**
 * CachePolicy.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

package ostore.cache;

public interface CachePolicy
{

  ////////////////////       CONSTANTS		////////////////////
  ////////////////////    STATIC VARIABLES 	////////////////////
  ////////////////////       VARIABLES		////////////////////
  ////////////////////  STATIC CONSTRUCTORS	////////////////////
  ////////////////////      CONSTRUCTORS	////////////////////
  ////////////////////     STATIC METHODS 	////////////////////
  //////////////////// 		METHODS		////////////////////

  /**
   * How many lines are in here? */
  public int size();

  /**
   * Are there any lines in here? */
  public boolean empty();

  /**
   * Insert into loop and return pointer. */
  public void insert(Line value);


  /**
   * Select the most appropriate line to remove. */
  public Line select_victim(boolean modified_ok);
    
  /**
   * Manually remove a line.*/
  public void remove(Line value);

  /**
   * Print out contents. */
  public String toString();

  ////////////////////     INNER CLASSES	////////////////////

  /** 
   * Opaque type used to reference line in cache.
   * 
   * <p>A note on locking: references to the _next and _prev pointers in
   * a Line are controled by locking the CachePolicy object.  This
   * nicely avoids deadlock during the remove function, for example,
   * where we want to lock several Lines at a time.  References to the
   * referenced and modified bits, on the other hand, are controled by
   * locking the line itself. */
  public static class Line
  {
    ////////////////////     CONSTANTS		////////////////////
    ////////////////////  STATIC VARIABLES 	////////////////////
    ////////////////////     VARIABLES		////////////////////
    ////////////////////    CONSTRUCTORS	////////////////////

    /**
     * Constructs a new <code>CachePolicy.Line</code>.
     *
     * @param modified  indicate if line has been modified. */
    public Line(boolean modified)
    {
      _mod = modified;
    }

    ////////////////////	STATIC METHODS 	////////////////////
    //////////////////// 	  METHODS	////////////////////
    public synchronized boolean referenced() { return _ref; }
    public synchronized boolean modified() { return _mod; }

    public synchronized void set_referenced() { _ref = true; }
    public synchronized void unset_referenced() { _ref = false; }

    public synchronized void set_modified() { _mod = true; }
    public synchronized void unset_modified() { _mod = false; }

    public synchronized boolean in_clock() { return _next != null; }

    public synchronized Line get_next() { return _next; }
    public synchronized Line get_prev() { return _prev; }

    public synchronized void set_next(Line next) { _next = next; } 
    public synchronized void set_prev(Line prev) { _prev = prev; } 

    private boolean _ref, _mod;
    public Line _next, _prev;
  }
}

