/**
 * Signal.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.dispatch;

import java.lang.reflect.Field;
import java.util.HashMap;
import seda.sandStorm.api.SignalIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;
import ostore.util.Carp;
import ostore.util.NodeId;
import ostore.util.Debug;
import ostore.util.DebugFlags;

/**
 * Contains a few static methods to mimic UNIX signals.
 * This class provides a unified interface for OceanStore- and
 * SandStorm-specific signals.  It sits atop both the
 * <code>Classifier</code> and <code>SignalMgr</code>.
 *
 * TODO: Coallesce all signals into single class with a "type" field.
 * Intercept SandStorm signals and replace with local equivalent.
 *
 * @author	Dennis Geels
 * @version $Id: Signal.java,v 1.14 2003/03/27 20:58:24 strib Exp $
 */

public class Signal
{
    ////////////////////       CONSTANTS	////////////////////

    private static final String class_tag = "Signal";
    private static final boolean DEBUG = DebugFlags.DISPATCH_SIGNAL;

    /**
     * The code for
     * {@link seda.sandStorm.api.StagesInitializedSignal}. 
     * Dispatched when the <code>init</code> method of each stage has
     * completed.
     */
    public static final int INIT = 4;	// Probably won't need SIGILL

    /**
     * The code for {@link Signal.RebuildSignal}.
     * This signal tells the receivers to resynchronize/optimize their
     * internal data structures.  Useful for starting time-consuming
     * optimizations when machine load is low. 
     */
    public static final int RBLD = 8;	// Probably won't need SIGFPE

    /**
     * The code for {@link Signal.AlarmSignal}.
     * Dispatched by <code>alarm</code>.
     */
    public static final int ALRM = 14;	// == SIGALRM

    /**
     * Used to size the signal tables. 
     */
    private static final int LAST_SIG = 14;
    
    ////////////////////    STATIC VARIABLES 	////////////////////
    
    /**
     * A <code>NodeId->Signal</code> map of signal objects.
     * Each virtual machine has a single <code>Signal</code>.
     */
    private static HashMap _signals;

    ////////////////////       VARIABLES	////////////////////

    /**
     * The <code>Classifier</code> for this machine.
     */
    private Classifier classifier;

    /**
     * The <code>Class</code> objects for each type of signal.
     */
    private Class[] classes;

    /**
     * The <code>Field</code> objects for the <code>target</code> field for
     * applicable signal type.
     */
    private Field[] targets;

    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	// Initialize the map of signals.
	_signals = new HashMap ();
    }
    
    ////////////////////      CONSTRUCTORS	////////////////////

    public Signal(NodeId node_id)
    {
	classifier = Classifier.getClassifier(node_id);

	// Initialize the <code>Class</code> and <code>Field</code> objects.
	classes = new Class[LAST_SIG+1];
	targets = new Field[LAST_SIG+1];
    	try {
	    classes[INIT] =
		Class.forName("seda.sandStorm.api.StagesInitializedSignal");

	    classes[RBLD] =
		Class.forName("ostore.dispatch.Signal$RebuildSignal");
	    targets[RBLD] = classes[RBLD].getField( "target" );

	    classes[ALRM] =
		Class.forName("ostore.dispatch.Signal$AlarmSignal");
	    targets[ALRM] = classes[ALRM].getField( "target" );

	} catch ( ClassNotFoundException cnfe ) {
	    Carp.die( cnfe );
	} catch ( NoSuchFieldException nsfe ) {
	    Carp.die( nsfe );
	}
    }
    
    ////////////////////     STATIC METHODS 	////////////////////

    /**
     * Returns the <code>Signal</code> instance corresponding to a
     * <code>NodeId</code>. 
     *
     * @param node_id	The <code>NodeId</code> which parameterizes a
     * virtual SandStorm instance.
     * @return the (possibly new) <code>Signal</code> instance
     * corresponding to <code>node_id</code>.
     */
    public static Signal getSignal(NodeId node_id)
    {
	Signal result = null;
	synchronized (_signals) {
	    result = (Signal) _signals.get (node_id);
	    if (result == null) {
		result = new Signal( node_id );
		_signals.put (node_id, result);
	    }
	}
	return result;
    }

    /**
     * Deletes the <code>Signal</code> instance corresponding to a
     * <code>NodeId</code>. Should be called only when a node is going down
     * and going to come up again with a new node id.  Use very carefully!
     *
     * @param node_id	The <code>NodeId</code> which parameterizes a
     * virtual SandStorm instance.
     */
    public static void removeSignal( NodeId node_id )
    {
	synchronized (_signals) {
	    _signals.remove (node_id);
	}
    }

    //////////////////// 	METHODS		////////////////////
    /**
     * Subscribes the specified sink to signals of the specified type which
     * have the specified target.  If <code>target</code> is
     * <code>null</code>, all signals of the specified type will be
     * enqueued on the sink.  For some signals (such as <code>INIT</code>),
     * the <code>target</code> parameter is ignored.
     *
     * @param type	currently one of {@link #INIT} or {@link #ALRM}.
     * @param target	the name of the <code>StageIF</code> to which the
     * 			signals are directed, or <code>null</code>.
     * @param sink	the <code>SinkIF</code> to receive the
     * 			<code>SignalIF</code>s.
     * @return true iff subscription succeeded.
     */
    public boolean register( int type, String target, SinkIF sink )
    {
	final String tag = class_tag+".register";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	Filter f = new Filter();
	
	switch( type ) {
	case INIT:
	    if( DEBUG ) Debug.printtagln( tag, sink +
					  " to receive INIT signals" );
	    f.requireType( classes[INIT] );
	    break;
	case RBLD:
	    if( DEBUG ) Debug.printtagln( tag, sink +
					  " to receive RBLD signals with "
					  + "target \"" + target + "\"" );
	    f.requireType( classes[RBLD] );
	    if( target != null )
		f.requireValue( targets[RBLD], target );
	    break;
	case ALRM:
	    if( DEBUG ) Debug.printtagln( tag, sink +
					  " to receive ALRM signals with "
					  + "target \"" + target + "\"" );
	    f.requireType( classes[ALRM] );
	    if( target != null )
		f.requireValue( targets[ALRM], target );
	    break;
	default:
	    if( DEBUG ) Debug.printtagln( tag, "unknown signal type: "+ type);
	    return false;
	}

	boolean worked = classifier.subscribe( f, sink );
	if( DEBUG ) Debug.printtagln( tag, "done: " +
				      ( worked ? "successful" : "failed" ) );
	return worked;
    }

    /**
     * Dispatches a signal of the specified type with the specified target.
     * For some signals, the <code>target</code> parameter is ignored.
     * For others, it is required to be non-<code>null</code>.
     *
     * @param type	currently must be {@link #RBLD} or {@link #ALRM}.
     * @param target	the name of the <code>StageIF</code> to which the
     * 			signals are directed, or <code>null</code>.
     *
     * @exception 	SinkFullException if the signal could not be
     * dispatched. 
     *
     * @see Classifier#dispatch
     */
    public void signal( int type, String target ) throws SinkFullException
    {
	final String tag = class_tag+".signal";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	SignalIF to_send = null;
	
	switch( type ) {
	case INIT:
	    if( DEBUG ) Debug.printtagln( tag, "cannot send INIT signal" );
	    break;
	case RBLD:
	    if( DEBUG ) Debug.printtagln( tag, "sending RBLD" +
					  " to \"" + target + "\"" );
	    to_send = new RebuildSignal( target );
	    break;
	case ALRM:
	    if( DEBUG ) Debug.printtagln( tag, "sending ALRM" +
					  " to \"" + target + "\"" );
	    to_send = new AlarmSignal( target );
	    break;
	default:
	    if( DEBUG ) Debug.printtagln( tag, "unknown signal type: "+ type);
	}

	// May throw a <code>SinkFullException</code>.
	classifier.dispatch( to_send );
	
	if( DEBUG ) Debug.printtagln( tag, "done" );	
    }
    
    /**
     * Dispatches a signal of the specified type with the specified target.
     * All <code>SinkFullException</code>s are quietly dropped.
     *
     * @param type	currently must be {@link #RBLD} or {@link #ALRM}.
     * @param target	the name of the <code>StageIF</code> to which the
     * 			signals are directed, or <code>null</code>.
     *
     * @see #signal
     */
    public void signal_lossy( int type, String target )
    {
	final String tag = class_tag+".signal_lossy";
	if( DEBUG ) Debug.printtagln( tag, "called" );
	try {
	    signal( type, target );
	} catch( SinkFullException sfe ) {
	    // Ignored
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }
    
    /**
     * Dispatch an <code>AlarmSignal</code> at a specified time
     * This method has similar functionality to the <code>alarm</code>
     * function in C.
     *
     * @see #cancel
     *
     * @param target	The name of the target <code>StageIF</code>.
     *			May <b>not</b> be <code>null</code>. 
     * @param millis	The number of milliseconds from now to dispatch the
     * 			event.
     * @return a token to pass to <code>cancel</code> in order to cancel
     * the alarm before it is dispatched.
     */
    public Object alarm( String target, long millis )
    {
	final String tag = class_tag+".alarm";
	if( DEBUG ) Debug.printtagln( tag, "called: " +
				      target + " @ " + millis );
	Object token =
	    classifier.dispatch_later( new AlarmSignal( target ), millis );
	if( DEBUG ) Debug.printtagln( tag, "done. token: " + token );
	return token;
    }

    /**
     * Cancel an <code>AlarmSignal</code> scheduled by <code>alarm</code>.
     * If the alarm has already been signaled, this method has no effect.
     *
     * @see #alarm
     *
     * @param token	The <code>Object</code> returned by <code>alarm</code>.
     * @exception IllegalArgumentException if <code>token</code> is not a
     * token returned by the internal timer.
     */
    public void cancel( Object token )
    {
	if( DEBUG ) Debug.printtagln( class_tag+".cancel",
				      "called: " + token );
	classifier.cancel( token );
    }

    ////////////////////     INNER CLASSES	////////////////////
    
    /**
     * This signal tells the receivers to resynchronize/optimize their
     * internal data structures.  Useful for starting time-consuming
     * optimizations when machine load is low. 
     *
     * @see Signal#register
     * @see Signal#signal
     */
    public static class RebuildSignal implements SignalIF
    {
	////////////////////     VARIABLES	////////////////////
	/**
	 * The name of the target.  This name is opaque to the
	 * <code>Classifier</code>.  By convention, it should equal
	 * <code>StageIF.getName()</code> for the receiving stage, to make
	 * <code>register</code>ing for the signal simple.
	 */
	public String target;

	////////////////////    CONSTRUCTORS	////////////////////

	public RebuildSignal( String target ) {
	    if( target == null )
		throw new NullPointerException( "null signal target" );
	    this.target = target;
	}

	//////////////////// 	  METHODS	////////////////////

	public String toString() {
	    return "<RebuildSignal for \"" + target + "\">";
	}
    }

    /**
     * The OceanStore equivalent of SIGALRM
     *
     * @see Signal#alarm
     * @see Signal#register
     */
    public static class AlarmSignal implements SignalIF
    {
	////////////////////     VARIABLES	////////////////////
	/**
	 * The name of the target.  This name is opaque to the
	 * <code>Classifier</code>.  By convention, it should equal
	 * <code>StageIF.getName()</code> for the receiving
	 * stage, to make <code>register</code>ing for the signal simple.
	 */
	public String target;

	////////////////////    CONSTRUCTORS	////////////////////

	public AlarmSignal( String target ) {
	    if( target == null )
		throw new NullPointerException( "null signal target" );
	    this.target = target;
	}

	//////////////////// 	  METHODS	////////////////////

	public String toString() {
	    return ("<AlarmSignal for \"" + target + "\">");
	}
    }
}
