/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.monitor;

import seda.sandStorm.core.BufferElement;
import ostore.dispatch.*;
import ostore.util.*;
import java.util.*;
import java.text.DecimalFormat;
import java.text.FieldPosition;

/**
 * Log for recording resource utilization deltas. Each log entry consists
 * of a time value in milliseconds and a tuple of utilization or status
 * values (doubles).
 *
 * @author  Jeff Pang
 * @version $Id: ResourceLog.java,v 1.3 2003/02/21 20:37:16 jeffpang Exp $
 **/
public class ResourceLog
{

    public static final boolean DEBUG = false;
    
    // initial buffer space
    public static final int   INITIAL_SIZE = 256;
    // factort of space to allocate when overrun buffer
    // new_space = alloc_factor*old_space
    public static final float ALLOC_FACTOR = 2.0F;
    
    // Id of Node (for time stamps)
    private NodeId id;
    
    // list of all log entry blocks: List<Pair<timeLog,utilLog>>
    private LinkedList log;
    // currently active log block: each log entry is a (time,util[]) pair
    private long[] timeLog;
    private double[][] utilLog;
    // skips the first entry in the log when dumping to disk; this is
    // when the log was truncated but the last element was put on the new log
    private boolean ignoreFirst;
    // next index into current active block
    private int index;
    // current size of log (including unused portion of current block)
    private int size;

    /**
     * Create a resource log
     *
     * @param id node id of current node (for time values)
     * @param initial initial resource utilization. The size of the
     * array determines the number of elements in the log (# of util numbers)
     */
    public ResourceLog(NodeId id, double[] initial)
    {
	this.id = id;

	log = new LinkedList();
	timeLog = new long[INITIAL_SIZE];
	utilLog = new double[initial.length][INITIAL_SIZE];
	index = 0;
	size = INITIAL_SIZE;
	ignoreFirst = false;
	
	log.addLast(new Pair((Object)timeLog, (Object)utilLog));
	log(initial);
    }

    /**
     * Log a new utilization value at this instant
     *
     * @param newUtil the new utilization numbers to log. This array is
     * copied, so you can reuse the same one each time.
     */
    public synchronized void log(double[] newUtil)
    {
	// see if we need to allocate a new log block
	if (index == timeLog.length) {
	    int toAlloc = (int)(size*ALLOC_FACTOR);
	    timeLog = new long[toAlloc];
	    utilLog = new double[utilLog.length][toAlloc];
	    index = 0;
	    size += toAlloc;
	    log.addLast(new Pair((Object)timeLog, (Object)utilLog));
	}

	timeLog[index] = Clock.current_date(id).getTime();
	for (int i=0; i<utilLog.length; i++)
	    utilLog[i][index] = newUtil[i];
	index++;
    }

    private int befPercision = 5;
    private int aftPercision = 5;
    private DecimalFormat fmt = new DecimalFormat("#####.#####");
    private FieldPosition pos = new FieldPosition(DecimalFormat.INTEGER_FIELD);

    /**
     * Set percision used in values printed out by dump(). Default is 5,5
     *
     * @param befPercision max number of digits to print before decimal point
     * @param aftPercision max number of digits to print after decimal point
     */
    public synchronized void setPercision(int befPercision, int aftPercision)
    {
	// FIXME: different percisions for each util number
	this.befPercision = befPercision;
	this.aftPercision = aftPercision;
	String format = "";
	for (int i=0; i<befPercision; i++)
	    format += "#";
	format += ".";
	for (int i=0; i<aftPercision; i++)
	    format += "#";
	fmt = new DecimalFormat(format);
    }

    /**
     * Construct a BufferElement from the current log entries in the 
     * following text format:
     * <pre>
     * time1 util1[0] util1[1] util1[2] ...
     * time2 util2[0] util2[1] util2[2] ...
     * time3 util3[0] util3[1] util3[2] ...
     * ...
     * </pre>
     */
    public synchronized BufferElement dump()
    {
	// FIXME: This allocation overestimates buffer size (always worst case)
	//System.out.println("size=" + size + ", log.size=" + log.size());
	int len = size()*(10+(befPercision+aftPercision+2)*utilLog.length + 1);
	StringBuffer data = 
	    new StringBuffer(len);
	//System.out.println("predicted=" + len);
	Iterator i=log.listIterator();
	boolean first = true;
	while ( i.hasNext() ) {
	    Pair p = (Pair)i.next();
	    long[] times = (long[])p.first();
	    double[][] utils = (double[][])p.second();
	    int sz = i.hasNext()?times.length:index;
	    //System.out.println("length="+times.length+" index=" + index +
	    //	       " sz="+sz);
	    for (int j=(ignoreFirst && first)?1:0; j<sz; j++) {
		data.append(times[j]);
		for (int k=0; k<utils.length; k++) {
		    data.append(' ');
		    fmt.format(utils[k][j], data, pos);
		}
		data.append('\n');
	    }
	    first = false;
	}
    
	//System.out.println("actual="+data.length());
	
	BufferElement ret = new BufferElement(data.toString().getBytes());
	return ret;
    }

    /**
     * Truncate the log. The initial value in the new log is the
     * last value in the old one.
     */
    public synchronized void clear()
    {
	int toAlloc = timeLog.length;
	double[] last = util();

	log = new LinkedList();
	timeLog = new long[toAlloc];
	utilLog = new double[utilLog.length][toAlloc];
	index = 0;
	size = toAlloc;
	ignoreFirst = true;
	log.add(new Pair(timeLog, utilLog));

	log(last);
    }

    /**
     * Number of utilization numbers per log entry
     */
    public int number()
    {
	return utilLog.length;
    }

    /**
     * Instantaneous utilization
     *
     * @param i which util number
     */
    public synchronized double util(int i)
    {
	return utilLog[i][index-1];
    }

    /**
     * Get all the utilization numbers. Warning: this function allocates
     * memory.
     */
    public synchronized double[] util()
    {
	double[] u = new double[utilLog.length];
	for (int i=0; i<utilLog.length; i++)
	    u[i] = util(i);
	return u;
    }
	
    /**
     * Exact number of deltas in log
     */
    public synchronized int size()
    {
	return size - timeLog.length + index - (ignoreFirst?1:0);
    }

    //----------------------------------------------------------------------
    // Test
    //----------------------------------------------------------------------

    public static void main(String[] args) {
	int iters  = 10;
	int deltas = 100000;

	double[] array = { 0, 0, 0, 0};
	ResourceLog log = new ResourceLog(null, array);
	log.setPercision(6,1);

	System.out.println("time to do " + deltas + " deltas:");

	for (int j=0; j<iters; j++) {
	    long start = Clock.current_date(null).getTime();
	    for (int i=0; i<deltas; i++) {
		array[0] = i;
		array[1] = i;
		array[2] = i;
		array[3] = i;
		log.log(array);
	    }
	    long time = Clock.current_date(null).getTime() - start;
	    System.out.println("[" + j + "] add:   " + time);
	    start = Clock.current_date(null).getTime();
	    BufferElement e = log.dump();
	    time = Clock.current_date(null).getTime() - start;
	    System.out.println("[" + j + "] dump:  " + time);

	    start = Clock.current_date(null).getTime();
	    log.clear();
	    System.gc();
	    time = Clock.current_date(null).getTime() - start;
	    System.out.println("[" + j + "] clear: " + time);
	}
    }
}
