/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.monitor;

import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.StagesInitializedSignal;
import seda.sandStorm.lib.aDisk.*;
import seda.sandStorm.core.BufferElement;
import ostore.util.*;
import java.util.*;

/**
 * Interface for logging resource monitoring information.
 * <p>
 * Config Variables:
 * <pre>
 * logname         - the name of resouce logs are: logname-ResourceName
 * sampleInterval  - base sample interval for dynamic resources in ms
 * dumpInterval    - mean interval between dumping logs to disk in ms 
 *                   (exponentially dist.)
 * </pre>
 *
 * @author  Jeff Pang
 * @version $Id: ResourceMonitorStage.java,v 1.5 2003/02/27 22:52:01 jeffpang Exp $
 **/
public class ResourceMonitorStage extends StandardStage
{
    public static final String class_tag = "ResourceMonitorStage";
    public static final boolean DEBUG = true;

    // Super:
    // protected Classifier classifier;
    // protected NodeId my_node_id;
    // protected SinkIF my_sink;
    // protected String tag;
    // protected String [] event_types = {};
    // protected String [] inb_msg_types = {};
    // protected String [] outb_msg_types = {};


    /**
     * Data Structure for each resource.
     */
    private static class ResourceInfo
    {
	// the resource
	public Resource rsrc;
	// the log
	public AFile log;
	// time of last sample
	public long lastTime;
	// various other state associated with this: Map<String,Object>
	public HashMap state;

	public ResourceInfo(Resource rsrc) 
	{ 
	    this.rsrc = rsrc; 
	    this.lastTime = 0;
	    this.state = new HashMap();
	}
    }

    /**
     * Request to sample resources
     */
    public static class SampleReq implements QueueElementIF
    {
	// empty
    }

    /**
     * Request to dump log of a resource to disk
     */
    public static class DumpLogReq implements QueueElementIF
    {
	public ResourceInfo info;

	public DumpLogReq(ResourceInfo which)
	{
	    info = which;
	}
    }

    // ------------------------------------------------------------------------
    // Config
    // ------------------------------------------------------------------------

    // The name of resouce logs are: logname-ResourceName
    private String logname;
    // Base sample interval for dynamic resources in ms
    // TODO: Also used as the default for now
    private int sampleInterval;
    // mean interval between dumping logs to disk in ms (exponentially dist.)
    private int dumpInterval;

    // ------------------------------------------------------------------------
    // State
    // ------------------------------------------------------------------------

    // Random
    private Random rand;
    // Resources recorded manually by the code
    // Map<String name, ResourceInfo>
    private Map managed; 
    // Resources recorded dynamically by OS or other third party
    // Map<String name, ResourceInfo>
    private Map dynamic;

    // Requests waiting for a resource we don't yet have registered.
    // MultiMap<String name, RequestResourceReq>
    private MultiMap pendingRequestResourceReqs;

    /**
     * Constructor
     */
    public ResourceMonitorStage()
    {
	rand = new Random();
	managed = new HashMap();
	dynamic = new HashMap();

	pendingRequestResourceReqs = new MultiMap();

	// events we listen to
	event_types = new String[] {
	    "seda.sandStorm.api.StagesInitializedSignal",
	    "ostore.monitor.RegisterResourceReq",
	    "ostore.monitor.RequestResourceReq",
	    "ostore.monitor.ResourceMonitorStage$SampleReq",
	    "ostore.monitor.ResourceMonitorStage$DumpLogReq",
	};
    }

    // Super:
    // protected final void BUG (String msg);
    // protected void dispatch (QueueElementIF item);

    /**
     * Sandstorm implementation.
     */
    public void init(ConfigDataIF config) throws Exception 
    {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called");

	super.init(config);

	logname = config.getString("logname");
	if (logname == null) {
	    Carp.die(tag + ": missing config variable 'logname'");
	}
	sampleInterval = config.getInt("sampleInterval");
	if (sampleInterval < 0) {
	    sampleInterval = 10000; // 10 sec
	}
	dumpInterval = config.getInt("dumpInterval");
	if (dumpInterval < 0) {
	    dumpInterval = 360000; // 5 min
	}

	if (DEBUG) {
	    Debug.printtagln( tag, "Config Variables:");
	    Debug.printtagln( tag, "logname: " + logname);
	    Debug.printtagln( tag, "sampleInterval: " + sampleInterval);
	    Debug.printtagln( tag, "dumpInterval: " + dumpInterval);
	}
	
	registerDefaultResources();
    }

    /**
     * Register the default resources we manage and open their corresponding
     * logs.
     */
    private void registerDefaultResources() throws Exception
    {
	final String tag = class_tag+".registerDefaultResources";
	if( DEBUG ) Debug.printtagln( tag, "called");

	// dynamic resources
	DynamicResource[] dr = new DynamicResource[] {
	    new LoadAvgResource(my_node_id, sampleInterval),
	    new SystemMemoryResource(my_node_id, sampleInterval),
	};

	// managed resources
	Resource[] mr = new Resource[] {
	};

	for (int i=0; i<dr.length; i++) {
	    ResourceInfo info = new ResourceInfo(dr[i]);
	    info.log = openResourceLog(dr[i].getName());
	    dynamic.put(dr[i].getName(), info);
	}

	for (int i=0; i<mr.length; i++) {
	    ResourceInfo info = new ResourceInfo(mr[i]);
	    info.log = openResourceLog(mr[i].getName());
	    managed.put(mr[i].getName(), info);
	}
    }

    /**
     * Open a new resource log for the resource named rsrcName.
     * <p>
     * The name of the logfile is <tt>logname-rsrcName</tt>.
     */
    private AFile openResourceLog(String rsrcName) throws Exception
    {
	final String tag = class_tag+".openResourceLog";
	if( DEBUG ) Debug.printtagln( tag, "called");

	String logFilename = logname + "-" + rsrcName;
	
	return new AFile(logFilename,
			 my_sink,     // completion queue
			 true,        // create if necessary
			 false);      // not read only
    }

    /**
     * Sample a exponential distribution with mean lambda.
     */
    private double nextExponential(double lambda) {
	double m = rand.nextDouble();
	return (-1.0/lambda)*(Math.log(m)/Math.log(Math.E));
    }

    /**
     * Start the timer for when to dump the logs to disk. Exponentially
     * distribute the times so they don't all clump together.
     */
    private void startDumpTimer(ResourceInfo which)
    {
	final String tag = class_tag+".startDumpTimer";
	if( DEBUG ) Debug.printtagln( tag, "called");
	
	// save this so we don't recreate it everytime
	DumpLogReq req = (DumpLogReq)which.state.get("DumpLogReq");
	if (req == null) {
	    req = new DumpLogReq(which);
	    which.state.put("DumpLogReq", req);
	}

	int time = (int)(nextExponential(1.0/(double)dumpInterval) + 0.5);
	classifier.dispatch_later(req, time);
    }

    /**
     * Start the timer for sampling (base interval).
     */
    private void startSampleTimer()
    {
	final String tag = class_tag+".startSampleTimer";
	if( DEBUG ) Debug.printtagln( tag, "called");

	final SampleReq req = new SampleReq();
	classifier.dispatch_later(req, sampleInterval);
    }

    /**
     * Sandstorm implementation.
     */
    public void handleEvent(QueueElementIF element) 
    {
	final String tag = class_tag+".handleEvent";
	if( DEBUG ) Debug.printtagln( tag, "called: " + element);

	// initialize
	if (element instanceof StagesInitializedSignal) {
	    startSampleTimer();
	    for (Iterator i=dynamic.values().iterator(); i.hasNext(); ) {
		startDumpTimer((ResourceInfo)i.next());
	    }
	    for (Iterator i=managed.values().iterator(); i.hasNext(); ) {
		startDumpTimer((ResourceInfo)i.next());
	    }

	    // tell everyone we're ready
	    dispatch (new ResourceMonitorReadySignal());
	    return;
	}
	
	// main loop
	if (element instanceof SampleReq) {
	    handleSampleReq();
	} else if (element instanceof DumpLogReq) {
	    handleDumpLogReq((DumpLogReq)element);
	} else if (element instanceof RegisterResourceReq) {
	    handleRegisterResourceReq((RegisterResourceReq)element);
	} else if (element instanceof RequestResourceReq) {
	    handleRequestResourceReq((RequestResourceReq)element);
	} else if (element instanceof AFileIOCompleted) {
	    // TODO: Check for error?
	} else {
	    Carp.die(tag + ": unknown element type: " + element);
	}
    }

    /**
     * Sample dynamic resource utilizations
     */
    private void handleSampleReq()
    {
	final String tag = class_tag+".handleSampleReq";
	if( DEBUG ) Debug.printtagln( tag, "called");

	// FIXME: This is not scalable. Allow for now since we're only
	// monitoring 2 dynamic resources. Change to use a prio queue later

	long time = Clock.current_date(my_node_id).getTime();

	for (Iterator i=dynamic.values().iterator(); i.hasNext(); ) {
	    ResourceInfo info = (ResourceInfo)i.next();
	    DynamicResource rsrc = (DynamicResource)info.rsrc;
	    if (info.lastTime + rsrc.getSampleInterval() <= time) {
		if (DEBUG) Debug.printtagln(tag, "sampling " + rsrc.getName());
		rsrc.sample();
		info.lastTime = time;
	    }
	}

	startSampleTimer();
    }

    /**
     * Dump log of a resource to disk
     */
    private void handleDumpLogReq(DumpLogReq req)
    {
	final String tag = class_tag+".handleDumpLogReq";
	if( DEBUG ) Debug.printtagln( tag, "called");

	ResourceLog log = req.info.rsrc.getLog();
	if (log.size() == 0) {
	    if( DEBUG ) Debug.printtagln( tag, "log empty");
	    
	    startDumpTimer(req.info);
	    return;
	}

	BufferElement data = null;
	synchronized(log) {
	    data = log.dump();
	    log.clear();
	}

	AFileWriteRequest write = new AFileWriteRequest(data, my_sink);
	try {
	    req.info.log.enqueue(write);
	} catch (SinkException se) {
	    Carp.die(tag + ": sink exception while writing to log");
	}

	startDumpTimer(req.info);
	return;
    }

    /**
     * Register a new resource for logging.
     */
    private void handleRegisterResourceReq(RegisterResourceReq req)
    {
	final String tag = class_tag+".handleRegisterResourceReq";
	if( DEBUG ) Debug.printtagln( tag, "called");

	Resource rsrc = req.getResource();
	
	if (dynamic.containsKey(rsrc.getName()) ||
	    managed.containsKey(rsrc.getName())) {
	    dispatch(new RegisterResourceResp(req, RegisterResourceResp
					      .ALREADY_REGISTERED));
	    return;
	}

	ResourceInfo info = new ResourceInfo(rsrc);
	try {
	    info.log = openResourceLog(rsrc.getName());
	} catch (Exception ex) {
	    ex.printStackTrace();
	    Carp.die(tag + ": Failed to open resource log for " + 
		     rsrc.getName());
	}

	if (rsrc instanceof DynamicResource) {
	    dynamic.put(rsrc.getName(), info);
	} else {
	    managed.put(rsrc.getName(), info);
	}
	startDumpTimer(info);

	dispatch(new RegisterResourceResp(req));

	// see if anyone is waiting for this and respond to them
	if (pendingRequestResourceReqs.containsKey(rsrc.getName())) {
	    Set s = (Set)pendingRequestResourceReqs.remove(rsrc.getName());
	    for (Iterator i=s.iterator(); i.hasNext(); ) {
		RequestResourceReq w = (RequestResourceReq)i.next();
		handleRequestResourceReq(w);
	    }
	}
    }

    /**
     * Request a handle to a resource.
     */
    private void handleRequestResourceReq(RequestResourceReq req)
    {
	final String tag = class_tag+".handleRequestResourceReq";
	if( DEBUG ) Debug.printtagln( tag, "called");

	String name = req.getName();

	ResourceInfo info = (ResourceInfo)dynamic.get(name);
	if (info == null)
	    info = (ResourceInfo)managed.get(name);

	if (info == null) {
	    if (req.block()) {
		pendingRequestResourceReqs.put(name, req);
		return;
	    } else {
		dispatch(new RequestResourceResp(req, RequestResourceResp
						 .NO_SUCH_RESOURCE));
		return;
	    }
	}

	dispatch(new RequestResourceResp(req, info.rsrc));
    }
}
