/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.monitor;

import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;
import ostore.util.*;
import java.util.*;
import java.io.*;

/**
 * Total system memory usage (used/total). This keeps track of two values:
 * <pre>
 * [0] = fraction of system memory used
 * [1] = fraction of system swap used
 * </pre>
 *
 * @author  Jeff Pang
 * @version $Id: SystemMemoryResource.java,v 1.2 2003/02/18 05:33:20 jeffpang Exp $
 **/
public class SystemMemoryResource extends DynamicResource
{

    private RandomAccessFile mem;
    private int memstart;
    private int memdigits;
    private long memtotal;

    private int swapstart;
    private int swapdigits;
    private long swaptotal;

    public SystemMemoryResource(NodeId id, int sampleInterval) 
	throws IOException
    {
	super(id, "sysmem", new double[] { 0, 0 }, 
	      sampleInterval); // beter way to do this?
	// TODO: how to do this on non-linux
	mem = new RandomAccessFile("/proc/meminfo", "r");
	getDigits();
	sample();
    }

    private void getDigits()
    {
	try {
	    memstart = 0;
	    memdigits = 0;
	    swapstart = 0;
	    swapdigits = 0;

	    int r;
	    mem.seek(0);
	    // where is total mem?
	    while ((r = mem.read()) >= 0) {
		char c = (char)r;
		if (c == '\n') {
		    mem.read(); // M
		    mem.read(); // e
		    mem.read(); // m
		    mem.read(); // :
		    mem.read(); //
		    mem.read(); //
		    memstart += 7;
		    swapstart += 7;
		    while ((r = mem.read()) >= 0) {
			char d = (char)r;
			swapstart++;
			if (Character.isDigit(d)) {
			    memdigits++;
			} else {
			    break;
			}
		    }
		    break;
		}
		memstart++;
		swapstart++;
	    }

	    // where is total swap?
	    while ((r = mem.read()) >= 0) {
		char c = (char)r;
		if (c == '\n') {
		    mem.read(); // S
		    mem.read(); // w
		    mem.read(); // a
		    mem.read(); // p
		    mem.read(); // :
		    mem.read(); //
		    swapstart += 7;
		    while ((r = mem.read()) >= 0) {
			char d = (char)r;
			if (Character.isDigit(d)) {
			    swapdigits++;
			} else {
			    break;
			}
		    }
		    break;
		}
		swapstart++;
	    }
	} catch (Exception ex) {
	    Carp.die("SystemMemoryResource.getDigits: exception: " + ex);
	}

	byte[] mt = new byte[memdigits];
	byte[] st = new byte[swapdigits];
	double val = 0;
	try {
	    mem.seek(memstart);
	    mem.read(mt);
	    memtotal = Long.parseLong(new String(mt));

	    mem.seek(swapstart);
	    mem.read(st);
	    swaptotal = Long.parseLong(new String(st));
	} catch (Exception ex) {
	    Carp.die("SystemMemoryResource.getDigits: exception: " + ex);
	}
	memstart += memdigits;
	swapstart += swapdigits;
    }
    
    /**
     * Take a sample at the current time.
     */
    public void sample()
    {
	final byte[] memused  = new byte[memdigits];
	final byte[] swapused  = new byte[swapdigits];
	final double[] val = new double[2];
	try {
	    int c, i;
	    boolean started;

	    i = 0;
	    started = false;
	    mem.seek(memstart);
	    while ((c = mem.read()) >= 0) {
		if ((char)c == ' ') {
		    if (started) {
			break;
		    } else {
			continue;
		    }
		} else {
		    memused[i] = (byte)c;
		    started = true; i++;
		}
	    }
	    val[0] = (double)Long.parseLong(new String(memused,0,i))/memtotal;

	    i = 0;
	    started = false;
	    mem.seek(swapstart);
	    while ((c = mem.read()) >= 0) {
		if ((char)c == ' ') {
		    if (started) {
			break;
		    } else {
			continue;
		    }
		} else {
		    swapused[i] = (byte)c;
		    started = true; i++;
		}
	    }
	    val[1] = (double)Long.parseLong(new String(swapused,0,i))/swaptotal;
	} catch (Exception ex) {
	    ex.printStackTrace();
	    Carp.die("SystemMemoryResource.sample: exception: " + ex);
	}
	log(val);
    }

    
    public static void main(String[] args) throws Exception
    {
	// Test
	SystemMemoryResource rsc = new SystemMemoryResource(null, 1000);

	while (true) {
	    Thread.currentThread().sleep(rsc.getSampleInterval());
	    rsc.sample();
	    System.out.println(rsc.getLog().util(0) + " " +
			       rsc.getLog().util(1));
	}
    }

}
