/**
 * LatencyPatch.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: LatencyPatch.java,v 1.8 2003/04/29 21:10:16 strib Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;

import java.util.HashMap;
import ostore.util.InvalidConfigFileException;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.util.NodeId;
import ostore.util.Clock;
import ostore.util.Pair;
import ostore.util.PriorityQueue;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;   
import seda.sandStorm.api.StagesInitializedSignal;   

/**
 * A <code>Patch</code> that measures round-trip latency (in microseconds).
 *
 * @author Jeremy Stribling
 * @version    $Id: LatencyPatch.java,v 1.8 2003/04/29 21:10:16 strib Exp $
 */
public class LatencyPatch extends Patch {

    public LatencyPatch( NodeId self_node_id, Classifier classifier, 
			 PatchworkTable table, FaultAdmin fault_admin, 
			 Period period ) {
	super( self_node_id, classifier, table, fault_admin, period );
	class_tag = "LatencyPatch[" + _self_node_id + "]";
	_type = TYPE_LATENCY;
	DEBUG = DebugFlags.NETWORK_PATCH_LATENCY;
    }

    protected void handleProbeResponse( ProbeMsg response, long timeRecvd ) {

	final String tag = class_tag + ".handleProbeResponse";
	
	// once we get a latency response, we should just record the latency
	long timeSent = registerResponse( response );

	// if it's -1, that means we have since timed out or never sent a
	// probe to this guy
	if( timeSent == -1 ) {
	    if( DEBUG ) {
		Debug.printtagln( tag, "Ignoring response: " + response );
	    }
	    return;
	}

	// otherwise, we have our latency
	long latency = timeRecvd - timeSent;
	if( !response.quick_stat ) {
	    makeObservation( response.peer, latency );
	    // we'll adjust the timeout for this node if it's way off
	    // for now, be dumb and if it's less than half the timeout, adjust
	    // remember to adjust for microseconds
	    long curr_timeout = _period.getTimeout( response.peer );
	    if( 3*latency/1000.0 < .5*curr_timeout ) {
		// remember that this sets the timeout for all patches 
		// of this node
		_period.setTimeout( response.peer, curr_timeout - 100 );
		if( DEBUG ) {
		    Debug.printtagln( tag, "Reducing timeout to " + 
				      (curr_timeout-100) + " for node " + 
				      response.peer );
		}
	    }
	} else {
	    handleQuickStatProbeResp( response.peer, latency );
	}

    }

    protected void handleAlarm( AlarmMsg alarm ) {

	final String tag = class_tag + ".handleAlarm";

	if( DEBUG ) {
	    Debug.printtagln( tag, "Got alarm: " + alarm );
	}

	// time to send a probe to someone.
	int seqNo = getAndUpdateSeqNo( alarm.peer );
	sendProbe( new ProbeMsg( alarm.peer, ProbeMsg.MINIMUM_SIZE,
				 _type, seqNo, false ) );
    }

    protected void handleTimeout( ProbeTimeout timeout ) {

	final String tag = class_tag + ".handleTimeout";

	if( registerTimeout( timeout ) ) {
	    // actually a valid timeout, so make a note and an observation
	    if( DEBUG ) {
		Debug.printtagln( tag, "Timed out." );
	    }
	    makeObservation( timeout.peer, LossPatch.VALUE_LOSS, TYPE_LOSS );

	}

    }

}

