/**
 * Patchwork.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: Patchwork.java,v 1.13 2003/06/19 21:38:06 srhea Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;

import java.net.DatagramPacket;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import ostore.network.NetworkMessage;
import ostore.network.NetworkNodeStatusMsg;
import ostore.network.RecordBytesMsg;
import ostore.util.InvalidConfigFileException;
import ostore.util.ByteArrayInputBuffer;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.DebugFlags;
import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.Clock;
import ostore.util.TypeTable;
import seda.nbio.NonblockingDatagramSocket;
import seda.nbio.Selectable;
import seda.nbio.SelectItem;
import seda.nbio.SelectSet;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.UnknownEventException;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SingleThreadedEventHandlerIF;
import seda.sandStorm.api.SinkIF;
import seda.sandStorm.api.SinkFullException;   
import seda.sandStorm.api.StagesInitializedSignal;   
import seda.util.MDWUtil;

/**
 * Stage that measures other nodes on a regular basis, for a variety of
 * specifiable conditions. Is able to measure anything that can be written as
 * an extension to the <code>Patch</code> class.<br><br>
 *
 * Is capable of monitoring nodes on a short-term or long-term basis (see
 * <code>QuickStatReq</code> and <code>MonitorMsg</code> respectively).
 * <br><br>
 *
 * The config file can contain the following options:<br><br>
 *
 * <code>numPatches</code>: The number of patches to be specified in the
 * config file for this stage.  If this is <i>N</i>, then Patch0 
 * through Patch<i>N-1</i> must be specified.  They can be in any order.
 * <br><br>
 *
 * <code>Patch<i>X</i></code>: A specification for a type of measurement
 * module.  This is expected to be in the following form:<br>
 * <code>Patch<i>X</i> patch_name [base_period] [op] [factor]</code><br>
 * Where <code>patch_name</code> can be "loss", "latency", "bandwidth",
 * or "recalc".  The other fields are increasingly optional; they specify
 * the periodic function by which nodes are probed.  <code>base_period</code>
 * is the base probing time in ms (default: 1 second) , <code>op</code> is 
 * the operation (can be "add"/"+", "mult"/"*", "sub"/"-", "div"/"/", or 
 * "exp"/"^") (default: "+"), and <code>factor</code> is the amount by which 
 * the base period is altered for each probing level (default: 0).  
 * For example, if "1000 * 2" is specified, then
 * nodes at probing level 0 get probed every 1 second, level 1 at 2 seconds,
 * level 2 at 4 seconds, level 3 at 8 seconds, etc.  The probing level
 * of a node is specified when a <code>MonitorMsg</code> is sent to Patchwork,
 * either with an explicit level, or an initial measurement value that can 
 * determine the type.<br><br>
 *
 * <code>timeout</code>: Milliseconds after which a probe is considered lost.
 * Default is 10 seconds.<br><br>
 *
 * <code>lossRespondEvery</code>: When monitoring loss rate, how often should
 * a receiver of probes respond.  Specified in number of probes.  The default
 * is 5 (meaning the receiver responds every five packets).<br><br>
 *
 * If another stage wishes to keep track of the long-term statistics, it should
 * create a <code>PatchworkTable<code> when it initializes, then apply deltas
 * to it every time a <code>DeltaMsg</code> is sent out.
 *
 * @author Jeremy Stribling
 * @version    $Id: Patchwork.java,v 1.13 2003/06/19 21:38:06 srhea Exp $
 */
public class Patchwork implements EventHandlerIF, 
				  SingleThreadedEventHandlerIF {

    private static final boolean DEBUG = DebugFlags.NETWORK_PATCHWORK;
    private String class_tag = "Patchwork";

    // config defaults
    private static final long TIMEOUT_DEFAULT = 10000;
    private static final int BASE_PERIOD_DEFAULT = 1000;
    private static final String OP_DEFAULT = "+";
    private static final int FACTOR_DEFAULT = 0;
    private static final int LOSS_RESPOND_EVERY_DEFAULT = 5;
    protected static final int MESSAGE_SIZE = 1500;

    // my node id
    private NodeId _self_node_id;

    // my classifier stuff
    private Classifier _classifier;
    private SinkIF _sink;

    // my table
    private PatchworkTable _table;

    // my fault admin
    private FaultAdmin _fault_admin;

    // measurement modules
    private Patch _patches[];

    // udp socket and thread
    private NonblockingDatagramSocket _udp_socket;
    private UdpListener _udp_thread;

    /**
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void handleEvents(QueueElementIF elemarr[]) 
	throws EventHandlerException
    {

	// make sure all events are handled at the same time (in us)
	long timeReceived;
	if( !Patch.sim_running ) {
	    timeReceived = MDWUtil.currentTimeUsec();
	} else {
	    timeReceived = 1000*Clock.current_date(_self_node_id).getTime();
	}
 
	final String tag = class_tag+".handleEvents";
	if( DEBUG ) Debug.printtagln( tag, "called: "
				      + elemarr.length + " elements" );
	for( int i=0; i<elemarr.length; i++ ) {
	    handleEvent( elemarr[i], timeReceived );
	}
	if( DEBUG ) Debug.printtagln( tag, "done" );
    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void handleEvent(QueueElementIF element) 
	throws EventHandlerException
    {

	if( !Patch.sim_running ) {
	    handleEvent( element, MDWUtil.currentTimeUsec() );
	} else {
	    handleEvent( element, 
			 1000*Clock.current_date(_self_node_id).getTime() );
	}
    }

    /**
     * Handle events given the exact time they were received by this stage
     */
    private void handleEvent( QueueElementIF element, long timeReceived ) {
      try
      {

	final String tag = class_tag+".handleEvent";
	if( DEBUG ) Debug.printtagln( tag, "called" );

        if (DEBUG) Debug.printtagln (tag, "got " + element.getClass () +
                " event at " + timeReceived + " ms: " + element + ".");

	if( element instanceof StagesInitializedSignal ) {

	    // do nothing at the moment

	} else if( element instanceof NetworkNodeStatusMsg ) {

	    // notify the recalc patch of the loss
	    if( _patches[Patch.TYPE_RECALC] != null ) {
		RecalcPatch rcp = (RecalcPatch) _patches[Patch.TYPE_RECALC];
		rcp.handleNetworkNodeStatusMsg((NetworkNodeStatusMsg) element);
	    } else if( DEBUG ) {
		Debug.printtagln( tag, "Got NetworkNodeStatusMsg: " + element
				  + ", but have no recalc patch so ignoring.");
	    }

	} else if( element instanceof InjectFaultMsg ) {	

	    _fault_admin.updateFaults( (InjectFaultMsg) element );

	} else if( element instanceof MonitorMsg ) {

	    handleMonitorMsg( (MonitorMsg) element );

	} else if( element instanceof MonitorAllMsg ) {

	    handleMonitorAllMsg( (MonitorAllMsg) element );

	} else if( element instanceof MonitorStopMsg ) {

	    handleMonitorStopMsg( (MonitorStopMsg) element );

	} else if( element instanceof PatchMsg ) {

	    // send it to the right patch
	    PatchMsg pm = (PatchMsg) element;
	    byte type = pm.getType();

	    if( _patches[type] == null ) {
		// bad message
		if( DEBUG ) {
		    Debug.printtagln( tag, "Patch " + type + 
				      " missing, ignoring message" );
		}
	    } else {
		_patches[type].handleEvent( pm, timeReceived );
	    }

	} else {

	  Carp.error( "Dropping unexpected message: " + element );

	}

	if( DEBUG ) Debug.printtagln( tag, "done" );
      }
      catch(Exception e)
      {
        Carp.dieGracefully(e, _self_node_id); 
      }
    }	

    private void handleMonitorMsg( MonitorMsg msg ) {

	final String tag = class_tag+".handleMonitorMsg";

	// start monitoring these nodes in all meausurement patches
	for( int i = 0; i < msg.peers.length; i++ ) {
	    
	    // if the node has values, initialize values:
	    if( msg.values != null ) {
		// this call returns false if node is already being monitored
		_table.initStat( msg.peers[i], msg.type, msg.values[i] );
	    }

	    // if there are no levels given, try to figure one out
	    int level;
	    if( msg.levels == null ) {
		level = determineLevel( msg.type, msg.values[i] );
	    } else {
		level = msg.levels[i];
	    }

	    // now start monitoring
	    for( int j = 0; j < _patches.length; j++ ) {
		if( _patches[j] != null ) {
		    if( DEBUG ) {
			Debug.printtagln( tag, "adding " + msg.peers[i] +
					  " to the " + j + " patch." );
		    }
		    _patches[j].addNode( msg.peers[i], level, msg.userData, 
					 msg.sink );
		}
	    }

	}

    }

    private void handleMonitorStopMsg( MonitorStopMsg msg ) {

	final String tag = class_tag+".handleMonitorStopMsg";

	// stop monitoring these nodes in all meausurement patches
	for( int i = 0; i < msg.peers.length; i++ ) {
	    
	    for( int j = 0; j < _patches.length; j++ ) {
		if( _patches[j] != null ) {
		    if( DEBUG ) {
			Debug.printtagln( tag, "removing " + msg.peers[i] +
					  " from the " + j + " patch." );
		    }
		    _patches[j].removeNode( msg.peers[i], msg.userData, 
					    msg.sink );
		}
	    }

	}

    }

    private void handleMonitorAllMsg( MonitorAllMsg msg ) {

	final String tag = class_tag+".handleMonitorAllMsg";

	for( int j = 0; j < _patches.length; j++ ) {
	    if( _patches[j] != null ) {
		if( DEBUG ) {
		    Debug.printtagln( tag, "monitoring all on the " +
				      j + " patch." );
		}
		_patches[j].addMonitorAll( msg.userData, msg.sink );
	    }
	}

    }

    /**
     * Figure out the probing level of a node, given an initial value of a 
     * certain type.
     */
    private int determineLevel( byte type, long value ) {

	switch( type ) {
	case Patch.TYPE_LOSS:
	    // simple:
	    if( value < 10 ) {
		return 0;
	    } else if( value < 50 ) {
		return 1;
	    } else if( value < 75 ) {
		return 2;
	    } else {
		return 3;
	    }
	case Patch.TYPE_LATENCY:
	    // simple heuristic for now:
	    if( value < 20000 ) {
		return 0;
	    } else if( value < 100000 ) {
		return 1;
	    } else if( value < 500000 ) {
		return 2;
	    } else {
		return 3;
	    }
	case Patch.TYPE_BANDWIDTH:
	    // just for now:
	    return 0;
	default:
	    Carp.die( "Bad type passed to determineLevel: " + type );
	    return -1;
	}

    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void init(ConfigDataIF config) throws Exception {
	final String tag = class_tag+".init";
	if( DEBUG ) Debug.printtagln( tag, "called" );

	///////////////////
	// get config info

	Patch.sim_running = config.getBoolean( "simulator_running" );
	Patch.use_tcp = config.getBoolean( "use_tcp" );
	_self_node_id = new NodeId( config.getString( "node_id" ) );
	_classifier = Classifier.getClassifier( _self_node_id );
	_sink = config.getStage().getSink();
	_table = new PatchworkTable();
	_fault_admin = new FaultAdmin( config.getBoolean( "acceptFaults" ) );

	String str_timeout = config.getString( "timeout" );
	long timeout;
	if( str_timeout == null ) {
	    timeout = TIMEOUT_DEFAULT;
	} else {
	    timeout = Long.parseLong( str_timeout );
	    if( timeout < 0 ) {
		Carp.die( "Negative value given for timeout in " + 
			  "Patchwork config: " + timeout );
	    }
	}
	
	int lossRespondEvery = config.getInt( "lossRespondEvery" );
	if( lossRespondEvery == -1 ) {
	    lossRespondEvery = LOSS_RESPOND_EVERY_DEFAULT;
	} else if( lossRespondEvery < 0 ) {
	    Carp.die( "Non-positive value given for lossRespondEvery in " + 
		      "Patchwork config: " + lossRespondEvery );
	}

	int numPatches = config.getInt( "numPatches" );
	if( numPatches < 0 ) {
	    Carp.die( "Must specify numPatches in Patchwork config." );
	}

	_patches = new Patch[ Patch.NUM_TYPES ];
	for( int i = 0; i < numPatches; i++ ) {
	    String[] patchConfig = config.getStringList("Patch" + i);
	    if( patchConfig == null ) {
		Carp.die( "No config info for Patch" + i );
	    }

	    // structure of config info:
	    // <patch_name> <base_period> <op> <factor> <patch_specifics>
	    byte type = getPatchType( patchConfig[0] );
	    // can't have two patches of the same type
	    if( _patches[type] != null ) {
		Carp.die( "Two patches of type " + type + " declared." );
	    } else {
		_patches[type] = makePatch( type, patchConfig, timeout, 
					    lossRespondEvery );
	    }
	    
	}
	if( !Patch.sim_running && !Patch.use_tcp ) {
	    startUdpSocket( _self_node_id.port()+1 );
	}
	class_tag += "[" + _self_node_id + "]";

	//////////////////
	// initialize filters

	// Subscribe to all my favorite events

	String[] eventNames = {
	    "seda.sandStorm.api.StagesInitializedSignal",
	    "ostore.network.NetworkNodeStatusMsg",
	    "ostore.network.patchwork.ProbeTimeout",
	    "ostore.network.patchwork.AlarmMsg",
	    "ostore.network.patchwork.QuickStatReq",
	    "ostore.network.patchwork.QuickStatTimeout",
	    "ostore.network.patchwork.MonitorMsg",
	    "ostore.network.patchwork.MonitorAllMsg",
	    "ostore.network.patchwork.MonitorStopMsg"
	};

	String[] inb_eventNames = {
	    "ostore.network.patchwork.ProbeMsg",
	    "ostore.network.patchwork.InjectFaultMsg"
	};

	Filter filter;

	// require local events
	for( int i = 0; i < eventNames.length; i++ ) {
	  filter = new Filter();
	  Class eventClass;
          if( DEBUG ) Debug.printtagln( tag, "Finding class: " +
                                               eventNames[i] );
          try {
            eventClass = Class.forName( eventNames[i] );
	    if( DEBUG ) Debug.printtagln( tag, "Requiring class: " +
	  				       eventNames[i] );
	    if( !filter.requireType( eventClass ) || 
	        !_classifier.subscribe( filter, _sink ) ) {

	      Carp.die( "Could not require class " + eventNames[i] ); 

	    } 

          } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
          }

        }

	// require inbounders
	for( int i = 0; i < inb_eventNames.length; i++ ) {
	  filter = new Filter();
	  Class eventClass;
          if( DEBUG ) Debug.printtagln( tag, "Finding class: " +
                                               inb_eventNames[i] );
          try {
            eventClass = Class.forName( inb_eventNames[i] );
	    if( DEBUG ) Debug.printtagln( tag, "Requiring class: " +
	  				       inb_eventNames[i] );
	    if( !filter.requireType( eventClass ) ||
		!filter.requireValue ( "inbound", new Boolean(true) ) || 
	        !_classifier.subscribe( filter, _sink ) ) {

	      Carp.die( "Could not require class " + inb_eventNames[i] ); 

	    } 

          } catch ( ClassNotFoundException cnfe ) {
            Carp.die( cnfe );
          }

        }


    }

    private byte getPatchType( String name ) {

	if( name.equalsIgnoreCase( "loss" ) ) {
	    return Patch.TYPE_LOSS;
	} else if( name.equalsIgnoreCase( "latency" ) ) {
	    return Patch.TYPE_LATENCY;
	} else if( name.equalsIgnoreCase( "bandwidth" ) ) {
	    return Patch.TYPE_BANDWIDTH;
	} else if( name.equalsIgnoreCase( "recalc" ) ) {
	    return Patch.TYPE_RECALC;
	} else {
	    Carp.die( "No name translation found for patch " + name );
	    return -1;
	}

    }

    private Patch makePatch( byte type, String[] config, long timeout, 
			     int lossRespondEvery ) {

	Period per;
	Patch rtPatch;

	int base_period = BASE_PERIOD_DEFAULT;
	String op = OP_DEFAULT;
	int factor = FACTOR_DEFAULT;
	
	if( config.length > 1 ) {
	    String str_base_period = config[1];
	    base_period = Integer.parseInt( str_base_period );
	    if( base_period < 0 ) {
		Carp.die( "Bad base period: " + base_period );
	    }
	}

	if( config.length > 2 ) {
	    op = config[2];
	}

	if( config.length > 3 ) {
	    String str_factor = config[3];
	    factor = Integer.parseInt( str_factor );
	}
	
	per = new Period( _self_node_id, base_period, op, factor, timeout );

	switch( type ) {
	case Patch.TYPE_LOSS: 
	    rtPatch =  new LossPatch( _self_node_id, _classifier, 
				      _table, _fault_admin, 
				      per, lossRespondEvery );
	    break;
	case Patch.TYPE_LATENCY:
	    rtPatch = new LatencyPatch( _self_node_id, _classifier, 
					_table, _fault_admin, per );
	    break;
	case Patch.TYPE_BANDWIDTH: 
	    rtPatch = null; // just for now
	    break;
	case Patch.TYPE_RECALC:
	    rtPatch = new RecalcPatch( _self_node_id, _classifier, 
				       _table, _fault_admin, per );
	    break;
	default:
	    rtPatch = null;
	    Carp.die( "Unknown patch type: " + type );
	} 

	return rtPatch;

    }

    private void startUdpSocket( int port ) {

	final String tag = class_tag+".startUdpSocket";
	try {
	    _udp_socket = new NonblockingDatagramSocket(port);
	} catch( java.io.IOException e ) {
	    Carp.die( "Got exception " + e + 
		      " when trying to create UDP socket." );
	}
	
	// now, just start a thread that's always listening to the port
	_udp_thread = new UdpListener();
	new Thread(_udp_thread).start();
    }

    /**
     * An internal method used for dispatching events
     */
    protected void dispatch( QueueElementIF event ) {

	final String tag = class_tag+".dispatch";
	if( DEBUG ) Debug.printtagln( tag, "called with event " + event );
	
	try {
	    _classifier.dispatch( event );
	} catch ( SinkFullException sfe ) {
	    Carp.die( "SinkFull: " + sfe );
	}                                         
	
    }

    /** 
     * Specified by seda.sandStorm.api.EventHandlerIF 
     */
    public void destroy()
    {
        final String tag = class_tag+".destroy";
	if( !Patch.sim_running && !Patch.use_tcp ) {
	    _udp_thread.running = false;
	}
        if( DEBUG ) Debug.printtagln( tag, "destroyed" );
    }

    public class UdpListener implements Runnable {

	public boolean running;

	public void run() {
	    running = true;
	    final String tag = class_tag+".UdpListener.run";
	    SelectSet read_selset = new SelectSet();
	    SelectItem read_selitem = 
		new SelectItem( _udp_socket, 
				(short)Selectable.READ_READY);
	    read_selset.add(read_selitem);

	    // for now, set the buffer size at 1500 bytes.
	    // ProbeMsgs should not be longer than this, however if they
	    // are they will be truncated.  That spells trouble.
	    byte[] data = new byte[MESSAGE_SIZE];
	    DatagramPacket dp = new DatagramPacket( data, MESSAGE_SIZE ); 
	    while( running ) {
		// now just loop through and read packets as they come in
		// NOTE: here I'm assuming these UDP messages are always 
		// less than one packet.  Hopefully this won't lead to any
		// trouble.
		while ((read_selitem.revents & Selectable.READ_READY) == 0) {
		    read_selset.select(-1); // block while waiting
		}
		if (DEBUG) Debug.printtagln( tag, "A read is ready." );
		// reset it after getting an event
		read_selitem.revents = 0;
		dp.setLength(MESSAGE_SIZE);
		int size = 0;
		try {
		    size = _udp_socket.nbReceive(dp);
		} catch( java.io.IOException e ) {
		    Carp.die( "Got IOException: " + e + 
			      " while nbReceiving on the udp socket." );
		}
		if( size == 0 ) {
		    // no packet was received for some reason, try again
		    if( DEBUG ) {
			Debug.printtagln( tag, "nbReceive returned 0" );
		    }
		    continue;
		}
		// otherwise, go ahead and unserialize the message
		InputBuffer ib = new ByteArrayInputBuffer( dp.getData(), 
							   0, size );
		NetworkMessage msg = null;   
		try {
                    NodeId wire_peer = new NodeId (ib);
                    NodeId wire_sender = new NodeId (ib);
                    msg = (NetworkMessage) TypeTable.new_object (ib);
		    msg.inbound = true;
                    msg.peer = wire_peer;
                    msg.sender = wire_sender;
		    // tell the network layer about these bytes
		    dispatch( new RecordBytesMsg( msg, size ) );
		    // now go ahead and handle this message
		    // we're assuming here that the peer has been set by the 
		    // sender, since we can't really know who it was
		    handleEvent( msg );
		}
		catch (Exception e) {
		    Carp.die( "Couldn't serialize or handle a message, " + 
			      "got exception " + e );
		}
	    }
	    _udp_socket.close();
	}

    }

}

