/**
 * PatchworkStats.java
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;

import ostore.util.InputBuffer;
import ostore.util.NodeId;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.QSException;
import ostore.util.StatRecorder;

/**
 * <p>
 * The <code>PatchworkStats</code> data structure contains info describing
 * thevalues of some statistic, both overall and current (over the last
 * measurement window).
 *
 *
 * @author	Jeremy Stribling
 * @version	$Id: PatchworkStats.java,v 1.4 2003/03/24 00:13:13 geels Exp $
 */
public class PatchworkStats implements QuickSerializable {

    // subject of stats
    private NodeId _node_id;
    private byte _type;

    // internal StatRecorder
    private StatRecorder _recorder;

    // have we done anything yet?
    private boolean _done_once = false;

    // the start value (0 if none)
    private long _start_value = 0;

    // the number of stats used in the last round, 
    // to get the "current" stats
    private long _count;

    // overall stats recorded
    private long _over_min = Long.MAX_VALUE;
    private long _over_max = 0;
    private double _over_mean = 0;
    private double _over_geo_mean = 0;

    // current stats recorded
    private long _curr_min = Long.MAX_VALUE;
    private long _curr_max = 0;
    private long _curr_median = 0;
    private double _curr_mean = 0;
    private double _curr_geo_mean = 0;

    public PatchworkStats( NodeId node_id, byte type) {
	_recorder = new StatRecorder( "" );
	_node_id = node_id;
	_type = type;
	_count = 0;
    }

    public PatchworkStats( NodeId node_id, byte type, long start_value ) {
	this( node_id, type );
	_start_value = start_value;
	_over_min = start_value;
	_over_max = start_value;
	_over_mean = (double) start_value;
	_over_geo_mean = (double) start_value;
	_curr_min = start_value;
	_curr_max = start_value;
	_curr_median = start_value;
	_curr_mean = (double) start_value;
	_curr_geo_mean = (double) start_value;
	_done_once = true;
    }

    /**
     * Specified by ostore.util.QuickSerializable 
     */
    public PatchworkStats( InputBuffer buffer ) throws QSException {
	_node_id = (NodeId) buffer.nextObject();
	_type = buffer.nextByte();
	_recorder = (StatRecorder) buffer.nextObject();
	_curr_min = buffer.nextLong();
	_curr_max = buffer.nextLong();	
	_curr_median = buffer.nextLong();
	_curr_mean = buffer.nextDouble();
	_curr_geo_mean = buffer.nextDouble();
	_over_min = buffer.nextLong();
	_over_max = buffer.nextLong();	
	_over_mean = buffer.nextDouble();
	_over_geo_mean = buffer.nextDouble();
	_done_once = buffer.nextBoolean();
	_count = buffer.nextLong();
    }

    /**
     * Specified by ostore.util.QuickSerializable 
     */
    public synchronized void serialize (OutputBuffer buffer)
    {
	buffer.add(_node_id);
	buffer.add(_type);
	buffer.add(_recorder);
	buffer.add(_curr_min);
	buffer.add(_curr_max);
	buffer.add(_curr_median);
	buffer.add(_curr_mean);
	buffer.add(_curr_geo_mean);
	buffer.add(_over_min);
	buffer.add(_over_max);
	buffer.add(_over_mean);
	buffer.add(_over_geo_mean);
	buffer.add(_done_once);
	buffer.add(_count);
    }

    protected void add( StatDelta delta ) {
	if( delta == null ) {
	    return;
	}
	_curr_min += delta.curr_min;
	_curr_max += delta.curr_min;	
	_curr_median += delta.curr_median;
	_curr_mean += delta.curr_mean;
	_curr_geo_mean += delta.curr_geo_mean;
	_over_min += delta.over_min;
	_over_max += delta.over_max;	
	_over_mean += delta.over_mean;
	_over_geo_mean += delta.over_geo_mean;
	_done_once = true;
    }

    protected void addStat( long value ) {
	_recorder.add( value );
    }

    public long getOverallMin() {
	return _over_min;
    }

    public long getOverallMax() {
	return _over_max;
    }

    public double getOverallMean() {
	return _over_mean;
    }

    public double getOverallGeometricMean() {
	return _over_geo_mean;
    }

    public long getCurrentMin() {
	return _curr_min;
    }

    public long getCurrentMax() {
	return _curr_max;
    }

    public long getCurrentMedian() {
	return _curr_median;
    }

    public double getCurrentMean() {
	return _curr_mean;
    }

    public double getCurrentGeometricMean() {
	return _curr_geo_mean;
    }

    /**
     * Returns the number of samples used in the last recalculation,
     * to obtain the "current" values.
     */
    public long getCount() {
	return _count;
    }

    protected StatDelta recalculate( double alpha, double threshold ) {

	// if there haven't been any sample, return an empty delta
	if( _recorder.getNumSamples() == 0 ) {
	    return null;
	}

	StatDelta delta = new StatDelta( _node_id, _type );
	
	// get the current values and clear
	long curr_min_tmp = _recorder.getMin();
	long curr_max_tmp = _recorder.getMax();
	long curr_median_tmp = _recorder.getMedian();
	double curr_mean_tmp = _recorder.getMean();
	double curr_geo_mean_tmp = _recorder.getGeometricMean();
	_count = _recorder.getNumSamples();
	_recorder.reset();
	boolean set_delta = false;
	
	long l_diff;
	double d_diff;

	// figure out the deltas
	l_diff = curr_min_tmp - _curr_min;
	if( ((double)Math.abs(l_diff))/((double)_curr_min) > threshold ) {
	    delta.curr_min = l_diff;
	    _curr_min = curr_min_tmp;
	    set_delta = true;
	}
	l_diff = curr_max_tmp - _curr_max;
	if( ((double)Math.abs(l_diff))/((double)_curr_max) > threshold ) {
	    delta.curr_max = l_diff;
	    _curr_max = curr_max_tmp;
	    set_delta = true;
	}
	l_diff = curr_median_tmp - _curr_median;
	if( ((double)Math.abs(l_diff))/((double)_curr_median) > threshold ) {
	    delta.curr_median = l_diff;
	    _curr_median = curr_median_tmp;
	    set_delta = true;
	}
	d_diff = curr_mean_tmp - _curr_mean;
	if( Math.abs(d_diff)/_curr_mean > threshold ) {
	    delta.curr_mean = d_diff;
	    _curr_mean = curr_mean_tmp;
	    set_delta = true;
	}
	d_diff = curr_geo_mean_tmp - _curr_geo_mean;
	if( Math.abs(d_diff)/_curr_geo_mean > threshold ) {
	    delta.curr_geo_mean = d_diff;
	    _curr_geo_mean = curr_geo_mean_tmp;
	    set_delta = true;
	}

	// now figure out overall values
	if( _curr_min < _over_min ) {
	    l_diff = _curr_min - _over_min;
	    if( ((double)Math.abs(l_diff))/((double)_over_min) > threshold ) {
		delta.over_min = l_diff;
		_over_min = _curr_min;
		set_delta = true;
	    }
	}
	if( _curr_max > _over_max ) {
	    l_diff = _curr_max - _over_max;
	    if( ((double)Math.abs(l_diff))/((double)_over_max) > threshold ) {
		delta.over_max = l_diff;
		_over_max = _curr_max;
		set_delta = true;
	    }
	}
	if( _done_once ) {
	    // exponentially weighted arithmetic mean
	    double over_mean_tmp = alpha*_curr_mean + (1.0-alpha)*_over_mean;
	    // exponentially weighted geometric mean
	    double over_geo_mean_tmp  = Math.pow( _curr_geo_mean, alpha ) *
		Math.pow( _over_geo_mean, 1.0 - alpha );

	    // if it beats the threshold, use it
	    d_diff = over_mean_tmp - _over_mean;
	    if( Math.abs(d_diff)/_over_mean > threshold ) {
		delta.over_mean = d_diff;
		_over_mean = over_mean_tmp;
		set_delta = true;
	    }
	    d_diff = over_geo_mean_tmp - _over_geo_mean;
	    if( Math.abs(d_diff)/_over_geo_mean > threshold ) {
		delta.over_geo_mean = d_diff;
		_over_geo_mean = over_geo_mean_tmp;
		set_delta = true;
	    }

	} else {

	    // if this is the first time, and there was a starting value,
	    // it needs to be added to all the diffs, so that updated models
	    // can have the same statistics
	    if( _start_value != 0 ) {
		delta.curr_min += _start_value;
		delta.curr_max += _start_value;
		delta.curr_mean += _start_value;
		delta.curr_geo_mean += _start_value;
		delta.curr_median += _start_value;
		delta.over_min += _start_value;
		delta.over_max += _start_value;
		delta.over_mean += _start_value;
		delta.over_geo_mean += _start_value;
	    }
	    
	    _over_mean = _curr_mean;
	    delta.over_mean = _curr_mean;
	    _over_geo_mean = _curr_geo_mean;
	    delta.over_geo_mean = _curr_geo_mean;
	    set_delta = true;
	    _done_once = true;
	}

	if( set_delta ) {
	    return delta;
	} else {
	    return null;
	}

    }

    public String toString() {
	
	String s = "";

	s += "\nStats for " + _node_id + " of type " + _type + ":\n";
	s += "Overall max=" + getOverallMax() + "/min=" + getOverallMin() +
	    "/mean=" + (getOverallMean()*100.0)/100.0 + "/geo_mean=" 
	    + (getOverallGeometricMean()*100.0)/100.0 + "\n";
	s += "Current max=" + getCurrentMax() + "/min=" + getCurrentMin() +
	    "/mean=" + (getCurrentMean()*100.0)/100.0 + "/geo_mean=" 
	    + (getCurrentGeometricMean()*100.0)/100.0 + "/median=" + 
	    getCurrentMedian() + "\n\n";

	return s;

    }

    public static class StatDelta implements QuickSerializable {

	// subject of stats
	public NodeId node_id;
	public byte type;

	// overall stats recorded
	public long over_min;
	public long over_max;
	public double over_mean;
	public double over_geo_mean;
	
	// current stats recorded
	public long curr_min;
	public long curr_max;
	public long curr_median;
	public double curr_mean;
	public double curr_geo_mean;
	
	public StatDelta( NodeId node_id, byte type) {
	    this.node_id = node_id;
	    this.type = type;
	}

	/**
	 * Specified by ostore.util.QuickSerializable 
	 */
	public StatDelta( InputBuffer buffer ) throws QSException {
	    node_id = (NodeId) buffer.nextObject();
	    type = buffer.nextByte();
	    curr_min = getLStat(buffer);
	    curr_max = getLStat(buffer);	
	    curr_median = getLStat(buffer);
	    curr_mean = getDStat(buffer);
	    curr_geo_mean = getDStat(buffer);
	    over_min = getLStat(buffer);
	    over_max = getLStat(buffer);
	    over_mean = getDStat(buffer);
	    over_geo_mean = getDStat(buffer);
	}

	private double getDStat( InputBuffer buffer ) throws QSException {
	    if( buffer.nextBoolean() ) {
		return buffer.nextDouble();
	    } else {
		return 0.0;
	    }
	}

	private long getLStat( InputBuffer buffer ) throws QSException {
	    if( buffer.nextBoolean() ) {
		return buffer.nextLong();
	    } else {
		return 0;
	    }
	}

	/**
	 * Specified by ostore.util.QuickSerializable 
	 */
	public synchronized void serialize (OutputBuffer buffer)
	{
	    buffer.add(node_id);
	    buffer.add(type);	    
	    addStat(buffer, curr_min);
	    addStat(buffer, curr_max);
	    addStat(buffer, curr_median);
	    addStat(buffer, curr_mean);
	    addStat(buffer, curr_geo_mean);
	    addStat(buffer, over_min);
	    addStat(buffer, over_max);
	    addStat(buffer, over_mean);
	    addStat(buffer, over_geo_mean);
	}
	
	private void addStat( OutputBuffer buffer, long value ) {
	    if( value == 0 ) {
		buffer.add( false );
	    } else {
		buffer.add( true );
		buffer.add( value );
	    }
	}

	private void addStat( OutputBuffer buffer, double value ) {
	    if( value == 0.0 ) {
		buffer.add( false );
	    } else {
		buffer.add( true );
		buffer.add( value );
	    }
	}

    }
}
