/**
 * PatchworkTable.java
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;

import ostore.util.QuickSerializable;
import ostore.util.QSException;
import ostore.util.InputBuffer;
import ostore.util.QSVector;
import ostore.util.QSWindow;
import ostore.util.TypeTable;
import ostore.util.Types;
import ostore.util.ByteUtils;
import ostore.util.QSArray;
import ostore.util.QSHashMap;
import ostore.util.Carp;
import ostore.util.Debug;
import ostore.util.NodeId;
import ostore.util.QSInt;
import ostore.util.OutputBuffer;
import ostore.oil.Matrix;
import ostore.oil.Array;
import ostore.oil.Model;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;

/**
 * <p>
 * The <code>PatchworkTable</code> is built by the <code>Patchwork</code>
 * code, and is used for collecting statistics about other nodes. 
 * It remembers observations sent to it and when <code>recalculate</code> 
 * is called it produces a <code>PatchworkTable.Delta</code> that contains 
 * the difference between the current model and the last one produced from 
 * a call to <code>recalculate</code>. </p><p>
 *
 * It combines new and old statistics using weighted function,
 * where the new statistic is considered as important as the "alpha"
 * value, default at 80%.</p><p>
 *
 * It only considers a change worthy of noting if it is greater than 
 * some threshold, default at 1%.</p><p>
 *
 * Currently, the measurement types are as
 * follows:</p><p>
 *
 * <b>latency (in microseconds)</b>: <code>Patch.LATENCY</code></p><p>
 *
 * <b>bandwidth</b>: <code>Patch.BANDWIDTH</code></p><p>
 *
 * <b>loss</b> (in percent, from 0 to 100, where 100 is complete loss): 
 *             <code>Patch.LOSS</code></p><p>
 *
 *
 * @author	Jeremy Stribling
 * @version	$Id: PatchworkTable.java,v 1.4 2003/03/24 00:13:13 geels Exp $
 */

public class PatchworkTable implements QuickSerializable {

    private final static boolean DEBUG = 
	ostore.util.DebugFlags.NETWORK_PATCHWORKTABLE;

    private final static String class_tag = "PatchworkTable";

    private final static int MAX_OBSERVATIONS = 10000;

    // Types of possible measurements
    private static final int NUM_MEASUREMENTS = 3;

    // the least percentage change a number must be from the current model
    // to count as a delta worthy of saving.
    private double _threshold = .01;

    // how much to emphasize the newest average in the calculations
    private double _alpha = .8;

    // a measurement matrix
    private QSHashMap _map;

    // a vector of nodes to be removed at the next recalculation
    private QSVector _to_be_removed;

    /**
     * Create a new <code>PatchworkTable</code>.
     */
    public PatchworkTable() {

	_map = new QSHashMap();
	_to_be_removed = new QSVector();

    }

    public PatchworkTable( InputBuffer buffer ) throws QSException {
	_map = (QSHashMap) buffer.nextObject();
	_to_be_removed = (QSVector) buffer.nextObject();
    }

    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    /**
     * Specified by ostore.oil.Model interface 
     */
    protected void record( Observation[] observations )
    {
	for( int i=0; i<observations.length; i++ ) {
	    record( observations[i] );
	}
    }

    /**
     * Specified by ostore.oil.Model interface 
     */
    protected synchronized void record( Observation obs )
    {

        final String tag = class_tag+".record";
        if( DEBUG ) Debug.printtagln( tag, "called with observation "
                                      + obs );
	QSArray nodeArray = (QSArray) _map.get( obs.peer );

	// if there are no stats for this node yet, make them
	if( nodeArray == null ) {
	    nodeArray = new QSArray( NUM_MEASUREMENTS );
	    _map.put( obs.peer, nodeArray );
	}

	// now add this stat
	PatchworkStats stats = (PatchworkStats) nodeArray.get( obs.type );
	if( stats == null ) {
	    stats = new PatchworkStats( obs.peer, obs.type );
	    nodeArray.set( obs.type, stats );
	}

	// add stat
	stats.addStat( obs.value );

    }

    /**
     * Get the current statistic data structure for a node.
     *
     * @param peer the node you're interested in
     * @param type  the type of statistic desired (see class description
     *              for valid types.
     */
    public synchronized PatchworkStats getStats( NodeId peer, byte type ) {

	QSArray nodeArray = (QSArray) _map.get( peer );

	if( nodeArray == null ) {
	    return null;
	}

	return (PatchworkStats) nodeArray.get( type );

    }

    /**
     * Initialize a value for some statistic, only if no values
     * have yet been recorded.
     *
     * @return whether this succeeded
     */
    protected synchronized boolean initStat( NodeId peer, byte type, long value ) {

	QSArray nodeArray = (QSArray) _map.get( peer );

	// if there are no stats for this node yet, make them
	if( nodeArray == null ) {
	    nodeArray = new QSArray( NUM_MEASUREMENTS );
	    _map.put( peer, nodeArray );
	}

	// now add this stat
	PatchworkStats stats = (PatchworkStats) nodeArray.get( type );
	if( stats == null ) {
	    stats = new PatchworkStats( peer, type, value );
	    nodeArray.set( type, stats );
	    return true;
	} else {
	    return false;
	}

    }

    /**
     * Tells the model to remove all statistics for this peer at the
     * time of the next recalculation.  This is best called when it is known
     * for certain that a node wil not be observed ever again, or at least
     * for a long time.
     *
     * @param peer      the node for which to remove the statistics
     */
    protected synchronized void remove( NodeId peer ) {

	_map.remove( peer );

	// make sure it's not already in it
	if( !_to_be_removed.contains( peer ) ) {
	    _to_be_removed.add( peer );
	}

    }

    /**
     * Update this copy of a table with a delta produced by Patchwork
     */
    public synchronized void add_delta( Delta d ) {

        final String tag = class_tag+".add_delta";
        if( DEBUG ) Debug.printtagln( tag, "called" );

	// removed those who need it
	QSVector removeThem = d.getToBeRemoved();
	for( int i = 0; i < removeThem.size(); i++ ) {
	    _map.remove( removeThem.get( i ) );
	}

	QSVector deltas = d.getDeltas();
	
	for( int i = 0; i < deltas.size(); i++ ) {
	    PatchworkStats.StatDelta currD = 
		(PatchworkStats.StatDelta) deltas.get(i); 

	    
	    // look up the stats, if it's not there, make it
	    QSArray nodeArray = (QSArray) _map.get( currD.node_id );
	    
	    // if there are no stats for this node yet, make them
	    if( nodeArray == null ) {
		nodeArray = new QSArray( NUM_MEASUREMENTS );
		_map.put( currD.node_id, nodeArray );
	    }

	    // now add this stat
	    PatchworkStats stats = (PatchworkStats) nodeArray.get(currD.type);
	    if( stats == null ) {
		stats = new PatchworkStats( currD.node_id, currD.type );
		nodeArray.set( currD.type, stats );
	    }

	    stats.add( currD );
	    
	}

    }

    /**
     * Specified by ostore.oil.Model interface 
     */
    protected synchronized Delta recalculate()
    {

        final String tag = class_tag+".recalculate";
        if( DEBUG ) Debug.printtagln( tag, "called" );

	Delta newDelta = new Delta( _to_be_removed );
	_to_be_removed = new QSVector();

	// to recalculate, go through every stats and recalculate
	Iterator keys = _map.keySet().iterator();
	while( keys.hasNext() ) {

	    NodeId currNode = (NodeId) keys.next();
	    QSArray currArray = (QSArray) _map.get( currNode );
	    for( int i = 0; i < NUM_MEASUREMENTS; i++ ) {
		PatchworkStats currStat = (PatchworkStats) currArray.get(i);
		if( currStat != null ) {
		    PatchworkStats.StatDelta d = 
			currStat.recalculate( _alpha, _threshold );
		    if( d != null ) {
			newDelta.addDelta( d );
		    }
		}
	    }

	}

	return newDelta;

    }

    /**
     * How important is the current value of a statistic when calculating
     * overall statistics?
     */
    protected double getAlpha( ) {
	return _alpha;
    }

    /**
     * How much does a statistic need to change for it to be considered worthy
     * of recording?
     */
    protected double getThreshold( ) {
	return _threshold;
    }

    protected void setAlpha( double newAlpha ) {
	_alpha = newAlpha;
    }

    protected void setThreshold( double newThresh ) {
	_threshold = newThresh;
    }

    /**
     * Specified by ostore.util.QuickSerializable 
     */
    public synchronized void serialize (OutputBuffer buffer) 
    {
	buffer.add(_map);
	buffer.add(_to_be_removed);
    }

    /**
     * Returns a string representation of the object 
     */
    public synchronized String toString() {

	String s = "PatchworkTable (current means):\n" + 
	    "\t\t\t\tLATENCY\t\tBANDWIDTH\tLOSS\n";
	
	Iterator iter = _map.keySet().iterator();
	while( iter.hasNext() ) {
	
	    NodeId peer = (NodeId) iter.next();
	    QSArray nodeArray = (QSArray) _map.get( peer );

	    double value;
	    PatchworkStats stats;

	    if( nodeArray == null ) {
		continue;
	    }

	    s += "\t" + peer + "\t";
	    
	    stats = (PatchworkStats) nodeArray.get( Patch.TYPE_LATENCY );
	    if( stats == null ) {
		s += "---\t\t";
	    } else {
		value = stats.getCurrentMean();
		s += Math.floor(value*100.0)/100.0 + "\t\t";
	    }

	    stats = (PatchworkStats) nodeArray.get( Patch.TYPE_BANDWIDTH );
	    if( stats == null ) {
		s += "---\t\t";
	    } else {
		value = stats.getCurrentMean();
		s += Math.floor(value*100.0)/100.0 + "\t\t";
	    }

	    stats = (PatchworkStats) nodeArray.get( Patch.TYPE_LOSS );
	    if( stats == null ) {
		s += "---\n";
	    } else {
		value = stats.getOverallMean();
		s += Math.floor(value*100.0)/100.0 + "\n";
	    }	

	}

	return s;

    }

    ////////////////////////////////////////////////////////////////
    ////////////////////     INNER CLASSES	////////////////////

    /**
     * A <code>PatchworkTable.Delta</code> object contains a diff 
     * of the sufficient statistics for a <code>PatchworkTable</code>.
     *
     * @author	Jeremy Stribling
     * @version	$Id: PatchworkTable.java,v 1.4 2003/03/24 00:13:13 geels Exp $
     */

    public static class Delta implements QuickSerializable {
	
	private QSVector _deltas;
	private QSVector _to_be_removed;
	
	////////////////////    CONSTRUCTORS	////////////////////

	/**
	 * Construct a <code>Delta</code> from its
	 * <code>QuickSerializable</code> form.
	 */
	public Delta( InputBuffer buffer ) throws QSException {
	    _deltas = (QSVector) buffer.nextObject();
	    _to_be_removed = (QSVector) buffer.nextObject();
	}

	/** 
	 * Construct a new <code>Delta</code>.
	 *
	 * @param toBeRemoved  a <code>QSVector</code> of <code>NodeId</code>s
	 *                     that should be removed from the 
	 *                     <code>Table</code>.
	 */
	public Delta( QSVector toBeRemoved ) {
	    _deltas = new QSVector();
	    _to_be_removed = toBeRemoved;
	}

	/**
	 */
	public void addDelta( PatchworkStats.StatDelta d ) {
	    _deltas.add( d );
	}

	/**
 	 * Returns a <code>QSVector</code> containing all the delta 
	 * information.
	 *
	 * @return the <code>QSVector</code> of deltas
	 */
	public QSVector getDeltas() {
	    return _deltas;
	}

	/**
	 * Get the list of <code>NodeId</code>s whose statistics should be 
	 * removed from the <code>Table</code>
	 */
	public QSVector getToBeRemoved() {
	    return _to_be_removed;
	}

	/**
	 * Get the list of NodeIds that have a change in a certain type of 
	 * statistics (i.e. Patch.TYPE_LATENCY)
	 */
	public NodeId[] getNodesChanged( byte type ) {
	    Vector tmp = new Vector();
	    for( int i = 0; i < _deltas.size(); i++ ) {
		PatchworkStats.StatDelta psd = 
		    (PatchworkStats.StatDelta) _deltas.get(i);
		if( psd.type == type ) {
		    tmp.add(psd.node_id);
		}
	    }
	    NodeId[] nodes = new NodeId[tmp.size()];
	    for( int i = 0; i < nodes.length; i++ ) {
		nodes[i] = (NodeId) tmp.get(i);
	    }
	    return nodes;
	}

	/**
	 * Specified by ostore.util.QuickSerializable 
	 */
	public synchronized void serialize (OutputBuffer buffer)
	{
	    buffer.add (_deltas);
	    buffer.add (_to_be_removed);
	}
	

	/**
	 * Clone this delta 
	 */
	public Object clone() throws CloneNotSupportedException {

	    Delta newDelta = new Delta( _to_be_removed );

	    newDelta._deltas = (QSVector) _deltas.clone();
	    newDelta._to_be_removed = (QSVector) _to_be_removed.clone();

	    return newDelta;

	}

	/**
	 * Returns a String representation of this <code>Delta</code> 
	 */
	public String toString() {

	    String s = "Delta:\t\t\t\tLATENCY\t\tBANDWIDTH\tLOSS\n";

	    return s;

	}

    }	// end class Delta
    

    /**
     * A <code>Patchwork.Observation</code> object contains an
     * observation recorded by the patchwork stage.
     *
     * @author	Jeremy Stribling
     * @version	$Id: PatchworkTable.java,v 1.4 2003/03/24 00:13:13 geels Exp $
     */

    public static class Observation {
	
	protected NodeId peer;
	protected long value;
	protected byte type;
	
	/**
	 * Record the observed double value of type <code>type</code>
	 * for the specified node id <code>peer</code>.
 	 * @param observed_peer  the node to which the statistic belongs
  	 * @param value   the value of the observed statistic
  	 * @param type    the type of statistic (i.e. LATENCY, etc)
	 */
	public Observation( NodeId observed_peer, long value, byte type ) {

	    peer = observed_peer;
	    this.value = value;
	    this.type = type;

	}

	/** 
	 * Returns a string representation of this 
         * <code>Observation</code> 
	 */
	public String toString() {

	    String s = "";
	    s += "{peer: " + peer + ", ";
	    s += "value: " + value + ", ";
	    s += "type: " + type + "}";

	    return s;

	}

    }	// end class Observation

}
