/**
 * ProbeMsg.java
 *
 * @author Jeremy Stribling
 * @version    $Id: ProbeMsg.java,v 1.3 2003/03/24 00:13:13 geels Exp $
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.network.patchwork;
import ostore.network.NetworkMessage;
import ostore.util.ByteUtils;
import ostore.util.NodeId;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QuickSerializable;
import ostore.util.QSException;
import ostore.tapestry.api.TapestryRouteMsg;

/**
 * This is a probing message sent out by Patchwork
 * in order to measure statistics of other nodes.
 * NOTE: This class cannot be extended, because the structure of the
 * serialized message (the extra filler) doesn't allow for it.
 *
 * @author Jeremy Stribling
 * @version    $Id: ProbeMsg.java,v 1.3 2003/03/24 00:13:13 geels Exp $
 */
public final class ProbeMsg extends NetworkMessage 
    implements QuickSerializable, PatchMsg {

    private int[] _filler;

    /**
     * The smallest size a <code>ProbeMsg</code> can be, given all its fields
     */
    public static final int MINIMUM_SIZE = 10;

    /**
     * The type of measurement being probed.  See <code>Patch</code>.
     */
    public byte type;

    /**
     * The current sequence number of this probe.
     */
    public int seq_no;
    
    /**
     * Is this a response to another probe?
     */
    public boolean response;
    /**
     * Should the receiver respond?
     */
    public boolean do_respond;
    /**
    * Is this a quick stat probe?
    */
    public boolean quick_stat;

    /**
    * Type-specific data
    */
    public int data;

    public ProbeMsg (NodeId peer, int totalSize, byte type, 
		     int seqNo, boolean response) {
	this( peer, totalSize, type, seqNo, response, false );
    }

    /**
     * Constucts a probing message of a certain type, with a given
     * approximate size.  The size needs to be at least 
     * <code>MINIMUM_SIZE</code>, and if difference between it and 
     * <code>MINIMUM_SIZE</code> is not divisible by 4 it will be 
     * rounded down to a number that is.
     *
     * @param peer  the node to probe
     * @param totalSize   how big the message should be
     * @param type  the type of the message
     * @param seqNo  the sequence number of this probe in the current stream 
     *               of probes to <code>peer</code>
     * @param response  whether this is a response
     * @param quickStat   whether this probe should be treated as "quick"
     */
    public ProbeMsg (NodeId peer, int totalSize, byte type, 
		     int seqNo, boolean response, boolean quickStat ) {

	super (peer, false /* outbound */);

	this.response = response;
	this.quick_stat = quickStat;
	this.seq_no = seqNo;

	// make sure the sender knows it needs to be at least some bytes
	if( totalSize < MINIMUM_SIZE ) {
	  throw new RuntimeException( "Size of ProbeMsg must " +
				       "be at least " + MINIMUM_SIZE + 
				       "bytes" );
	}

	this.type = type;

	// make sure the fillerSize is divisible by 4
	int fillerSize = totalSize - MINIMUM_SIZE;
	int remainder = fillerSize % 4;
	fillerSize -= remainder;

	// make filler to fill the rest of the array size
	_filler = new int[ fillerSize/4 ];

    }

    public ProbeMsg (InputBuffer buffer) throws QSException {
	super( buffer );
	type = buffer.nextByte();
	seq_no = buffer.nextInt();
	{
	    byte bitmap = buffer.nextByte();
	    response = ( (bitmap & 0x01) != 0 );
	    do_respond = ( (bitmap & 0x02) != 0 );
	    quick_stat = ( (bitmap & 0x04) != 0 );
	}
	data = buffer.nextInt();
    }

    public void serialize (OutputBuffer buffer) {
	super.serialize( buffer );
	buffer.add( type );
	buffer.add( seq_no );
	{
	    // Store all the booleans in a single byte.
	    byte bitmap = 0;
	    if( response ) { bitmap |= 0x01; }
	    if( do_respond ) { bitmap |= 0x02; }
	    if( quick_stat ) { bitmap |= 0x04; }
	    buffer.add( bitmap );
	}
	buffer.add( data );
	for (int i = 0; i < _filler.length; ++i)
	    buffer.add (_filler [i]);
    }

    public Object clone () throws CloneNotSupportedException {
	ProbeMsg newone = (ProbeMsg) super.clone ();
	newone.type = type;
	newone.seq_no = seq_no;
	newone.response = response;
	newone.do_respond = do_respond;
	newone.quick_stat = quick_stat;
	newone.data = data;
	newone._filler = new int[_filler.length];
	return newone;
    }

    public byte getType() {
	return type;
    }

    public String toString () {
	return "(ProbeMsg, peer " + peer + ", type " + type + 
	    ", fillerSize " + /*_filler.length +*/ ", response: " + 
	    response + ", do_respond: " + do_respond + ", quick_stat: " + 
	    quick_stat +", seqNo: " + seq_no + " " + ", data " 
	    + data + ")";
    }
}
