/**
 * SignedQS.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security;
import ostore.util.*;
import ostore.dataobj.*;
import java.security.*;

/**
 * A class to provide the same basic functionality as
 * java.security.SignedObject, but using OceanStore's QuickSerializable
 * class rather than the Java serialization mechanism.  
 * 
 * <p>
 * This class serves two purposes.  It encapsulates the code for generating
 * and serializing digital certificates and the signatures over them, and
 * it handles the encoding of the particular algorithm used to sign them so
 * that they can be verified after decoding.  This latter functionality is
 * accomplished with the aid of a helper class, {@link QSSignature}, which
 * encodes sufficient information about an object of type 
 * {@link java.security.Signature} to allow an equivalent object to be
 * reproduced on decode.
 *
 * <p> 
 * Note that the decoding operation of this class <i>does not</i> verify
 * the signature, since the receiving node may not recognize the algorithm
 * used to create the signature.  Instead, the functions
 * {@link #verify(java.security.PublicKey, java.security.Signature)}
 * and {@link #user_data()} are separated.  The former should be called before
 * trusting the data from the latter, although in some cases the data may
 * be useful even if the signature cannot be verified.  Alternatively,
 * since the verification operation is expensive, it may be worth first
 * checking the relevance of the data before bothering to verify its
 * integrity or source.
 *
 * <p>
 * Finally, the user of this class should take care to include the public
 * key of the pair used sign this message in the signed data, either in
 * full or as a digest, as otherwise certain attacks are possible.
 *
 * @author   Sean C. Rhea
 * @version  $Id: SignedQS.java,v 1.35 2002/08/16 00:00:09 srhea Exp $
 */
public class SignedQS implements QuickSerializable {

    public class SignatureBuffer extends OutputBufferImpl {
	
	public Signature engine;

	public SignatureBuffer (Signature engine) {
	    this.engine = engine;
	}

	public void add( byte input ) {
	    try { 
		engine.update (input);
	    }
	    catch (SignatureException e) {
		throw new AssertionViolatedException (
			"signature object should be initialized.");
	    }
	}

	public void add (byte[] input, int offset, int length) {
	    try { 
		engine.update (input, offset, length);
	    }
	    catch (SignatureException e) {
		throw new AssertionViolatedException (
			"signature object should be initialized.");
	    }
	}
    }

    protected SignedQS () {}

    protected final void init (
	    QuickSerializable user_data,
	    QSSignature qs_signature,
	    PrivateKey privkey,
	    SecureRandom random,
	    Signature engine) throws InvalidKeyException, SignatureException {

	if (! qs_signature.algorithm ().equals (engine.getAlgorithm ())) {
	    throw new InvalidKeyException ();
	}

	_user_data = user_data;
	_qs_signature = qs_signature;

	if (random != null)
	    engine.initSign (privkey, random);
	else
	    engine.initSign (privkey);

	SignatureBuffer sb = new SignatureBuffer (engine);
	sb.add (_user_data);
	// unnecessary: sb.add (_qs_signature);

	_signature = engine.sign ();
    }

    protected final void init (
	    QuickSerializable user_data,
	    SecureHash user_data_digest,
	    PrivateKey privkey,
	    RawNativeRSASignature_PKCS1_SHA1 engine) 
	    throws InvalidKeyException, SignatureException {

	_user_data = user_data;
	_qs_signature = new QSSha1RsaSignature ();

	engine.engineInitSign (privkey);
	_signature = engine.engineSign (user_data_digest);
    }

    /**
     * The standard constructor.  This constructor actually produces a
     * signature, so if <code>user_data</code> is changed after this object
     * is constructed, those changes will <i>not</i> be reflected in the
     * signed message.
     *
     * @param user_data the message to sign
     * @param privkey   the private key to sign the message with
     * @param random    a random number generator.  If this parameter is
     *                  <code>null</code>, the default random number 
     *                  generator will be used.
     * @param engine    a signature generating engine such that 
     * <pre>provider  ().equals (engine.getProvider ().getName ())</pre>
     * and
     * <pre>algorithm ().equals (engine.getAlgorithm ())</pre>
     *                  These can be obtained through the function {@link
     *                  java.security.Signature#getInstance(String)} or
     *                  {@link java.security.Signature#getInstance(String,
     *                  String)}.
     */
    public SignedQS (
	    QuickSerializable user_data,
	    QSSignature qs_signature,
	    PrivateKey privkey,
	    SecureRandom random,
	    Signature engine) throws InvalidKeyException, SignatureException {

	init (user_data, qs_signature, privkey, random, engine);
    }

    public SignedQS (
	    QuickSerializable user_data,
	    SecureHash user_data_digest,
	    PrivateKey privkey,
	    RawNativeRSASignature_PKCS1_SHA1 engine) 
	    throws InvalidKeyException, SignatureException {

	init (user_data, user_data_digest, privkey, engine);
    }

    public SignedQS (InputBuffer buffer) throws QSException {
	_user_data = buffer.nextObject ();
	try {
	    _qs_signature = (QSSignature) buffer.nextObject ();
	}
	catch (TypeTable.NoSuchTypeCode e) {
	    _qs_signature = null;
	    _qs_signature_type = e.type_code;
	}
	int siglen = buffer.nextInt ();
	_signature = new byte [siglen];
	buffer.nextBytes (_signature, 0, siglen);
    }

    /**
     * The data which this object signs.  Please see the note about this in
     * {@link #SignedQS (QuickSerializable, KeyPair, SecureRandom,
     * Signature)}.  Also, note that
     * {@link #verify(java.security.PublicKey, java.security.Signature)}
     * should be called to verify that the signature is correct.
     *
     * @return the data which is signed
     */
    public QuickSerializable user_data () { 
	return _user_data; 
    }

    /**
     * The type of signature over this object.
     * 
     * @return a subtype QSSignature such that if an object of type 
     *         {@link java.security.Signature} is generated where with a
     *         matching algorithm and provider, that engine can be safely
     *         passed to the verify function without fear of the latter
     *         throwing an exception.
     *
     * @exception TypeTable.NoSuchTypeCode if the type code included in the
     *            encoding of this message does not correspond to a known
     *            type on this host.  In other words, if the message was
     *            signed with an unknown algorithm.
     */
    public QSSignature qs_signature () throws TypeTable.NoSuchTypeCode {
	if (_qs_signature == null)
	    throw new TypeTable.NoSuchTypeCode (_qs_signature_type);
	else 
	    return _qs_signature;
    }

    public void serialize (OutputBuffer buffer) {
	buffer.add (_user_data);
	// I would like to combine the signature type with its length,
	// since generally length is a function of type, but if the
	// receiving party is not aware of the type, this technique would
	// prohibit them from properly decoding any of the remaining
	// information in the message.  Thus, we waste a few bytes and
	// include the length explicitly.
	buffer.add (_qs_signature);
	buffer.add (_signature.length);
	buffer.add (_signature);
    }

    /**
     * Verify that the signature over {@link #user_data()} is valid.
     *
     * @param pubkey The public key of the alleged signer of this message.
     * @param engine a signature generating engine such that 
     * <pre>provider  ().equals (engine.getProvider ().getName ())</pre>
     * and
     * <pre>algorithm ().equals (engine.getAlgorithm ())</pre>
     *                  These can be obtained through the function {@link
     *                  java.security.Signature#getInstance(String)} or
     *                  {@link java.security.Signature#getInstance(String,
     *                  String)}.
     *
     * @return <code>true</code> if the signature is valid,
     *         <code>false</code> otherwise
     *
     * @exception TypeTable.NoSuchTypeCode if the type code included in the
     *            encoding of this message does not correspond to a known
     *            type on this host.  In other words, if the message was
     *            signed with an unknown algorithm.
     */
    public boolean verify (PublicKey pubkey, Signature engine) 
    throws InvalidKeyException, SignatureException, TypeTable.NoSuchTypeCode {

	// make sure the signature type is decoded
	qs_signature ();

	if (! _qs_signature.algorithm ().equals (engine.getAlgorithm ()))
	    throw new InvalidKeyException ();

	// Check the signature
	engine.initVerify (pubkey);

	SignatureBuffer sb = new SignatureBuffer (engine);
	sb.add (_user_data);
	// unnecessary: sb.add (_qs_signature);

	return engine.verify (_signature);
    }

    /** 
     * A verify function that can take advantage of the
     * RawNativeRSASignature_PKCS1_SHA1 class.
     */
    public boolean verify (PublicKey pubkey, 
	    RawNativeRSASignature_PKCS1_SHA1 engine,
	    SecureHash digest) 
    throws InvalidKeyException, SignatureException, TypeTable.NoSuchTypeCode {

	// make sure the signature type is decoded
	qs_signature ();

	if (! _qs_signature.algorithm ().equals (
		    (new QSSha1RsaSignature ()).algorithm ()))
	    throw new InvalidKeyException ();

	// Check the signature
	engine.engineInitVerify (pubkey);
	return engine.engineVerify (_signature, digest);
    }

    public String toString () {
	return "(SignedQS " + user_data () + ")";
    }

    protected QuickSerializable _user_data;
    protected byte [] _signature;
    protected QSSignature _qs_signature;
    protected long _qs_signature_type;

    /**
     * Test harness for SignedQS.
     */
    public static void main (String [] args) throws Exception {

	Types.registerAllTypes ();

	System.out.println ("Generating key pair...");
	KeyPair keypair = null;
	Signature engine = null;

	int keysize = Integer.parseInt (args [0]);
	Security.addProvider (new cryptix.jce.provider.CryptixCrypto ());
	KeyPairGenerator kpgen = 
	    KeyPairGenerator.getInstance ("RSA", "CryptixCrypto");
	kpgen.initialize(keysize);
	keypair = kpgen.generateKeyPair();

	if (NativeRSAAlgorithm.available) {
	    Security.addProvider (new ostore.security.NativeRSAProvider ());
	    engine = Signature.getInstance (
		    "SHA-1/RSA/PKCS#1", "NativeRSA");
	}
	else {
	    engine = Signature.getInstance (
		    "SHA-1/RSA/PKCS#1", "CryptixCrypto");
	}

        PRCert message = new PRCert (new SHA1Hash ("Hello, world."),
		new QSRsaPublicKey (keypair.getPublic ()));

	System.out.println ("Creating signed message...");
	SignedQS signed_message = null;
	long start = System.currentTimeMillis ();
	for (int i = 0; i < 1000; ++i) {
	    try {
		signed_message = new SignedQS (
			message, new QSSha1RsaSignature (), 
			keypair.getPrivate (), null, engine);
	    }
	    catch (Exception e) {
		Carp.out ("Could not generate signed message: " + e);
		System.exit (1);
	    }
	}
	long stop = System.currentTimeMillis ();
	System.out.println ("Average signing time = " + 
		(((double) stop-start) / 1000.0) + " ms");

	System.out.println ("QSCloning signed message...");
	SignedQS decoded_message = (SignedQS) QSClone.qs_clone (signed_message);
	System.out.println ("Verifying cloned message...");

	start = System.currentTimeMillis ();
	for (int i = 0; i < 1000; ++i) {
	    try {
		if (! decoded_message.verify (keypair.getPublic (), engine)) {
		    Carp.out ("Signature did not verify.");
		    System.exit (1);
		}
	    }
	    catch (Exception e) {
		Carp.out ("Could not verify decoded signed message:" + e);
		System.exit (1);
	    }
	}
	stop = System.currentTimeMillis ();
	System.out.println ("Average verifying time = " + 
		(((double) stop-start) / 1000.0) + " ms");

	if (! decoded_message.user_data ().equals (message)) {
	    Carp.out ("Message was not the same.");
	    System.exit (1);
	}

	System.out.println ("Success!");
	System.exit (0);
    }
}

