/**
 * SignedQSTree.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security;
import java.io.*;
import java.security.*;
import java.util.*;
import ostore.util.*;
import ostore.dataobj.*;
import ostore.security.thresh.*;

/**
 * SignedQSTree
 *
 * @author   Sean C. Rhea
 * @version  $Id: SignedQSTree.java,v 1.15 2002/08/16 00:00:09 srhea Exp $
 */
public class SignedQSTree extends SignedQS {

    public static class Root implements QuickSerializable {
	int elt_cnt;
	SecureHash left, right;
	
	Root (int c, SecureHash l, SecureHash r) {
	    elt_cnt = c; left = l; right = r;
	}

	public Root (InputBuffer buffer) throws QSException {
	    elt_cnt = buffer.nextInt ();
	    left = (SecureHash) buffer.nextObject ();
	    right = (SecureHash) buffer.nextObject ();
	}

	public void serialize (OutputBuffer buffer) {
	    buffer.add (elt_cnt);
	    buffer.add (left);
	    buffer.add (right);
	}

	public boolean equals (Object rhs) {
	    if (! getClass ().equals (rhs.getClass ())) return false;
	    Root other = (Root) rhs;
	    return (elt_cnt == other.elt_cnt) &&
		left.equals (other.left) &&
		right.equals (other.right);
	}
		
	public int hashCode () {
	    return elt_cnt ^ left.hashCode () ^ right.hashCode ();
	}
    }

    public static int next_highest_power_of_two (int i) {
	// (((i | ((i - 1) << 1)) + 1) >> 1) ?
	int result = 2;
	while (result < i) result *= 2;
	return result;
    }

    protected SignedQSTree () {}

    public SignedQSTree (
	    QSArray elts,
	    QSSignature qs_signature,
	    PrivateKey privkey,
	    SecureRandom random,
	    Signature engine) throws InvalidKeyException, SignatureException {

	_elts = elts;
	init (compute_root (_elts), qs_signature, privkey, random, engine);
    }

    public static SecureHash [] compute_tree (QSArray elts) {
	//System.err.println ("compute tree");
	int leaves = next_highest_power_of_two (elts.length ());
	//System.err.println ("leaves = " + leaves);
	SecureHash [] tree = new SecureHash [2*leaves-1];
	int j = 0;
	for (int i = tree.length - leaves; i < tree.length; ++i) {
	    if (j < elts.length ()) 
		tree [i] = new SHA1Hash ((QuickSerializable) elts.get (j));
	    else 
		tree [i] = new SHA1Hash ();
	    //System.out.println ("tree[" + i + "] = " + tree [i]);
	    ++j;
	}
	for (int i = tree.length - leaves - 1; i > 0; --i) {
	    SecureHashBuffer b = new SHA1HashBuffer ();
	    // Don't want typecodes, so call like this.
	    tree [i*2+1].serialize (b);
	    tree [i*2+2].serialize (b);
	    tree [i] = b.toHash ();
	    //System.out.println ("tree[" + i + "] = " + tree [i]);
	}
	return tree;
    }

    public static Root compute_root (QSArray elts) {
	SecureHash [] tree = compute_tree (elts);
	return new Root (elts.length (), tree [1], tree [2]);
    }
	
    public SignedQSTree (InputBuffer buffer) throws QSException {
	super (buffer);
	_elts = (QSArray) buffer.nextObject ();
    }

    public QuickSerializable user_data () { 
	return _elts;
    }



    public void serialize (OutputBuffer buffer) {
	super.serialize (buffer);
	buffer.add (_elts);
    }

    public boolean verify (PublicKey pubkey, Signature engine) 
    throws InvalidKeyException, SignatureException, TypeTable.NoSuchTypeCode {
	
	Root r = compute_root (_elts);
	if (! r.equals (_user_data)) return false;
	return super.verify (pubkey, engine);
    }

    public String toString () {
	return "(SignedQSTree " + user_data () + ")";
    }

    protected QSArray _elts;

    public static void main (String [] args) throws Exception {
	String tag = "SignedQSTree";

	Types.registerAllTypes ();

	Security.addProvider (new cryptix.jce.provider.CryptixCrypto ());
	Security.addProvider (new ThresholdProvider ());

	ThresholdPublicKey public_key = null;
	String keyfilename = args [0];
	try {
	    FileInputStream keyfile = new FileInputStream (keyfilename);
	    public_key = (ThresholdPublicKey)  QSIO.read (keyfile);
	    keyfile.close ();
	}
	catch (Exception e) {
	    Carp.die (tag + ": caught exception " + e +
			" while trying to read pr keys from disk.");
	}

	KeyShare[] keys = new KeyShare [4];
	for (int i = 0; i < 4; ++i) {
	    String ksfilename = args [i+1];
	    try {
		FileInputStream keyfile = new FileInputStream (ksfilename);
		keys [i] = (KeyShare)  QSIO.read (keyfile);
		keyfile.close ();
	    }
	    catch (Exception e) {
		Carp.die (tag + ": caught exception " + e +
			" while trying to read " + ksfilename + " from disk.");
	    }
	}
	int NUM_OBJS = Integer.parseInt (args [5]);
	int k = 2;

	long qs_sig_type = TypeTable.type_code ((new ThresholdSignature ()));
	Random rand = new Random (1);

	SignedQS bar = null;
	SignedQS [] foo = null;
	{
	    QSArray qs_array = new QSArray (NUM_OBJS);
	    for (int i = 0; i < NUM_OBJS; ++i) {
		byte[] data = new byte[1024];
		rand.nextBytes (data);
		qs_array.set (i, new QSByteArray (data));
	    }

	    SigShare[] sigs = new SigShare[k];

	    for (int i=0; i<k; i++)
		sigs[i] = ThresholdSignedQSTree.create_sig_share (
			keys[i], qs_array, qs_sig_type);

	    byte [] signature = SigShare.package_up (sigs, public_key);
	    SignedQS signed_data = 
		new ThresholdSignedQSTree (qs_array, signature);

	    bar = signed_data;
	    foo = new SignedQS [NUM_OBJS];
	    for (int i = 0; i < NUM_OBJS; ++i) {
		foo [i] = new SignedQSTreePath (
			(SignedQSTree) signed_data, i);
	    }
	}

	Signature engine = null;
	try {
	    engine = Signature.getInstance ("Threshold", "Threshold");
	}
	catch (Exception e) {
	    if (engine == null) Carp.die( "Caught exception " + e + 
		    " while instanciating signature " +
		    "engine.  Dying...");
	}
	if (engine == null) Carp.die( "Could not instanciate signature " +
		"engine.  Dying...");

	
	SignedQS signed_data = (SignedQS) QSClone.qs_clone (bar);
	if (signed_data.verify (public_key, engine))
	    System.out.println ("Signature Correctly Verified");
	else 
	    Carp.die ("Signature Verification Failed!");

	for (int i = 0; i < NUM_OBJS; ++i) {
	    SignedQS path = (SignedQS) QSClone.qs_clone (foo [i]);
	    if (path.verify (public_key, engine))
		System.out.println ("Signature Correctly Verified");
	    else 
		Carp.die ("Signature Verification Failed!");
	}
    }
}

