/**
 * KeyShare.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.security.thresh;
import cryptix.jce.provider.random.DevRandom;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.math.BigInteger;
import java.security.*;
import java.util.Random;
import ostore.security.SafeBigInt;
import ostore.util.ByteUtils;
import ostore.util.Carp;
import ostore.util.QuickSerializable;
import ostore.util.QuickSerializable;
import ostore.util.InputBuffer;
import ostore.util.OutputBuffer;
import ostore.util.QSByteArray;
import ostore.util.QSException;
import ostore.util.QSIO;
import ostore.util.QuickSerializable;
import ostore.util.SHA1Hash;
import ostore.util.TypeTable;

/**
 * A Secret Key Share for an RSA (k,l) Threshold Scheme.
 *
 * <p>Reference: "Practical Threshold Signatures",
 * Victor Shoup (sho@zurich.ibm.com), IBM Research Paper RZ3121, 4/30/99
 *
 * @author  <a href=mailto:sweis@theory.lcs.mit.edu>Steve Weis</a> 
 * @version $Id: KeyShare.java,v 1.23 2002/08/02 19:39:56 geels Exp $
 */
public class KeyShare implements QuickSerializable {

    private static final String class_tag = "KeyShare";

    static {
	final String tag = class_tag + ".static";
	Provider cryptixRandom = new cryptix.jce.provider.CryptixRandom ();
	Security.addProvider (cryptixRandom);
	if (!DevRandom.isAvailable ()) Carp.die( tag + ": need DevRandom.");
    }

    public KeyShare (InputBuffer buffer) throws QSException {
	secret = (SafeBigInt) buffer.nextObject ();
	verifier = (SafeBigInt) buffer.nextObject ();
	groupVerifier = (SafeBigInt) buffer.nextObject ();
	n = (SafeBigInt) buffer.nextObject ();
	delta = (SafeBigInt) buffer.nextObject ();
	signVal = (SafeBigInt) buffer.nextObject ();
	id = buffer.nextInt ();
    }

    public void serialize (OutputBuffer buffer) {
	buffer.add (secret);
	buffer.add (verifier);
	buffer.add (groupVerifier);
	buffer.add (n);
	buffer.add (delta);
	buffer.add (signVal);
	buffer.add (id);
    }

// Constants and variables
//............................................................................
    private final static SafeBigInt TWO = SafeBigInt.valueOf(2l);
    private final static SafeBigInt FOUR = SafeBigInt.valueOf(4l);

    /** An arbitrary security parameter for generating secret shares */
    private final static int L1 = 128;

    /** Secret key value */
    private SafeBigInt secret;

    /** Verifier used to authenticate self to other shares */
    private SafeBigInt verifier;
    private SafeBigInt groupVerifier;

    // TODO: This information should be moved to a GroupKey object
    // It's redundant to put into every share
    private SafeBigInt n;
    private final SafeBigInt delta;

    /** The secret key value used to sign messages. Should not be exposed */
    private SafeBigInt signVal;

    // TODO: Maybe get a better id scheme. Use verifier
    private int id;

// Constructors
//............................................................................
   
    /**
     * Create a new share
     *
     * @param id - the identifier of this share
     * @param secret - a secret value generated by a Dealer

     * @param n - the modulo of the group public key
     * @param delta - l! (group size factorial)
     */
    public KeyShare(int id, SafeBigInt secret,
		    SafeBigInt n, SafeBigInt delta)
    {
	this.id = id;
	this.secret = secret;

	// Verifier must be set later using setVerifier()
	// TODO: Merge Dealer.generateShares and Dealer.generateVerifiers
	// and generate them simultaneously
	this.verifier = null;

	this.n = n;
	this.delta = delta;
	this.signVal = FOUR.multiply(delta).multiply(secret);
    }

// Public Methods
//............................................................................

    public int getId() { return id; }

    public SafeBigInt getSecret() { return secret; }


    public void setVerifiers(SafeBigInt verifier, SafeBigInt groupVerifier) {
	this.verifier = verifier;
	this.groupVerifier = groupVerifier;
    }

    public SafeBigInt getVerifier() { return verifier; }

    public SafeBigInt getSignVal() { return signVal; }

   
    public String toString() {
	return "KeyShare[" + id + "]";
    }

    // public SigShare sign(SafeBigInt x) {
    //	return new SigShare(this.id, x.modPow(signVal,n));
    //    }

    /**
     * Create a SigShare and a Verifier for byte[] b<BR>
     *
     * Refer to Shoup pg. 8 <BR>
     * @return a sig share with a verifier
     */
    public SigShare sign(byte [] b) {
	
	final String tag = class_tag + ".sign";
	
	SecureRandom random = null;
	try {
	    random = SecureRandom.getInstance ("DevRandom", "CryptixRandom");
	    if (random == null) Carp.die( tag + ": need DevRandom.");
	}
	catch (Exception e) {
	    if (random == null) Carp.die( tag + ": caught " + e + 
		    " while trying to init DevRandom.");
	}
    
	SafeBigInt x = (new SafeBigInt(b)).mod(n);

	int randbits = n.bitLength()+3*L1;

	// r \elt (0, 2^L(n)+3*l1)
	SafeBigInt r = (new SafeBigInt(randbits,random));

	SigShare from_java = null, from_c = null;

	if (NativeIF.available) {
	    from_c = sign_using_c (x, n, r, false);
	}
	
	if (DEBUG || !NativeIF.available) {

	    SafeBigInt vprime = groupVerifier.modPow(r,n);
	    SafeBigInt xtilde = x.modPow(FOUR.multiply(delta),n);
	    SafeBigInt xprime = xtilde.modPow(r,n);

	    SafeBigInt c = null;
	    SafeBigInt z = null;
	    SafeBigInt v = null;
	    SafeBigInt vi = null;
	    SafeBigInt xi = null;
	    SafeBigInt xisqr = null;
	    byte [] dig = null;
	    // Try to generate C and Z
	    try {
		MessageDigest md = MessageDigest.getInstance("SHA");
		md.reset();

		v = groupVerifier.mod(n);
		md.update(v.toByteArray ());

		md.update(xtilde.toByteArray());

		vi = verifier.mod(n);
		md.update(vi.toByteArray());

		xi = x.modPow(signVal,n);
		xisqr = xi.modPow(TWO,n);
		md.update(xisqr.toByteArray());

		md.update(vprime.toByteArray());

		md.update(xprime.toByteArray());

		dig = md.digest ();
		c = new SafeBigInt (dig).mod(n);
		z = (c.multiply(secret)).add(r);
	    } catch (Exception e) { e.printStackTrace(); }

	    Verifier ver = new Verifier(this.id,z,c,verifier,groupVerifier);

	    from_java = new SigShare(this.id, xi, ver);

	    if (DEBUG && NativeIF.available && ! from_c.equals (from_java)) {
		System.out.println ("NOT EQUAL");
		System.out.println ("java: x = " + x.toString (16));
		System.out.println ("java: n = " + n.toString (16));
		System.out.println ("java: r = " + r.toString (16));
		System.out.println ("java: grouperifier = " + groupVerifier.toString (16));
		System.out.println ("java: delta = " + delta.toString (16));
		System.out.println ("java: verifier = " + verifier.toString (16));
		System.out.println ("java: signval = " + signVal.toString (16));
		System.out.println ("java: secret = " + secret.toString (16));
		System.out.println ("java: v = " + ByteUtils.print_bytes (v.toByteArray ()));
		System.out.println ("java: xtilde = " +
		ByteUtils.print_bytes (xtilde.toByteArray
		()));
		System.out.println ("java: vi = " + ByteUtils.print_bytes (vi.toByteArray ()));
		System.out.println ("java: xi^2 = " + ByteUtils.print_bytes (xisqr.toByteArray ()));
		System.out.println ("java: v' = "+ ByteUtils.print_bytes (vprime.toByteArray ()));
		System.out.println ("java: x' = " + ByteUtils.print_bytes (xprime.toByteArray ()));
		
		System.out.println ("java: hash = " + ByteUtils.print_bytes (dig));
		System.out.println ("java: c = " + c.toString (16));
		System.out.println ("java: z = " + z.toString (16));
		System.out.println ("java: xups = " + xi.toString (16));
		System.out.flush ();

		SigShare dummy = sign_using_c (x, n, r, true);

		System.exit (1);
	    }
	}

	return (NativeIF.available ? from_c : from_java);
    }

    public SigShare sign_using_c (SafeBigInt x, SafeBigInt n, SafeBigInt r, boolean debug) {
	BigInteger [] results = { null, null, null };
	NativeIF.sign_using_c_helper (
		x.toBigInteger (), n.toBigInteger (), r.toBigInteger (), 
		groupVerifier.toBigInteger (), delta.toBigInteger (), 
		verifier.toBigInteger (), signVal.toBigInteger (), 
		secret.toBigInteger (), results, debug);
	SafeBigInt x_modpow_signval_n = new SafeBigInt (results [0]);
	SafeBigInt z = new SafeBigInt (results [1]);
	SafeBigInt c = new SafeBigInt (results [2]);

	Verifier ver = new Verifier(this.id,z,c,verifier,groupVerifier);
	return new SigShare(this.id, x_modpow_signval_n, ver);
    }

    // Debugging
    //............................................................................

    private final static boolean DEBUG = false;

    private static void debug(String s)
    {
	System.err.println("KeyShare: " + s);
    }

    static void main (String [] args) throws Exception {
	Security.addProvider (new cryptix.jce.provider.CryptixCrypto ());
	Security.addProvider (new ThresholdProvider ());
	ostore.util.Types.registerAllTypes ();
	long qs_sig_type = TypeTable.type_code ((new ThresholdSignature ()));
	KeyShare ks = (KeyShare) QSIO.read ("sample_keyshare");

	byte[] data = new byte[1024];
	(new Random()).nextBytes(data);
	QSByteArray qsdata = new QSByteArray (data);

	SigShare ss = ThresholdSignedQS.create_sig_share (
			ks, qsdata, qs_sig_type);
    }
}



