/*
 * TapestryVerifier.java
 *
 *  Copyright (c) 2002 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.tapestry.impl;

import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedList;
import ostore.util.Carp;
import ostore.util.NodeId;
import ostore.util.SecureHash;
import ostore.util.SHA1Hash;

/**
 *
 * An object of this class can be used to verify certain properties
 * about a Tapestry network, given complete knowledge about the network.
 * For example, given everyone's Guids and RoutingTables, it can 
 * verify the correctness of each RoutingTable.<br><br>
 *
 * Once a TapestryVerifier has first been populated with data (either through
 * the constructor or a call to addData), that will be the only kind
 * of data the object will accept.  For example, if you construct a 
 * TapestryVerifier with the (guids[], tables[]) constructor, you cannot
 * later call the addData( guid, table, revpointers) method.
 *
 * @author     Jeremy Stribling
 * @version    $Id: TapestryVerifier.java,v 1.6 2003/03/17 23:58:30 ravenben Exp $
 */

public class TapestryVerifier {

    // Possible errors:
    
    // NodeId to guid mapping doesn't exist in routing table
    public static final String ERROR_NO_GUID = "NO_GUID";
    // No data for this node has been provided.
    public static final String ERROR_NO_DATA = "NO_DATA";
    // guid found where it should not be in someone's routing table
    public static final String ERROR_BAD_GUID_DIGITS = "BAD_GUID_DIGITS";
    // no latency found for this node in the table
    public static final String ERROR_NO_LAT = "NO_LAT";
    // nodes out of order with respect to latencies in routing table
    public static final String ERROR_BAD_LAT_ORDER = "BAD_LAT_ORDER";
    // guid not found in revpointers, when expected 
    public static final String ERROR_GUID_NOT_IN_REVS = "GUID_NOT_IN_REVS";
    // hole found in routing table
    public static final String ERROR_ROUTING_HOLE = "ROUTING_HOLE";
    // revpointer found that wasn't actually in other's table
    public static final String ERROR_BAD_REVPOINTER = "BAD_REVPOINTER";
    // duplicate guid in routing table entry
    public static final String ERROR_DUPLICATE_GUID = "DUPLICATE_GUID";

    /**
     * Possible types
     */
    private static final int NO_TYPE = -1;
    private static final int TYPE_GT = 1;
    private static final int TYPE_GTR = 2;

    /**
     * SecureHash -> TapestryVerifier.DataNode
     */
    private HashMap data;
    
    /**
     * What kind of data does this Verifier have?
     */
    private int type = NO_TYPE;

    /**
     * Keeps track of guid relationships
     */
    private Tree guidTree;

    /**
     * Info created and cached by a call to verify. When more
     * data is added, this info is invalidated.
     */
    private String errors;
    private boolean verified;

    /**
     * Create an empty TapestryVerifier, and fill it up with data one at
     * a time later.
     */
    public TapestryVerifier() {

	reset();

    }

    /**
     * Create a TapestryVerifier that can verify that every 
     * RoutingTable is correct as well.  All holes are filled that
     * should be filled, all nodes are sorted correctly according to the
     * latencies recorded, no nodes are in incorrect spots, etc.
     *
     * @param guids  An array of all known guids in the network.
     * @param tables  An array of all RoutingTables for all nodes, in the
     *                same order as the corresponding guids.
     */
    public TapestryVerifier( SecureHash guids[], RoutingTable tables[] ) {

	this();
	// all arrays need to be the same length
	if( guids.length != tables.length ) {
	    Carp.die( "Array lengths don't match in TapestryVerifier" );
	}
	for( int i = 0; i < guids.length; i++ ) {
	    addData( guids[i], tables[i] );
	}

    }

    /**
     * Create a TapestryVerifier that can verify that every 
     * RoutingTable is correct as well, and also verify that every
     * revpointers map is correct.  All nodes in each revpointers
     * are actually being pointed to in the corresponding RoutingTable,
     * all nodes in RoutingTable are in corresponding revpointers, etc.
     *
     * @param guids  An array of all known guids in the network.
     * @param tables  An array of all RoutingTables for all nodes, in the
     *                same order as the corresponding guids.
     * @param revpointers    An array of the revpointer HashMaps for all
     *                       nodes, in the same order as the 
     *                       corresponding guids.
     */
    public TapestryVerifier( SecureHash guids[], RoutingTable tables[],
			     HashMap revpointers[] ) {


	this();
	// all arrays need to be the same length
	if( (guids.length != tables.length) || 
	    (guids.length != revpointers.length)  ) {
	    Carp.die( "Array lengths don't match in TapestryVerifier" );
	}
	for( int i = 0; i < guids.length; i++ ) {
	    addData( guids[i], tables[i] );
	}

    }


    /**
     * Consider more data.
     *
     * @param guid    a guid of a node in the network
     * @param table   the RoutingTable of the node
     */
    public synchronized void addData( SecureHash guid, RoutingTable table ) {

	if( type == NO_TYPE ) {
	    type = TYPE_GT;
	} else if( type != TYPE_GT ) {
	    Carp.die( "Trying to add data not of TYPE_GT " + 
		      "in TapestryVerifier!" );
	}

	errors = null;
	data.put( guid, new DataNode( guid, table, null ) );
	guidTree.add( guid );

    }

    /**
     * Consider more data.
     *
     * @param guid    a guid of a node in the network
     * @param table   the RoutingTable of the node
     * @param revpointers     the reverse pointers of the node
     */
    public synchronized void addData( SecureHash guid, RoutingTable table, 
				      HashMap revpointers ) {

	if( type == NO_TYPE ) {
	    type = TYPE_GTR;
	} else if( type != TYPE_GTR ) {
	    Carp.die( "Trying to add data not of TYPE_GTR " +
		      "in TapestryVerifier!" );
	}

	errors = null;
	data.put( guid, new DataNode( guid, table, revpointers ) );
	guidTree.add( guid );

    }

    /**
     * Prints the errors from the last verify() call to STDOUT. 
     */
    public void printErrors() {


	Carp.out( "***********************************" );
	Carp.out( "TapestryVerifier errors:\n" );

	if( errors == null ) {
	    Carp.out( "No errors, or verify() not called" );
	} else {
	    Carp.out( errors );
	}
	
	Carp.out( "***********************************" );
	
	
    }

    /**
     * Return the String of all the errors from the last verify.  This is 
     * useful if the caller wants to parse it and take corrective measures.
     * Returns null if no errors or if verify() hasn't been called.
     */
    public String getErrors() {
	return errors;
    }

    /**
     * This should be called when getting ready to populate this Verifier
     * with a new set of data, or if the guids of any nodes have changed.
     */
    public synchronized void reset() {
	errors = null;
	guidTree = new Tree();
	data = new HashMap();
    }

    /**
     * Verify all the info we have about this Tapestry network.  Returns
     * true if everything was correct.  Otherwise, it will return false
     * and the errors can be printed to STDOUT with a call to
     * printErrors.
     *
     * verify will cache its answer after the first call, and that answer
     * will remain valid until more data is added.
     *
     * @return whether all data is consistent for this Tapestry network
     */
    public synchronized boolean verify() {

	// is the cached answer valid?
	if( errors != null ) {
	    return verified;
	}

	// assume it's ok, and be proved wrong later
	verified = true;

	// Algorithm for verification:
	// Let's do this a dumb way first.  The dumb way:
	// For each node in data
	//   For each entry in routing table
	//    if entry is non-null and full
	//      make sure each node in the entry belongs
	//      make sure they're sorted right by latency
	//      make sure the current node is in their revpointers
	//    else
	//      make sure there is no node that could be here
	//   For each node in revpointer
	//      make sure the current node is in the routing table of this node
	//
	// TODO: There's got to be a smarter way

	errors = "";
	Iterator nodes = data.values().iterator();
	while( nodes.hasNext() ) {
	    
	    DataNode currNode = (DataNode) nodes.next();
	    
	    for(int digit = 0; digit<RoutingTable.DIGITS_PER_GUID; digit++) {
		for(int value = 0; value<RoutingTable.DIGIT_VALUES; value++) {
		
		    RouteEntry re = currNode.table.getEntry( digit, value ); 
		    if( re != null ) {

			double lastLatency = Double.MIN_VALUE;

			for( int i = 0; i < re.size(); i++ ) {

			    NodeId reNodeId = re.getAtIndex(i);
			    
			    SecureHash reGuid = (SecureHash)
				currNode.table.get_node_map().get( reNodeId );
			    if( reGuid == null ) {

				// ERROR!
				// Don't know guid
				verified = false;
				errorNoGuid( currNode.guid, reNodeId );
				continue;

			    }

			    DataNode reNode = (DataNode) data.get( reGuid );
			    if( reNode == null ) {

				// ERROR!
				// Don't have data
				verified = false;
				errorNoData( currNode.guid, reNodeId, reGuid );
				continue;

			    }

			    // make sure this node belongs here
			    // first, make sure they match on everything
			    // up to this digit
			    boolean gotError = false;
			    for( int j = 0; j < digit; j++ ) {
				if( reNode.digits[j] != currNode.digits[j] ) {
				  
				    // ERROR!
				    // Guid doesn't belong
				    verified = false;
				    errorBadGuidDigits( currNode.guid, 
							reNodeId, reGuid,
							digit, value );
				    gotError = true;
				    break;

				}

			    }
			    if( gotError ) {
				continue;
			    }
			    // next, make sure the current value matches where
			    // it should be
			    if( reNode.digits[digit] != value ) {
				
				// ERROR!
				// Guid doesn't belong
				verified = false;
				errorBadGuidDigits( currNode.guid, 
						    reNodeId, reGuid,
						    digit, value );
				continue;
				
			    }

			    // does this node appear multiple times?
			    for( int j = 0; j < i; j++ ) {
				if (reNodeId.equals( re.getAtIndex(j) ) ) {
				    
				    // ERROR!
				    // duplicate nodeid
				    verified = false;
				    errorDuplicateGuid( currNode.guid,
							reNodeId, reGuid,
							digit, value );
				    continue;

				}
			    }

			    // don't need to check latency numbers 
			    // or revpointers for ourself
			    if( reGuid.equals( currNode.guid ) ) {
				continue;
			    }

			    // now, make sure the latencies are in order
			    Double currLatencyLong = 
				currNode.table.node_to_latency( reNodeId );
			    if( currLatencyLong == null ) {
				
				// ERROR!
				// can't find latency
				verified = false;
				errorNoLat( currNode.guid, reNodeId, reGuid,
					    digit, value );
				continue;

			    }
			    double currLatency = currLatencyLong.doubleValue();
			    if( lastLatency > currLatency ) {

				// ERROR!
				// previous latency was bigger!
				verified = false;
				errorBadLatOrder( currNode.guid, reNodeId, 
						  reGuid, digit, value,
						  lastLatency, currLatency );
				continue;

			    }

			    // finally, if we have revpointers, make sure
			    // current node exists in correct revpointers
			    if( type == TYPE_GTR ) {
				HashMap hm = (HashMap) 
				    reNode.revpointers.get(new Integer(digit));

				if( hm == null ) {

				    // ERROR
				    // no map for this level
				    verified = false;
				    errorGuidNotInRevs( reGuid, currNode.guid,
							digit );
				    continue;

				}

				if( !hm.containsKey( currNode.guid ) ) {

				    // ERROR
				    // not in this level's map
				    verified = false;
				    errorGuidNotInRevs( reGuid, currNode.guid,
							digit );
				    continue;

				}

			    }


			}

		    } // if( re != null )

		    // if RouteEntry is null or not full
		    //if(re == null || re.size() < RouteEntry.REDUNDANT_LVL ) {
		    // if RouteEntry is null or empty
		    if( re == null || re.size() == 0 ) {

			int size = 0;
			if( re != null ) {
			    size = re.size();
			}

			// figure out if there should any nodes in here
			// that are conspicuously missing
			LinkedList matchers = guidTree.matches( currNode.guid,
								 digit, 
								 value );
			if( matchers.size() > size ) {

			    // ERROR!
			    // Hole in routing table
			    verified = false;

			    // figure out ones that are missing
			    SecureHash matches[] = 
				new SecureHash[matchers.size() - size];
			    int matchesIndex = 0;
			    for( int i = 0; i < matchers.size(); i++ ) {
				SecureHash currGuid = 
				    (SecureHash) matchers.get(i);
				if( size == 0 ) {
				    matches[matchesIndex] = currGuid;
				    matchesIndex++;
				} else {
				    // make sure this one isn't there already
				    for( int j = 0; j < re.size(); j++ ) {
					NodeId nid = re.getAtIndex(j);
					SecureHash reGuid = (SecureHash)
					    currNode.table.get_node_map().
					                     get( nid );
					if( !currGuid.equals( reGuid ) &&
					    matchesIndex < matches.length ) {
					    matches[matchesIndex] = currGuid;
					    matchesIndex++;
					}
				    }
				}
			    }

			    // make error
			    errorRoutingHole( currNode.guid, digit, value,
					      matches );

			} // end routing hole error

		    }

		}

	    } // end routing table verification

	    if( type == TYPE_GTR ) {
		// now let's check the revpointers, and see if there are any
		// extra nodes in there
		Iterator revlevels = currNode.revpointers.keySet().iterator();
		while( revlevels.hasNext() ) {
		    Integer level = (Integer) revlevels.next();
		    int intLevel = level.intValue();
		    HashMap guids = (HashMap) currNode.revpointers.get(level);

		    Iterator guidIter = guids.keySet().iterator();
		    while( guidIter.hasNext() ) {

			SecureHash revGuid = (SecureHash) guidIter.next();
			
			// see if we are really in this node's routing table
			// at this level
			DataNode revNode = (DataNode) data.get( revGuid );
			if( revNode == null ) {
			    // ERROR!
			    // Don't have data
			    verified = false;
			    errorNoData( currNode.guid, null, revGuid );
			    continue;
			}
			RouteEntry re = revNode.table.getEntry( intLevel,
						   currNode.digits[intLevel] );

			if( re == null ) {
			    
			    // ERROR!
			    // null RouteEntry where non-null is expected
			    verified = false;
			    errorBadRevpointer( currNode.guid, revGuid,
						intLevel );
			} else {

			    // see if it's in there
			    boolean found = false;
			    for( int i = 0; i < re.size(); i++ ) {
				NodeId nid = re.getAtIndex(i);
				SecureHash reGuid = (SecureHash)
				    revNode.table.get_node_map().get( nid );
				if( reGuid == null ) {
				    
				    // ERROR!
				    // Don't know guid
				    verified = false;
				    errorNoGuid( revNode.guid, nid );
				    continue;
				    
				}
				if( reGuid.equals( currNode.guid ) ) {
				    found = true;
				}
			    }
			    if( !found ) {
				
				// ERROR!
				// couldn't find it in the entry
				verified = false;
				errorBadRevpointer( currNode.guid, revGuid,
						    intLevel );

			    }

			}
			
		    }

		}


	    } // end revpointer verification

	} // end looping through data

	return verified;

    }

    private String label( String error ) {
	return error + ": ";
    }

    private void errorNoGuid( SecureHash guid, NodeId badNode ) {

	errors += label(ERROR_NO_GUID) + "Node " + badNode + " has no guid " +
	    "mapping in the table of " + guid + "\n";

    }

    private void errorNoData( SecureHash guid, NodeId badNode, 
			      SecureHash badGuid ) {

	errors += label(ERROR_NO_DATA) + "Node " + badNode + " with guid " +
	    badGuid + " has no data reported, but was found in table of " + 
	    guid + "\n";

    }
    
    private void errorBadGuidDigits( SecureHash guid, NodeId badNode, 
				     SecureHash badGuid, int digit, 
				     int value ) {

	errors += label(ERROR_BAD_GUID_DIGITS) + "Node " + badNode + 
	    " with guid " + badGuid + " does not belong in table of " + 
	    guid + " at index (" + digit + "," + value + ")\n";
	

    }

    private void errorNoLat( SecureHash guid, NodeId badNode, 
			     SecureHash badGuid, int digit, int value ) {

	errors += label(ERROR_NO_LAT) + "Node " + badNode + 
	    " with guid " + badGuid + " does not have latency in table of " + 
	    guid + " at index (" + digit + "," + value + ")\n";
	

    }

    private void errorBadLatOrder( SecureHash guid, NodeId badNode, 
				   SecureHash badGuid, int digit, int value,
				   double lastLat, double currLat ) {

	errors += label(ERROR_BAD_LAT_ORDER) + "Node " + badNode + 
	    " with guid " + badGuid + " has a smaller latency (" + currLat + 
	    ") than a preceding node (" + lastLat + ") in table of " + 
	    guid + " at index (" + digit + "," + value + ")\n";
	

    }

    private void errorGuidNotInRevs( SecureHash guid, SecureHash badGuid, 
				     int level ) {

	errors += label(ERROR_GUID_NOT_IN_REVS) + "Guid " + badGuid +
	    " not in revpointers of " + guid + " at level " + level +
	    " although it should be\n";

    }

    private void errorRoutingHole( SecureHash guid, int digit, int value,
				   SecureHash[] matches ) {

	errors += label(ERROR_ROUTING_HOLE) + "Guids found that could fill " +
	    "hole in table of " + guid + " at index (" + digit + "," +
	    value + ") -- { ";
	for( int i = 0; i < matches.length; i++ ) {
	    errors += matches[i] + " "; 
	}
	errors += "}\n";

    }

    private void errorBadRevpointer( SecureHash guid, SecureHash badGuid,
				     int level ) {

	errors += label(ERROR_BAD_REVPOINTER) + "Guid " + badGuid + 
	    " found in revpointers of " + guid + ", but " + guid + 
	    " was not in level " + level + " of its table\n";

    }

    private void errorDuplicateGuid( SecureHash guid, NodeId badNode, 
				     SecureHash badGuid, int digit, 
				     int value ) {

	errors += label(ERROR_DUPLICATE_GUID) + "Node " + badNode + 
	    " with guid " + badGuid + " appears twice in table of " + 
	    guid + " at index (" + digit + "," + value + ")\n";
	

    }

    // lame test function
    public static void main( String[] args ) {

	// make a few fake nodes, and give them fake routing tables
	SecureHash guid1 = new SHA1Hash( "guid1" ); // 2c336
	SecureHash guid2 = new SHA1Hash( "guid2" ); // 45934
	SecureHash guid3 = new SHA1Hash( "guid3" ); // 3637f
	SecureHash guid4 = new SHA1Hash( "guid4" ); // c468c
	SecureHash guid5 = new SHA1Hash( "guid5" ); // 81954

	NodeId nodeid1 = null;
	NodeId nodeid2 = null;
	NodeId nodeid3 = null;
	NodeId nodeid4 = null;
	NodeId nodeid5 = null;

	try {
	    nodeid1 = new NodeId( "192.168.0.1:3631" );
	    nodeid2 = new NodeId( "192.168.0.1:3632" );
	    nodeid3 = new NodeId( "192.168.0.1:3633" );
	    nodeid4 = new NodeId( "192.168.0.1:3634" );
	    nodeid5 = new NodeId( "192.168.0.1:3635" );
	} catch( Exception e ) {
	    Carp.die( "Got NodeId exception: " + e );
	}

	//Carp.out( "guid1: " + guid1 );
	//Carp.out( "guid2: " + guid2 );
	//Carp.out( "guid3: " + guid3 );
	//Carp.out( "guid4: " + guid4 );

	RoutingTable rt1 = new RoutingTable( nodeid1, guid1, null );
	RoutingTable rt2 = new RoutingTable( nodeid2, guid2, null );
	RoutingTable rt3 = new RoutingTable( nodeid3, guid3, null );
	RoutingTable rt4 = new RoutingTable( nodeid4, guid4, null );

	rt1.write( 0, 4, nodeid2, guid2, new Double(1) );
	rt1.write( 0, 5, nodeid3, guid3, new Double(1) );
	rt1.write( 1, 2, nodeid4, guid4, new Double(1) );
	rt1.addSelf( guid1 );

	rt2.write( 0, 4, nodeid1, guid1, new Double(1) );
	rt2.write( 0, 5, nodeid3, guid3, new Double(1) );
	rt2.write( 1, 6, nodeid4, guid4, new Double(1) );
	rt2.addSelf( guid2 );

	rt3.write( 0, 2, nodeid1, guid1, new Double(1) );
	rt3.write( 0, 3, nodeid2, guid2, new Double(1) );
	rt3.write( 0, 10, nodeid4, guid4, new Double(1) );
	rt3.addSelf( guid3 );

	rt4.write( 0, 2, nodeid1, guid1, new Double(1) );
	rt4.write( 0, 4, nodeid2, guid2, new Double(1) );
	rt4.write( 0, 3, nodeid3, guid3, new Double(1) );
	rt4.addSelf( guid4 );

	HashMap hm1 = new HashMap();
	HashMap hm2 = new HashMap();
	HashMap hm3 = new HashMap();
	HashMap hm4 = new HashMap();

	HashMap level0 = new HashMap();
	level0.put( guid4, nodeid4 );
	level0.put( guid2, nodeid2 );
	hm1.put( new Integer(0), level0 );
	
	TapestryVerifier tv = new TapestryVerifier();
	tv.addData( guid1, rt1, hm1 );
	tv.addData( guid2, rt2, hm2 );
	tv.addData( guid3, rt3, hm3 );
	tv.addData( guid4, rt4, hm4 );

	if( !tv.verify() ) {
	    tv.printErrors();
	}

    }

    /**
     * A private inner class that just holds the table and revpointers for
     * a guid
     */
    private class DataNode {

	SecureHash guid;
	RoutingTable table;
	HashMap revpointers;
	int digits[];

	public DataNode( SecureHash guid, RoutingTable table, 
		     HashMap revpointers ) {

	    this.guid = guid;
	    this.table = table;
	    this.revpointers = revpointers;
	    digits = RoutingTable.digits( guid );

	}

    }

    /**
     * Organizes guids into a tree-like data structure that makes it
     * easy to tell who should know about each other.
     */
    public class Tree {

	private TreeNode root = new TreeNode();;

	/** 
	 * Add this new node into the graph
	 */
	public void add( SecureHash guid ) {

	    // traverse the graph, adding it into the right place every time
	    TreeNode lastNode = root;
	    int[] digits = RoutingTable.digits(guid);

	    for( int digit = 0; digit<RoutingTable.DIGITS_PER_GUID; digit++ ) {

		TreeNode currNode = lastNode.links[digits[digit]];
		if( currNode == null ) {
		    currNode = new TreeNode();
		    lastNode.links[digits[digit]] = currNode;
		}

		// leave it here, and move on
		if( !currNode.matchers.contains( guid ) ) {
		    currNode.matchers.add( guid );
		}
		lastNode = currNode;

	    }

	}

	/**
	 * Return all nodes that could be in this guid's routing table
	 * at that particular digit, for that value, including itself
	 */
	public LinkedList matches( SecureHash guid, int digit, int value ) {

	    TreeNode lastNode = root;
	    int[] digits = RoutingTable.digits(guid);

	    // traverse the tree up to this digit
	    for( int level = 0; level<digit; level++ ) {
		lastNode = lastNode.links[digits[level]];
		if( lastNode == null ) {
		    Carp.die( "Followed bad link in TapestryVerifier.Tree" );
		}
	    }
	    
	    // we've arrived at the interesting node.  Check down the
	    // link and see what we have (if anything).
	    TreeNode finalNode = lastNode.links[value];
	    if( finalNode == null ) {
		return new LinkedList();
	    } else {
		return finalNode.matchers;
	    }

	}


    }

    /**
     * Private inner class for keeping track of graph stuff
     */
    private class TreeNode {

	LinkedList matchers = new LinkedList();
	TreeNode[] links = new TreeNode[RoutingTable.DIGIT_VALUES];

    }


}
