/**
 * FaultController.java
 *
 * @author	Jeremy Stribling
 * @version	$Id: FaultController.java,v 1.3 2003/01/29 22:40:50 strib Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.tapestry.patchwork;

import java.util.HashMap;
import java.util.Random;
import ostore.util.NodeId;
import ostore.util.Debug;
import ostore.util.DebugFlags;

/**
 * This is a data structure that decides whether or not a message
 * should be dropped on its way through the Router.
 *
 * @author Jeremy Stribling
 * @version    $Id: FaultController.java,v 1.3 2003/01/29 22:40:50 strib Exp $
 */
public class FaultController {

    private boolean DEBUG = DebugFlags.TAPESTRY_PATCHWORK_FAULTCONTROLLER;
    private static final String class_tag = "FaultController";

    private HashMap faultRates;
    private Random rng;

    private RouterFaultMsg allNodesRFM;
    
    public FaultController() {

	final String tag = class_tag+".constructor";

	faultRates = new HashMap();
	Random preRng = new Random();
	long rng_starter = preRng.nextLong();
	if( DEBUG ) Debug.printtagln( tag, "Constructing FaultController "
				      + "with the random seed: " + 
				      rng_starter );
	rng = new Random( rng_starter );
	allNodesRFM = null;
    }

    /**
     * Called when a new <code>RouterFaultMsg</code> is received 
     * by the <code>Router</code>
     */
    public void updateFaults( RouterFaultMsg rfm ) {

	final String tag = class_tag+".updateFaults";
	if( DEBUG ) Debug.printtagln( tag, "received fault: " + rfm );
	if( rfm.allNodes ) {
	    allNodesRFM = rfm;
	} else {
	    if( rfm.nodeToDrop == null ) {
		faultRates.put( rfm.peer, rfm );
	    } else {
		faultRates.put( rfm.nodeToDrop, rfm );
	    }
	}
    }

    /**
     * Decide whether or not to drop a message, based on previously
     * received loss rates.
     */
    public boolean dropMessage( NodeId peer ) {

	final String tag = class_tag+".dropMessage";
	// first, check the allNodes rfm
	if( allNodesRFM != null ) {
	    if( allNodesRFM.isPercent && 
		(Math.random()*100 < allNodesRFM.value) ) {
		// drop is as a percent
		if( DEBUG ) Debug.printtagln( tag, 
					      "rejecting msg/percent =" 
					      + allNodesRFM.value );
		return true;
	    } else if( !allNodesRFM.isPercent && 
		       (allNodesRFM.value > 0) ) {
		// drop it as an absolute value
		allNodesRFM.value--;
		if( DEBUG ) Debug.printtagln( tag, 
					      "rejecting msg/# left =" 
					      + allNodesRFM.value );
		return true;
	    }
	}
	// then, try the individual faults
	RouterFaultMsg rfm = (RouterFaultMsg) faultRates.get( peer );
	if( rfm != null ) {
	    if( rfm.isPercent && 
		(Math.random()*100 < rfm.value) ) {
		// drop is as a percent
		if( DEBUG ) Debug.printtagln( tag, 
					      "rejecting msg/percent =" 
					      + rfm.value );
		return true;
	    } else if( !rfm.isPercent && 
		       (rfm.value > 0) ) {
		// drop it as an absolute value
		rfm.value--;
		if( DEBUG ) Debug.printtagln( tag, 
					      "rejecting msg/# left =" 
					      + rfm.value );
		return true;
	    }
	}
	return false;
    }

}

