/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

import java.util.Iterator;
import java.util.HashSet;
import ostore.dispatch.Classifier;
import ostore.tapestry.api.TapestryNodeDetachMsg;
import seda.sandStorm.api.SinkException;
import sun.misc.Signal;
import sun.misc.SignalHandler;

/**
 * A convenience class for dying and generating error messages, complete
 * with "carp" and "croak", borrowed from PERL.
 * <p>
 * Now extended to control a graceful shutdown.  Calls to
 * <code>dieGracefully</code> cause this class to detach a node from
 * Tapestry.  UNIX signals (TERM, INT, HUP) detach all virtual nodes.
 *
 * @author	Dennis Geels
 * @version	$Id: Carp.java,v 1.16 2003/01/04 00:15:18 geels Exp $
 */
public abstract class Carp
{
    ////////////////////       CONSTANTS	////////////////////
    ////////////////////    STATIC VARIABLES 	////////////////////

    /**
     * <code>true</code> iff the <code>croak</code> and <code>die</code>
     * methods should throw a <code>RuntimeException</code> instead of
     * calling <code>System.exit</code>.
     */
    private static boolean _throwException = false;

    /**
     * The set of <code>NodeId</code>s for nodes that are already dying
     * gracefully.
     */
    private static HashSet _dyingNodes = new HashSet();

    ////////////////////       VARIABLES	////////////////////
    ////////////////////  STATIC CONSTRUCTORS	////////////////////

    static {
	// Register UNIX signal handlers.
	// TODO: Can we generalize this functionality beyond the
	// not-quite-official sun.misc package?
	
	String[] signal_names = {"INT", "HUP", "TERM"};
	SignalHandler handler = new ShutdownHandler();
	for( int i=0; i<signal_names.length; i++ ) {
	    try {
		Signal.handle( new Signal( signal_names[i] ), handler );
	    } catch( IllegalArgumentException iae ) {
		out( "Cannot catch " + signal_names[i] + " signal" );
	    }
	}
    }
    
    ////////////////////      CONSTRUCTORS	////////////////////
    ////////////////////     STATIC METHODS 	////////////////////

    /**
     * Sets the flag that controls whether <code>croak</code> and
     * <code>die</code> should throw a <code>RuntimeException</code>
     * (<code>true</code>) or call <code>System.exit</code>
     * (<code>false</code>). 
     */
    public static void setThrowException( boolean throwException ) {
	_throwException = throwException;
    }

    /**
     * Returns the flag that controls whether <code>croak</code> and
     * <code>die</code> should throw a <code>RuntimeException</code>
     * (<code>true</code>) or call <code>System.exit</code>
     * (<code>false</code>). 
     */
    public static boolean getThrowException() {
	return _throwException;
    }

    /**
     * Dispatches an event to the Tapestry stages of the specified virtual
     * node telling them to disconnect the node from the Tapestry network.
     * <p>
     * If the <code>_throwException</code> flag is <code>false</code>, this
     * method just calls <code>croak</code>. 
     */
    public static void dieGracefully( Throwable e, NodeId node_id ) {
	String tag = "Carp.dieGracefully";
	if(!_throwException) {
	    // Graceful handling is disabled.  This call will exit:
	    error( e );
	    carp( tag + ": exiting" );
	    exit();
	}

	synchronized(_dyingNodes) {
	    if( _dyingNodes.contains( node_id ) ) {
		// Already dying. Don't confuse Tapestry.
		return;
	    } // else continue
	    _dyingNodes.add( node_id );
	}

	try {
	    error( e );
	    out( tag + ": " + node_id + " detaching from tapestry.");
	    Classifier c = Classifier.getClassifier( node_id );
	    c.dispatch( new TapestryNodeDetachMsg() );
	} catch( SinkException se ) {
	    carp( se );
	    exit();
	}
    }

    
    /**
     * Print a message to standard out.
     */
    public static void out( String msg ) {
	System.out.println(  msg );
    }

    /**
     * Print a message to standard error.
     */
    public static void carp( String msg ) {
	System.err.println( "ERROR:\t" + msg );
    }

    /**
     * Print an exception's message to standard error.
     */
    public static void carp( Throwable e ) {
	carp( "Caught " + e.getClass().getName() );
	carp( "  message: " + e.getMessage() );
    }

    /**
     * Print a message and current stack trace to standard error.
     */
    public static void error( String msg ) {
	carp( msg );
        Thread.dumpStack();
    }

    /**
     * Print an exception's message and stack trace to standard error.
     */
    public static void error( Throwable e ) {
	carp( e );
	e.printStackTrace();
    }

    /**
     * Kill the whole jvm.
     */
    private static void exit() {
	System.exit( -1 );
    }

    /**
     * Kill the whole jvm or throw a <code>RuntimeException</code>.
     */
    public static void croak() {
	if( _throwException ) {
	    throw new RuntimeException();
	} else {
	    exit();
	}
    }

    /**
     * Print a message to standard error and croak.
     */
    public static void croak( String msg ) {
	carp( msg );
	croak();
    }

    /**
     * Print an exception's message to standard error and croak.
     */
    public static void croak( Throwable e ) {
	carp( e );
	croak();
    }
    
    /**
     * Carp the current stack trace, then croak.
     */
    public static void die( ) {
	Thread.dumpStack();
	croak();
    }

    /**
     * Carp a message and current stack trace, then croak.
     */
    public static void die( String msg ) {
	carp( msg );
	die();
    }

    /**
     * Carp an exception's message and stack trace, then croak.
     */
    public static void die( Throwable e ) {
	error( e );
	die();
    }

    /**
     * Print out the stack trace of the current thread.  The stack trace will
     * include the call to <code>Carp.dumpStack</code>.
     */
    public static void dumpStack() {
	Thread.dumpStack();
    }
    
    //////////////////// 	METHODS		////////////////////
    ////////////////////     INNER CLASSES	////////////////////

    /**
     * A simple {@link sun.misc.SignalHandler} that calls
     * {@link #dieGracefully} for each virtual node in this jvm.
     */
    private static class ShutdownHandler implements SignalHandler
    {
	/**
	 * Call <code>dieGracefully</code> for each virtual node.
	 */
	public void handle( Signal sig ) {
	    carp( "ShutdownHandler caught signal: " + sig +
		  ".  All nodes dying." );
	    RuntimeException e = new RuntimeException();

	    Iterator node_iter = Classifier.getNodeIds().iterator();
	    while( node_iter.hasNext() ) {
		dieGracefully( e, (NodeId)node_iter.next() );
	    }
	}
    }
}
