/**
 * OutputBufferImpl.java
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.util;
import java.math.BigInteger;

/**
 * A buffer into which <code>QuickSerializable</code> objects can be
 * written.  
 *
 * Implements all of the non-essential functions of
 * <code>OutputBuffer</code>, so that other classes only need to implement
 * <code>add (byte)</code> and <code>add (byte, int, int)</code>.
 *
 * <p>The implementations of most of these functions were cribbed from
 * Dennis's SecureHashBuffer class.
 * 
 * @see QuickSerializable
 *
 * @author  Sean C. Rhea
 * @version $Id: OutputBufferImpl.java,v 1.12 2002/09/05 08:26:00 hweather Exp $
 */

public abstract class OutputBufferImpl implements OutputBuffer
{
    /**
     * Adds the specified byte to the digest.
     */
    public abstract void add (byte input);

    /**
     * Adds <code>length</code> bytes of the specified array to the digest,
     * starting at <code>offset</code>.
     */
    public abstract void add (byte[] input, int offset, int length);

    /**
     * Adds the specified byte array to the digest.
     */
    public void add (byte[] input) {
	add (this, input);
    }

    /**
     * Adds the specified object to the digest.
     * This method invokes the <code>QuickSerializable.serialize</code>
     * method; be careful to avoid infinite loops.
     */
    public void add (QuickSerializable value) {
	add (this, value);
    }

    /**
     * Adds the remaining bytes in the given input buffer to this output
     * buffer.
     */
    public void add (InputBuffer in) {
	add (this, in);
    }

    public void add (boolean b) {
        add (this, b);
    }

    public void add (short i) {
	add (this, i);
    }

    public void add (int i) {
	add (this, i);
    }

    public void add (long i) {
	add (this, i);
    }

    public void add (double x) {
	add (this, x);
    }

    public void add (String s) {
        add (this, s);
    }

    public void add (BigInteger value) {
	add (this, value);
    }

    public static final void add (OutputBuffer buffer, byte[] input) {
	buffer.add (input, 0, input.length);
    }

    /**
     * Adds the specified object to the digest.
     * This method invokes the <code>QuickSerializable.serialize</code>
     * method; be careful to avoid infinite loops.
     *
     * @throws AssertionViolatedException
     *         if <code>value</code> is <code>null</code>
     */
    public static final void add (OutputBuffer buffer,
                                  QuickSerializable value)
        throws ostore.util.AssertionViolatedException
    {
        if (value == null)
        {
            String msg = "Attempt to serialize null object.";
            throw new ostore.util.AssertionViolatedException(msg);
        }

	buffer.add (TypeTable.type_code_qs (value));
	value.serialize (buffer);
    }

    /**
     * Adds the remaining bytes in the given input buffer to this output
     * buffer.
     */
    public static final void add (OutputBuffer out, InputBuffer in) {
	in.nextBytes (out);
    }

    public static final void add (OutputBuffer buffer, boolean b) {
        buffer.add ((byte) (b ? 1 : 0));
    }

    public static final void add (OutputBuffer buffer, short i) {
	/** 
	 * Commented out must be MSBF, so that we use ByteUtils and
	 * let ByteUtils specify whether we use <I>network</I>-order
	 * or <I>intel</I>-order.
	 *
	 * // must be MSBF
	 *
	 * for (int j = 1; j >= 0; --j) {
	 *    buffer.add ((byte) (0xff & (i>>(8*j))));
	 * }
	 **/

	byte[] data  = {0,0};
	int[] offset = {0};
	ByteUtils.shortToBytes(i, data, offset);
	offset[0] = 0;
	buffer.add(data, offset[0], ByteUtils.SIZE_SHORT);
    }

    public static final void add (OutputBuffer buffer, int i) {
	/** 
	 * Commented out must be MSBF, so that we use ByteUtils and
	 * let ByteUtils specify whether we use <I>network</I>-order
	 * or <I>intel</I>-order. 
	 *
	 * // must be MSBF
	 *
	 * for (int j = 3; j >= 0; --j) {
	 *    buffer.add ((byte) (0xff & (i>>(8*j))));
	 * }
	 **/
	byte[] data  = {0,0,0,0};
	int[] offset = {0};
	ByteUtils.intToBytes(i, data, offset);
	offset[0] = 0;
	buffer.add(data, offset[0], ByteUtils.SIZE_INT);
    }

    public static final void add (OutputBuffer buffer, long i) {
	/** 
	 * Commented out must be MSBF, so that we use ByteUtils and
	 * let ByteUtils specify whether we use <I>network</I>-order
	 * or <I>intel</I>-order. 
	 *
	 * 
	 * // must be MSBF
	 *
	 * for (int j = 7; j >= 0; --j) {
	 *    buffer.add ((byte) (0xff & (i>>(8*j))));
	 * }
	 **/
	byte[] data = {0,0,0,0,0,0,0,0};
	int[] offset = {0};
	ByteUtils.longToBytes(i, data, offset);
	offset[0] = 0;
	buffer.add(data, offset[0], ByteUtils.SIZE_LONG);
    }

    public static final void add (OutputBuffer buffer, double x) {
	long bits = Double.doubleToLongBits (x);
	buffer.add (bits);
    }

    public static final void add (OutputBuffer buffer, String s) {
        byte s_bytes[] = s.getBytes ();
        buffer.add (s_bytes.length);
        if (s_bytes.length > 0)
        {
            buffer.add (s_bytes);
        }
        return;
    }

    public static final void add (OutputBuffer buffer, BigInteger value) {
	// TODO: I wish I could do this without allocating memory, but the
	// BigInteger interface doesn't allow for it.

	byte [] bytes = value.toByteArray ();
	buffer.add (bytes.length);
	buffer.add (bytes);
    }
}

