/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

/**
 * A <code>QSVector</code> is a cousin of <code>java.lang.Vector</code> which
 * implements <code>ostore.util.QuickSerializable</code>.  All elements of the
 * <code>QSVector</code> must also be <code>QuickSerializable</code>.
 *
 * Most of the functionality from the <code>Vector</code> has not been
 * ported.  If you require the use of a missing method, implement it.
 *
 *  @author  Dennis Geels
 *  @version $Id: QSVector.java,v 1.27 2002/12/10 09:51:18 hweather Exp $
 */
public class QSVector implements QuickSerializable, Cloneable
{
    ////////////////////       CONSTANTS	////////////////////

    public final static int DEFAULT_CAPACITY = 10;
    
    ////////////////////    STATIC VARIABLES 	////////////////////
    ////////////////////       VARIABLES	////////////////////

    /**
     * The array of objects stored in this <code>QSVector</code>.  All must by
     * <code>QuickSerializable</code>. 
     */
    protected QuickSerializable elementData[];

    /**
     * The number of elements currently in the <code>QSVector</code>.
     */
    protected int elementCount;

    /**
     * The number of elements by which to grow the storage array when
     * necessary.  A value of zero causes the storage to double each time.
     */
    protected int capacityIncrement;
    
    ////////////////////      CONSTRUCTORS	////////////////////

    /**
     * Constructs a <code>QSVector</code> of the specified capacity and
     * capacity increment.
     *
     * @param initial_capacity	The initial size for the storage array.
     * @param capacity_increment	The number of elements by which to
     * 				increase the storage array when necessary.
     */
    public QSVector( int initial_capacity, int capacity_increment ) {
	elementData = new QuickSerializable[initial_capacity];
	elementCount = 0;
	capacityIncrement = capacity_increment;
    }

    /**
     * Constructs a <code>QSVector</code> of the specified capacity.
     *
     * @param initial_capacity	The initial size for the storage array.
     */
    public QSVector( int initial_capacity ) {
	this( initial_capacity, 0 );
    }

    /**
     * Constructs a small <code>QSVector</code>.
     */
    public QSVector() {
	this( DEFAULT_CAPACITY, 0 );
    }

    /**
     * Constructs a <code>QSVector</code> containing the contents of the
     * specified array
     *
     * @param array	The array whose contents to add
     */
    public QSVector( QuickSerializable[] array )
    {
	elementData = new QuickSerializable[array.length];
	System.arraycopy( array, 0, elementData, 0, array.length );
	elementCount = array.length;
	capacityIncrement = 0;
    }

    public QSVector(InputBuffer buffer) throws QSException
    {
        capacityIncrement = buffer.nextInt();
        elementCount = buffer.nextInt();

        int len = buffer.nextInt();
        elementData = new QuickSerializable[len];

        for (int i = 0; i < elementCount; ++i) {
            elementData[i] = buffer.nextObject();
        }
    }

    ////////////////////     STATIC METHODS 	////////////////////
    //////////////////// 	METHODS		////////////////////

    public void serialize(OutputBuffer buffer)
    {
	buffer.add( capacityIncrement );
	buffer.add( elementCount );
	buffer.add( elementData.length );

	for( int i=0; i<elementCount; i++ ) {
	    buffer.add( elementData[i] );
	}
    }
    
    //  end methods for ostore.util.QuickSerializable interface

    /**
     * Create an exact copy of this hash
     * @return new SHA1Hash identical in value to this one
     * @exception CloneNotSupportedException if clone() is not supported */
    public Object clone() throws CloneNotSupportedException 
    {
      QSVector result = new QSVector(elementData.length, capacityIncrement);
      System.arraycopy(elementData, 0, result.elementData, 0, elementCount);
      result.elementCount = elementCount;
      result.capacityIncrement = capacityIncrement;
      return result;
    }

    ////////////////////////////////////////////////////
    // Methods which mirror java.util.Vector:

    /**
     * Return the number of elements in this <code>QSVector</code>
     *
     * @return elementCount;
     */
    public int size()
    {
	return elementCount;
    }

    /**
     * Append the specified object to the <code>QSVector</code>.
     *
     * @param object	The <code>QuickSerializable</code> to add.
     * @return <code>true</code>, just like <code>Vector.add()</code>.
     */
    public synchronized boolean add(QuickSerializable object)
    {
	ensureCapacityHelper( elementCount+1 );
	elementData[elementCount] = object;
	elementCount++;
	return true;
    }

    /**
     * Append the specified object to the <code>QSVector</code> at index 
     * <code>index</code>, shifting up the objects following the index, if any.
     *
     * @param object	The <code>QuickSerializable</code> to add.
     * @param index     The index at which to add the 
     *			<code>QuickSerializable</code>.  Must be valid.
     * @return <code>true</code>, just like <code>Vector.add()</code>.
     * @throws ArrayIndexOutOfBoundsException  if index is out of range
     */
    public synchronized boolean add( int index, QuickSerializable object ) {

	ensureCapacityHelper( elementCount + 1 );

	if( (index<0) || (index>=elementCount+1) ) {
	    throw new ArrayIndexOutOfBoundsException(index);
	}

	for( int i = elementCount - 1; i >= index; i-- ) {
	  elementData[i+1] = elementData[i];
	}
	elementData[index] = object;
	elementCount++;
	return true;

    }

    /**
     * Append the specified objects, in order, to the
     * <code>QSVector</code>.
     *
     * @param objects	The <code>QuickSerializable</code>s to add.
     * @return <code>true</code>, just like <code>Vector.add()</code>.
     */
    public synchronized boolean add(QuickSerializable[] objects)
    {
	ensureCapacityHelper( elementCount + objects.length );
	System.arraycopy( objects, 0, elementData, elementCount,
			  objects.length );
	elementCount += objects.length;
	return true;
    }

    /**
     * Grow the storage array if necessary.
     *
     * @param size	The minimum acceptable capacity for the storage array.
     */
    public void ensureCapacity( int size ) {

	ensureCapacityHelper( size );

    }

    /**
     * Grow the storage array if necessary.
     *
     * @param size	The minimum acceptable capacity for the storage array.
     */
    private void ensureCapacityHelper( int size )
    {
	int len = elementData.length;
	
	if( len < size ) {
	    QuickSerializable new_data[];
	    int new_len;
	    if( capacityIncrement <= 0 ) {
		new_len = len*2;
	    } else {
		new_len = len + capacityIncrement;
	    }
	    if( new_len < size ) {
		new_len = size;
	    }
	    new_data = new QuickSerializable[new_len];

	    for( int i=0; i<elementCount; i++ ) {
		new_data[i] = elementData[i];
	    }
	    elementData = new_data;
	}
	return;
    }

    /**
     * Insert the specified object at the specified index.
     *
     * @param index	the index for the element.
     * @return the <code>QuickSerializable</code> object previously stored
     * at the specified index.
     */
    public synchronized QuickSerializable set(int index,
					      QuickSerializable object )
    {
	if( (index<0) || (index>=elementCount) ) {
	    throw new ArrayIndexOutOfBoundsException(index);
	}
	QuickSerializable old = elementData[index];
	elementData[index] = object;	
	return old;
    }

    /**
     * Return the object at the specified index.
     *
     * @param index	the index of the element to return.
     * @return the specified <code>QuickSerializable</code> object, or
     * <code>null</code> if <code>index</code> is out of range.
     */
    public synchronized QuickSerializable get(int index)
    {
	if( (index<0) || (index>=elementCount) ) {
	    throw new ArrayIndexOutOfBoundsException(index);
	}
	return elementData[index];
    }

    /**
     * Remove the object at the specified index.
     *
     * @param index	the index of the element to remove.
     * @return the <code>QuickSerializable</code> object previously stored
     * at the specified index.
     */
    public synchronized QuickSerializable remove(int index)
    {
	if( (index<0) || (index>=elementCount) ) {
	    throw new ArrayIndexOutOfBoundsException(index);
	}
	QuickSerializable old = elementData[index];
	removeRange( index, index+1 );
	return old;
    }


    /**
     * Returns the component at the specified index.<p>
     *
     * This method is identical in functionality to the get method
     * (which is part of the List interface).
     *
     * @param      index   an index into this vector.
     * @return     the component at the specified index.
     * @exception  ArrayIndexOutOfBoundsException  if the <tt>index</tt> 
     *             is negative or not less than the current size of this 
     *             <tt>Vector</tt> object.
     *             given.
     * @see	   #get(int)
     * @see	   java.util.List
     */
    public synchronized Object elementAt(int index) {
	if (index >= elementCount) {
	    throw new ArrayIndexOutOfBoundsException(index + " >= " + elementCount);
	}
	/* Since try/catch is free, except when the exception is thrown,
	   put in this extra try/catch to catch negative indexes and
	   display a more informative error message.  This might not
	   be appropriate, especially if we have a decent debugging
	   environment - JP. */
	try {
	    return elementData[index];
	} catch (ArrayIndexOutOfBoundsException e) {
	    throw new ArrayIndexOutOfBoundsException(index + " < 0");
	}
    }

    /**
     * Remove a range of elements.
     *
     * @param from  The index of the first element to remove.
     * @param to    The index after the last element to remove.
     */
    public synchronized void removeRange( int from, int to ) 
    {
	int num_to_shift = elementCount - to;
	if( num_to_shift > 0 ) 
	    System.arraycopy( elementData, to, elementData, from, num_to_shift );
	
	int oldCount = elementCount;
	elementCount -= (to-from);
	for( int i=elementCount; i<oldCount; i++ ) {
	    elementData[i] = null;
	}
    }

    /**
     * Remove all elements.
     *
     * @see	#clear
     */
    public synchronized void removeAllElements()
    {
	for( int i=0; i<elementCount; i++ ) {
	    elementData[i] = null;
	}
	elementCount = 0;
    }

    /**
     * Tests if the specified <code>QuickSerializable</code> object
     * is a component in this vector.
     *
     * @param elem  a <code>QuickSerializable</code> object
     * @return <code>true</code> if and only if the specified object is
     *   the same as a component of this vector, as determined by the equals
     *   method, <code>false</code> otherwise.
     */
    public synchronized boolean contains( QuickSerializable elem ) {

	for( int i=0; i<elementCount; i++ ) {
	    if( elem.equals( elementData[i] ) ) {
		return true;
	    }
	}
	return false;

    }

    /**
     * Returns the elements of this <code>QSVector</code> as an array.
     * Useful for repeated iterations over the contents.
     *
     * @return an array of size <code>this.size()</code> containing the
     * elements of this <code>QSVector</code> in order.
     */
    public synchronized QuickSerializable[] toArray()
    {
	QuickSerializable array[] = new QuickSerializable[elementCount];
	System.arraycopy( elementData, 0, array, 0, elementCount );
	return array;
    }

    /**
     * Return a hashcode, generated as in <code>java.util.Vector</code>, which
     * was in turn specified by <code>java.util.List</code>.
     *
     * @return a hashcode which incorporates the hashcodes of all the elements
     * in the <code>QSVector</code>
     */
    public synchronized int hashCode()
    {
	int hashCode = 1;
	QuickSerializable object;
	for( int i=0; i<elementCount; i++ ) {
	    object = elementData[i];
	    hashCode = 31*hashCode + (object==null ? 0 : object.hashCode());
	}
	return hashCode;
    }

    /**
     * Compare this <code>QSVector</code> to another <code>Object</code>.
     *
     * @return <code>true</code> iff the other <code>Object</code> is also a
     * <code>QSVector</code> and contains <code>equal</code> elements.
     */ 
    public synchronized boolean equals(Object o)
    {
	boolean matched = false;
	
	if( o == this ) {
	    matched = true;
	} else if( o instanceof QSVector ) {
	    QSVector other = (QSVector)o;
	    if( elementCount != other.elementCount ) {
		matched = false;
	    } else {
		matched = true;
		for( int i=0; i<elementCount; i++ ) {
		    if( !( elementData[i].equals(other.elementData[i])) ) {
			matched = false;
			break;
		    }
		}
	    }
	}
	return matched;
    }

    /**
     * Removes all of the elements from this Vector.  The Vector will
     * be empty after this call returns (unless it throws an exception).
     *
     * @since 1.2
     */
    public void clear() {
        removeAllElements();
    }

    /**
     * Produce a human-readable version of this <code>QSVector</code>,
     * which includes all the elements.
     *
     * @return a <code>String</code> representation of this
     * 		<code>QSVector</code>.
     */
    public synchronized String toString()
    {
	StringBuffer sbuf = new StringBuffer( "[" );
	for( int i=0; i<size(); i++ ) {
	    sbuf.append( ' ' );
	    sbuf.append( '(' );
	    sbuf.append(  i  );
	    sbuf.append( ')' );
	    sbuf.append( '<' );
	    sbuf.append( elementData[i].toString() );
	    sbuf.append( '>' );
	}
	sbuf.append( ']' );
	return sbuf.toString();
    }
}

