/**
 * RandomExponential.java
 *
 * Copyright (c) 2003 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 */

package ostore.util;
import java.util.Random;

/**
 * A utility class for generating random values from a negative exponential
 * distribution.  <code>PDF( x ) = -lambda * exp( -lambda * x )</code>.
 * The mean and standard deviation are both <code>1/lambda</code>.
 *
 * @author	Dennis Geels
 * @version	$Id: RandomExponential.java,v 1.1 2003/04/19 00:55:41 strib Exp $
 */

public class RandomExponential extends Random
{
    ////////////////////       VARIABLES	////////////////////

    /**
     * Equal to <code>-1/lambda</code>.
     */
    private double negative_mean;
    
    ////////////////////      CONSTRUCTORS	////////////////////

    public RandomExponential( double lambda )
    {
	super();
	init( lambda );	
    }

    public RandomExponential( double lambda, long seed )
    {
	super( seed );
	init( lambda );
    }
    
    ////////////////////     STATIC METHODS 	////////////////////

    public static void main( String args[] )
    {
	double lambda = .01;
	RandomExponential r = new RandomExponential(lambda);
	System.out.println( "negative_mean = " + r.negative_mean );	
	double sum = 0;
	double iterations = 100000;
	for( int i=0; i<iterations; i++ ) {
	    double x = r.nextExponential();
	    System.out.println( "x = " + x );
	    sum += x;
	}
	System.out.println( "mean = " + (sum/iterations) );
    }
    
    //////////////////// 	METHODS		////////////////////

    final private void init( double lambda )
    {
	if( lambda <= 0 ) {
	    throw new IllegalArgumentException( "non-positive lambda: " +
						lambda );
	}
	this.negative_mean = (-1.0/lambda);		
    }

    /**
     * Returns a positive number drawn pseudo-randomly from the exponential
     * distribution with parameter <code>this.lambda</code>. <p>
     * This method uses the "inverse transform" method.  Given a number
     * <code>u</code> drawn uniformly at random from the interval
     * <code>[0,1)</code>, we calculate a value <code>x</code> from our
     * distribution as <code> x = (-1/lambda) * log( 1 - u ) </code>.
     */
    public double nextExponential()
    {
	double u = nextDouble();
	double x = this.negative_mean * StrictMath.log( 1 - u );
	return x;
    }
}
