/**
 * SHA1NativeIF.java
 *
 * @author   Hakim Weatherspoon
 * @version  $Id: SHA1NativeIF.java,v 1.3 2002/09/07 02:19:28 strib Exp $
 *
 * Copyright (c) 2001 Regents of the University of California.
 * All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *  3. Neither the name of the University nor the names of its contributors
 *     may be used to endorse or promote products derived from this software
 *     without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 *  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 *  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 *  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 *  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 *  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 *  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 *  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 *  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 *  SUCH DAMAGE.
 **/
package ostore.util;

import java.security.DigestException;
import java.security.MessageDigestSpi;
   
/**
 * <CODE>SHA1HashNativeIF</CODE> is an <CODE>interface</CODE> to a 
 * <CODE>native</CODE> <CODE>C</CODE> implementation of 
 * a <CODE>SHA-1</CODE> {@link java.security.MessageDigest}.  
 *
 * The result of <CODE>SHA1NativeIF</CODE> is used by the 
 * {@link ostore.util.SHA1Hash} and {@link ostore.util.SHA1HashBuffer} 
 * class and by all entities which are SHA1 hashes.
 * (In the prototype, this means Guids and VHashes).
 *
 * @author   Hakim Weatherspoon
 * @version  $Id: SHA1NativeIF.java,v 1.3 2002/09/07 02:19:28 strib Exp $
 **/
public class SHA1NativeIF extends MessageDigestSpi 
{
  ////////////////////       CONSTANTS		////////////////////
  ////////////////////    STATIC VARIABLES 	////////////////////

  /**
   * <CODE>available</CODE> indicates if the <CODE>C</CODE> implementation
   * of <CODE>SHA-1</CODE> is available. */
  public static boolean available = false;

  ////////////////////       VARIABLES		////////////////////

  /**
   * <CODE>mdState</CODE> is an instance variable used to hold
   * message digest state between calls. */
  private int[] mdState;

  ////////////////////  STATIC CONSTRUCTORS	////////////////////

  static
  {
    try
    {
      System.loadLibrary ("ostore_util_sha1_c");
      available = true;
      ostore.util.Carp.out("SHA1NativeIF.<static_init>:" +
			   " Successfully loaded JNI library" +
			   " ostore_util_sha1_c");
    }
    catch (UnsatisfiedLinkError e1)
    {
      System.err.println("Could not load JNI SHA1.  " +
			 "Using Java implementation instead.  (This will "
			 + "affect performance, but not correctness.)");
      available = false;
    }
  }

  ////////////////////      CONSTRUCTORS	//////////////////// 

  /**
   * Construct according to the given fields. */
  public SHA1NativeIF() {
      engineInit();
  }

  ////////////////////     STATIC METHODS 	////////////////////
  ////////////////////	 	METHODS		////////////////////

  /**
   * <CODE>engineInit</CODE> initializes the digest for further use. */
  public native void engineInit();

  /**
   * <CODE>engineReset</CODE> resets the digest for further use. */
  public native void engineReset();

  /**
   * <CODE>engineGetDigestLength</CODE> returns the digest length in 
   * bytes.
   *
   * @return The digest length in bytes. */
  public int engineGetDigestLength() {
    return SHA1Hash.NUM_BYTES;
  }
    
  /**
   * <CODE>engineUpdate</CODE> updates the digest using the specified byte.
   *
   * @param input The byte to use for the update. */
  public native void engineUpdate(byte input);

  /**
   * <CODE>engineUpdate</CODE> updates the digest using the specified 
   * array of bytes, starting at the specified offset.
   *
   * @param input  The array of bytes to use for the update.
   * @param offset The offset to start from in the array of bytes.
   * @param len    The number of bytes to use, starting at 
   *               <code>offset</code>. */
  public native void engineUpdate(byte[] input, int offset, int len);

  /**
   * <CODE>engineDigest</CODE> completes the hash computation by 
   * performing final operations such as padding. Once 
   * <code>engineDigest</code> has been called, the engine should is 
   * reset (see {@link #engineReset() engineReset}).  
   *
   * @return The array of bytes for the resulting hash value.  */
  public native byte[] engineDigest();

  /**
   * <CODE>engineDigest</CODE> completes the hash computation by 
   * performing final operations such as padding. Once 
   * <code>engineDigest</code> has been called, the engine is 
   * reset (see {@link #engineReset() engineReset}).  
   *
   * @param buf    The output buffer in which to store the digest.
   * @param offset Offset to start from in the output buffer.
   * @param len    Number of bytes within buf allotted for the digest.
   *                 The presence of this parameter is solely
   *                 for consistency in our API's.  If the value of 
   *                 this parameter is less than the actual digest 
   *                 length, the method will throw a DigestException.
   *                 This parameter is ignored if its value is greater 
   *                 than or equal to the actual digest length.
   *
   * @return the length of the digest stored in the output buffer.
   * 
   * @exception DigestException if an error occurs. */
  public native int engineDigest(byte[] buf, int offset, int len) throws DigestException;


  /**
   * <CODE>engineUpdateAndDigest</CODE> in one call updates the 
   * digest using the specified array of bytes, starting at the 
   * specified offset.  And completes the hash computation by 
   * performing final operations such as padding. Once 
   * <code>engineUpdateAndDigest</code> has been called, the engine is 
   * reset (see {@link #engineReset() engineReset}).  
   *
   * @param buf     The output buffer in which to store the digest.
   * @param out_off Offset to start from in the output buffer.
   * @param out_len Number of bytes within buf allotted for the digest.
   *                  The presence of this parameter is solely
   *                  for consistency in our API's.  If the value of 
   *                  this parameter is less than the actual digest 
   *                  length, the method will throw a DigestException.
   *                  This parameter is ignored if its value is greater 
   *                  than or equal to the actual digest length.
   * @param input   The array of bytes to use for the update.
   * @param in_off  The offset to start from in the array of bytes.
   * @param in_len  The number of bytes to use, starting at 
   *                  <code>offset</code>.
   *
   * @return The length of the digest stored in the output buffer.
   * 
   * @exception DigestException if an error occurs. */
  public native int engineUpdateAndDigest(byte[] buf, int out_off, int out_len, byte[] input, int in_off, int in_len) throws DigestException;

  /**
   * Tests the throughput of SHA1Native from entirely within C.  Creates
   * random byte arrays of size <code>arrSz</code> and times how long it
   * takes to hash <code>iterations</code> of them.
   *
   * @param arrSz           The number of bytes to hash
   * @param iterations      The number of times to hash
   * @return   The number of milliseconds it took to perform the hashes
   */
  public native long testTput(int arrSz, int iterations );

  ////////////////////     INNER CLASSES	////////////////////

}

