/*
 *  Copyright (c) 2001 Regents of the University of California.  All
 *  rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *  1. Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *  2. Redistributions in binary form must reproduce the above
 *     copyright notice, this list of conditions and the following
 *     disclaimer in the documentation and/or other materials provided
 *     with the distribution.
 *  3. Neither the name of the University nor the names of its
 *     contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 *  IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *  FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 *  REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 *  (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 *  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 *  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 *  EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;

/**
 * SecureHash is the interface used to represent classes which
 * are secure hashes
 * An example is the SHA1 hash
 *
 * @author	Chris WElls
 * @version	$Id: SecureHash.java,v 1.16 2002/09/24 23:02:46 eaton Exp $
 */
public interface SecureHash 
extends Cloneable, Comparable, QuickSerializable {

	/**
	 * Return the number of bytes used in the representation of this hash
	 * (for example, SHA1 uses 20 bytes)
	 * @return number of bytes used in the implementation
	 */
	public int size();

	/**
	 * Determines if this SecureHash is a NULL hash or not.
	 * @return true iff this Hash has a NULL value
	 */
	public boolean isNull();

	/**
	 * Returns the implementation's NULL hash
         * (typically all zeroes).
	 * @return the implementation's null hash
	 */
	public SecureHash nullHash();

	/**
	 * Returns the bytes contained in this hash
	 * @return the bytes of this hash
	 */
	public byte[] bytes();

	/**
	 * Compares this SecureHash with another hash, and returns
	 * true iff they are the same type of hash and have the
	 * same byte representation
	 * @param o the object with which to compare this SecureHash
	 * @return true iff this and o contain the same byte values
	 */
	public boolean equals(Object o);

	/**
	 * Calculates a Java hash code for this SecureHash object
	 * (typically by taking its lower order bytes)
	 * @return the Java hash code for this SecureHash
	 */
	public int hashCode();

	/**
	 * Hashes over the input bytes, starting at offset, going for len
	 * @param bytes byte array over which to hash
	 * @param offset offset in bytes[] at which to begin hashing
	 * @param len number of bytes over which to hash
	 */
	public void hash(byte bytes[], int offset, int len);

	/**
	 * Hashes over the input bytes
	 * @param bytes byte array over which to hash
	 */
	public void hash(byte bytes[]);

	/**
	 * Hashes over a QuickSerializable object
	 * @param o the object over which to hash
	 */
	public void hash(QuickSerializable o);

	/**
	 * Hashes over the bytes of the input String
	 * @param str the String over which to hash
	 */
	public void hash(String str);

	/**
	 * Verifies a QuickSerializable object
	 * @param o the object over which to hash
	 * @return true iff the input object hashes to
	 * a hash value equal to this hash value
	 */
	public boolean verify(QuickSerializable o);
	
	/**
	 * Converts a few bytes of this SecureHash to a human-readable String
	 * @return human-readable String representation of this Hash
	 */
	public java.lang.String toString();

	/**
	 * Converts all bytes of this SecureHash to a human-readable String
	 * @return human-readable String representation of this Hash
	 * containing every nibble of this Hash's value
	 */
	public java.lang.String fullString();

    /**
     * This is the clone routine for Fragment.  It deeply clones a fragment
     * (meaning that it copies all of the data, not just references). */
    public Object clone() throws CloneNotSupportedException;
}
