/**
 * Copyright (c) 2002 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package ostore.util;
import ostore.dispatch.Classifier;
import ostore.dispatch.Filter;
import seda.sandStorm.api.ConfigDataIF;
import seda.sandStorm.api.EventHandlerIF;
import seda.sandStorm.api.EventHandlerException;
import seda.sandStorm.api.QueueElementIF;
import seda.sandStorm.api.SinkException;
import seda.sandStorm.api.SinkIF;

/**
 * Implements functionality common to most stages.
 *
 * @author	Sean C. Rhea
 * @version	$Id: StandardStage.java,v 1.7 2003/05/02 17:08:36 srhea Exp $
 */

public abstract class StandardStage implements EventHandlerIF {
    
    protected Classifier classifier;
    protected NodeId my_node_id;
    protected SinkIF my_sink;
    protected String tag;
    protected String [] event_types = {};
    protected String [] inb_msg_types = {};
    protected String [] outb_msg_types = {};

    protected boolean DEBUG;

    protected long now_ms () {
	return Clock.current_date (my_node_id).getTime ();
    }

    protected final void BUG (String msg) {
	Carp.die (tag + ": " + msg +
                 "  This is a bug in " + this.getClass ().getName () +
		 ".  Please report it to the class's author.");
    }

    public void init(ConfigDataIF config) throws Exception {

	tag = config.getStage().getName();
	Debug.turnOn(tag);
	my_sink = config.getStage().getSink();

	my_node_id = new NodeId (config.getString ("node_id"));
	classifier = Classifier.getClassifier (my_node_id);

	Filter filter;
	for (int i = 0; i < event_types.length; ++i) {
	    filter = new Filter ();
	    if (! filter.requireType (Class.forName (event_types [i])))
		BUG (tag + ": could not require type " + event_types [i]);
	    if( DEBUG )
		Debug.printtagln( tag, "subscribing to " + event_types [i] );
	    classifier.subscribe (filter, my_sink);
	}

	for (int i = 0; i < inb_msg_types.length; ++i) {
	    ostore.util.TypeTable.register_type (inb_msg_types [i]);
	    filter = new Filter ();
	    if (! filter.requireType (Class.forName (inb_msg_types [i])))
		BUG (tag + ": could not require type " + inb_msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (true)))
		BUG (tag + ": could not require inbound = true for " 
			+ inb_msg_types [i]);
	    if( DEBUG )
		Debug.printtagln( tag, "subscribing to " + inb_msg_types[i]);
	    classifier.subscribe (filter, my_sink);
	}

	for (int i = 0; i < outb_msg_types.length; ++i) {
	    filter = new Filter ();
	    if (! filter.requireType (Class.forName (outb_msg_types [i])))
		BUG (tag + ": could not require type " + outb_msg_types [i]);
	    if (! filter.requireValue ("inbound", new Boolean (false)))
		BUG (tag + ": could not require inbound = false for " 
			+ outb_msg_types [i]);
	    if( DEBUG )
		Debug.printtagln( tag, "subscribing to " + outb_msg_types[i]);
	    classifier.subscribe (filter, my_sink);
	}
    }

    public void destroy() {}

    public void handleEvents(QueueElementIF element_array[]) 
    throws EventHandlerException {
	for (int i = 0; i < element_array.length; ++i)
	    handleEvent(element_array[i]);
    }

    protected void dispatch (QueueElementIF item) {
	try {
	    if( DEBUG ) Debug.printtagln( tag+".dispatch", "dispatching: "
					  + item + " type=" +
					  item.getClass().getName() );
	    classifier.dispatch (item);
	}
	catch (SinkException e) {
	    BUG ("could not dispatch " + item + ".");
	}
    }
}


