# Tools.pm
#
# Copyright (c) 2001 Regents of the University of California.
# All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#  3. Neither the name of the University nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
#  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
#  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#  SUCH DAMAGE.

#
# $Id: Tools.pm,v 1.73 2003/06/23 02:26:11 geels Exp $
#
# Useful common subroutines and data structures for running OceanStore servers.

package Tools;

use Exporter();
use File::Path "rmtree";
use strict;
use vars qw( @ISA @EXPORT $ring_size %ostore_dirs $seed_counter $disable_jit
	     $use_nohup $use_tapas $num_parallel_calls $ignore_remote_pids
	     $tmp_dir $hostname @child_pids @logs @configs @dirs @profiles
	     @remote_kill_cmd @concat_cmd @running_hosts);
use POSIX qw( sys_wait_h );

@ISA = ("Exporter");
@EXPORT = qw( &is_localhost &remote_tmp_dir &wait_for
	      &wait_for_ready &wait_for_tapestry_ready &wait_for_network_ready
	      &wait_for_return &returned_zero &returned_zero_or_timeout 
	      &is_clean &override &insert &interpolate
	      &fork_process &fork_sandstorm &fork_java
	      &assign_keys &create_node &create_host &ssh_args &remote_run_dir
	      &start_local &start_remote &start_multiple_remote &fork_remote
	      &extract_remote_pid &call_remote_sub
	      &exit_all &clean_up &die &catch_sig &user_exit
	      $ring_size %ostore_dirs $seed_counter $disable_jit
	      $use_nohup $use_tapas $tmp_dir $hostname $num_parallel_calls
	      @child_pids @logs @configs @dirs @profiles @remote_kill_cmd
	      $ignore_remote_pids  @running_hosts);

$| = 1;

############################################################
# GLOBAL VARIABLES

# A counter for unique random number generator seeds.
$seed_counter = ($ENV{"OSTORE_SEED"} or 1000);

# The number of servers in an inner ring (mostly constant).
$ring_size = 4;

# Determines which set of keyshares the inner ring uses.
#  Default: 512-bit keys (ir.pub, ir-0.priv, etc.)
my $keyshare_prefix = ($ENV{"KEYSHARE_PREFIX"} or "ir");

# Stores the keypair used by each Router, keyed by the node's tag.
# This keypair determines the NodeId of the server, but is independent of
# the keys used to sign client updates and inner ring responses.
my %keys_by_tags;

# True (1) iff the JIT should be disabled
$disable_jit = 0;

# True (1) iff remote processes are to be daemonized
$use_nohup = 0;

# True (1) iff it is ok to forget about remote daemonized processes.
# Currently only &fork_remote reads this variable.
$ignore_remote_pids = 0;

# True (1) iff Tapestry stages should be replaced by Tapas stages.
$use_tapas = 0;

# Stage includes required to use Tapas.
my @tapas_stage_includes = ("TapasRouter.stg", "TapestryEmu.stg", "Patchwork.stg");

# Stages required to use Tapestry.
my @tapestry_stages = ("StaticTClient", "TClient", "DynamicTClient",
		       "DTClient", "Router", "Federation");

# Defined if remote logs should _not_ be copied back to local machine on failure.
# Set to a positive number if logs should not be copied even on success.
my $disable_copyback = $ENV{"OSTORE_DISABLE_COPYBACK"};

# True (1) iff the logs, configs, etc. should <b>not</b> be cleaned up on a
# successful exit.
my $disable_cleanup = $ENV{"OSTORE_DISABLE_CLEANUP"};

# parameterize the local /tmp or alternative.
$tmp_dir = ($ENV{"OSTORE_TMP"} or "/tmp");

# PIDs of all children forked (will be killed)
@child_pids = ();

# Names of all config files (will be deleted)
@configs = ();

# Names of all log files (will be deleted)
@logs = ();

# Names of all profiling output files (will be deleted)
@profiles = ();

# Names of all temp dirs (will be deleted)
@dirs = ();

# The number of remote calls (ssh/scp) to run in parallel during experiment
# startup and shutdown.
$num_parallel_calls = 20;

# Get the hostname.
$hostname = `hostname`;
chomp $hostname;

########## Remote Host information

# A list of host objects for all remote hosts.
@running_hosts = ();

########## Environment variables:

# A few useful directories
if( $ENV{"OSTORE_RUN_DIR"} ) {
    $ostore_dirs{"run_dir"} = $ENV{"OSTORE_RUN_DIR"};
} elsif( $ENV{"PWD"} ) {
    $ostore_dirs{"run_dir"} = $ENV{"PWD"};
} else {
    $ostore_dirs{"run_dir"} = `pwd`;
    chomp $ostore_dirs{"run_dir"};
}

if( $ENV{"OSTORE_CFG_DIR"} ) {
    $ostore_dirs{"cfg_dir"} = $ENV{"OSTORE_CFG_DIR"};
} else {
    $ostore_dirs{"cfg_dir"} = $ostore_dirs{"run_dir"} . "/cfg";
}

if( $ENV{"OSTORE_GRAPH_DIR"} ) {
    $ostore_dirs{"graph_dir"} = $ENV{"OSTORE_GRAPH_DIR"};
} else {
    $ostore_dirs{"graph_dir"} = $ostore_dirs{"run_dir"} . "/graphs";
}

$ostore_dirs{"key_dir"} = $ostore_dirs{"run_dir"} . "/keys";

##### Executables:

my @genkey_cmd = ($ostore_dirs{"run_dir"}."/run-java",
		  "ostore.security.GenerateRSAKeyPair");

my $sandstorm_main = "seda.sandStorm.main.Main";

my $ssh_cmd = "ssh";

my @scp_cmd = ("scp", "-q");

my $nohup_cmd = "nohup";

my @perl_exp_cmd = ("perl", "-e");

@concat_cmd = $ostore_dirs{"run_dir"} . "/cfgcat.pl";

@remote_kill_cmd = ("kill");

my @remote_rm_cmd = ("rm", "-rf");

############################################################
# SUBROUTINES

# Substitute runtime values into configuration files.
# Any stand-alone occurrence of a key is replaced inline with its value.
# If value is not present in placeholder map, environment is checked for
# value before failing.
# Unless the optional fourth parameter (<code>$ignore_missing_vars</code>) is
# true, choke and die on any undefined variables.
#
# Pass in a placeholder->value hashtable,
#	(e.g. $map{"\$\{FederationID\}"} = "$hostname:3630").
# Usage: &interpolate( <template>, <output_filename>, <map_ref>,
#		[<ignore_missing_vars>] )

sub interpolate {
    my ($template, $output, $map_ref, $ignore_missing_vars) = @_;
    my %map = %{$map_ref};

    open IN, "<$template" or &die( "can't open IN=$template");
    open OUT, ">$output"  or &die( "can't open OUT=$output");

    # Write each line of the template into the output file, replacing
    # special keys with runtime values.
    my $line_no = 0;
    while( <IN> ) {
	$line_no++;
	while( /^[^\#]*\$\{(.*)\}/ ) {
	    # found a variable - replace from map.
	    my $variable = $1;
	    my $value = $map{$variable};
            $value = $ENV{$variable} if ( !defined $value); # SEC Hack ...
	    if( defined $value ) {
		s/\$\{$variable\}/$value/;
	    } elsif( ! $ignore_missing_vars ) {
		print "ERROR: undefined variable \"$variable\"" .
		    " on line $line_no of file $template\n";
		kill "TERM", $$;
	    } else {
#		print "Ignoring undefined variable \"$variable\"" .
#		    " on line $line_no of file $template\n";
		last;
	    }
	}
	print OUT $_;
    }

    close( IN );
    close( OUT );
}

# Swap in new values for named variables
# Any line defining a specified key will be rewritten.  The new line will
#	set the key to its provided value.
#
# Pass in a variable->value hashtable, (e.g. $map{"Profile"} = "false").
# Usage: &override( <cfg_filename> <%hashtable> )

sub override {
    my ($cfg, %map) = @_;
    my $tmp_name = "$cfg.$$.tmp";
    
    open( IN, "<$cfg" );
    open( OUT, ">$tmp_name" );

    # Fix here, too.
    my $key;
  LINE: while( <IN> ) {
      foreach $key (keys %map) {
	  if( /^(\s*)$key\b.*/ ) {
	      print OUT "$1$key\t$map{$key}\n";
	      next LINE;
	  }
      }
      print OUT $_;	# no match
  }
    
    close( IN );
    close( OUT );

    # now replace the old copy with the new one.
    rename( $tmp_name, $cfg );
}

# Adds new variable-value mappings to a config file
#
# All fields will be added to the "<initargs>" section of the named stage,
#  to the global initargs if "global" is specified,
#  or to EVERY stage if "ALL" is specified.
# "ALL" is stronger than "global", because the global initargs can be
#  overridden by a stage's local initargs.
#
# Pass in a variable->value hashtable, (e.g. $map{"Profile"} = "false").
# Usage: &insert( <cfg_filename>, (<stagename>|"global"|"ALL"), <%hashtable> )

sub insert {
    my ($cfg, $stage, %map) = @_;
    my $tmp_name = "$cfg.$$.tmp";
    my $key;
    
    open( IN, "<$cfg" );
    open( OUT, ">$tmp_name" );

    # find stage/global
  STAGE_LOOP: while( 1 ) {
      while( <IN> ) {
	  print OUT $_;
	  if (m/^([^\#]*)<([^\/]+)>(.*)$/) {
	      last if( ($stage eq $2) or ($stage eq "ALL") );
	  } elsif (m/^([^\#]*)<\/stages>(.*)$/) {
	      last STAGE_LOOP;	# no more stages.
	  }
      }
      # find initargs
      while( <IN> ) {
	  print OUT $_;
	  last if( /^\s*<initargs>/ );
      }
      # replace old mappings
      my %remaining = %map;	# Don't modify original map.
    LINE: while( <IN> ) {
	foreach $key (keys %remaining) {
	    if( /^(\s*)$key\b.*/ ) {
		# found one
		print OUT "$1$key\t$remaining{$key}\n";
		delete $remaining{$key};	# avoid repeats
		next LINE;
	    }
	}
	
	if( /^(\s*)<\/initargs>/ ) {
	    # insert remaining mappings
	    foreach $key (keys %remaining) {
		print OUT "$1    $key\t$remaining{$key}\n";
	    }
	    print OUT $_;	# "</initargs>"
	    next STAGE_LOOP;	# start looking for next stage.
	}
	print OUT $_;	# non-matching mapping
    }
  }
    # Copy rest of file
    while( <IN> ) {
	print OUT $_;
    }
    
    close( IN );
    close( OUT );

    # now replace the old copy with the new one.
    rename( $tmp_name, $cfg );
}

# Adds a "<!include>" directive to a config file.
# All directives are inserted at end of "<stages>" section of file.
#
# Usage: &add_components( <cfg_filename>, <component list> );

sub add_components {
    my ($cfg, @components) = @_;
#    print "Tools::add_components: $cfg\n";
    $cfg =~ m%/?([^/]+?)(\.tmp)?$%;
#    print "Tools::add_components: $1\n";
    my $new_cfg = "$tmp_dir/experiment-$$-$1.tmp";
#    print "Tools::add_components: $new_cfg\n";
    
    open( IN, "<$cfg" );
    open( OUT, ">$new_cfg" );

    my %key_words = ();

    # copy to end of "stages" section.
    my $indentation;
    while( <IN> ) {
	if(/^\s*<\!include\s+(.*)>/) {
	  $key_words{$1} = $1;
	}

        if( m%^(\s*)</stages>% ) {
	    $indentation = $1;
	    last;
	} else {
	    print OUT $_;
	}
    }

    # Add directive for each component.
    my $component;
    foreach $component (@components) {
	if(!(defined $key_words{$component})) {
	    print OUT "$indentation    <!include $component>\n";
	}
    }
    
    # copy rest of file
    print OUT $_;	# "</stages>"
    while( <IN> ) {
	print OUT $_;
    }
    
    close( IN );
    close( OUT );
    return $new_cfg;
}


# Removes stage definitions from a configuration file.
# Only works on expanded definitions, not "<!include>" directives.
#
# Usage: &remove_stages( <cfg_filename>, <stage list> );

sub remove_stages {
    my ($cfg, @remove_list) = @_;

    # Strip out base filename, add proper ".tmp" suffix:
#    print "Tools::remove_stages: $cfg\n";
    $cfg =~ m%/?(experiment-\d+-)*([^/]+?)(\.tmp)?(\.(\d))?$%;
#    print "Tools::remove_stages: \"$1\" $2 $3 \"$4\" \"$5\"\n";    
    my $new_cfg = "$tmp_dir/experiment-$$-$2.tmp";
    if( $3 ) {
	my $suffix = (defined $4) ? (".".($5+1)) : ".1";
	$new_cfg .= $suffix;
    }
#    print "Tools::remove_stages: $new_cfg\n";
    
    open( IN, "<$cfg" );
    open( OUT, ">$new_cfg" );

    my %should_remove;
    foreach (@remove_list) {
	$should_remove{$_} = 1;
    }
    
    # copy to end of "stages" section.
    my $indentation;
    STAGE_LOOP: while( <IN> ) {
	if (m/^([^\#]*)<([^\/]+)>(.*)$/) {
	    # The start of a stage entry
	    my $stage = $2;
	    if( $should_remove{$stage} ) {
		# Skip to the end of this stage.
		while( <IN> ) {
		    next STAGE_LOOP if (m/^([^\#]*)<\/$stage>(.*)$/);
		}
	    }
	}
	# Fell through if clauses: keep this line:
	print OUT $_;
    }
    
    close( IN );
    close( OUT );
    return $new_cfg;
}

# Returns true (1) if passed the name of the current host, otherwise false (0).
# This method compares the named machine's IP address to that of
# `hostname`.  It also looks for "localhost" or "127.0.0.1".

sub is_localhost {
    my $hostname = shift;
    if( ($hostname eq "localhost") or
	($hostname eq "127.0.0.1") or
	(gethostbyname( $hostname ) eq gethostbyname( $Tools::hostname )) ) {
	return 1;
    } else {
	return 0;
    }
}

# parameterize the remote /tmp or alternative.
#
# Reads a value from an environment variable if possible.
# chooses a preset /tmp alternative for certain machines.
# Otherwise, uses /tmp.
sub remote_tmp_dir {
    my $remote_host = shift;
    my $remote_tmp;
    
    if( &is_localhost( $remote_host ) ) {
	$remote_tmp = $tmp_dir;
    } elsif( $ENV{"OSTORE_REMOTE_TMP"} ) {
	# deprecated
	die( "warning: OSTORE_REMOTE_TMP is obsolete.  ".
	     " Use REMOTE_OSTORE_TMP instead\n" );
    } elsif( $ENV{"REMOTE_OSTORE_TMP"} ) {
	$remote_tmp = $ENV{"REMOTE_OSTORE_TMP"};
    } elsif( &in_local_cluster( $remote_host ) ) {
	# X or ibm cluster
	$remote_tmp = "/stripe";
    } else {
	# who knows.
	$remote_tmp = "/tmp";
    }
    return $remote_tmp;
}

# fork a process; return its PID.
# redirect STDOUT and STDERR to the named log.
# automatically adds the PID to @child_pids and the log filename to @logs.
#
# Usage: &fork_process( <log_filename>, <executable>, [<arg1>, ..., <argN>] );

sub fork_process {
    my $log = shift( @_ );
    push @logs, $log;

#    print "fork_process: calling (@_)\n";
    
    my $pid = fork();
    if( $pid > 0 ) {
	#parent;  return PID for child.
	push @child_pids, $pid;
	return $pid;
	
    } elsif( defined $pid ) {
	# child; run the named process..
	
	open STDOUT, ">$log" or &die( "can't redirect STDOUT");
	open STDERR, ">&STDOUT" or &die( "can't redirect STDERR");

	# now exec the process as specified.
	exec (@_);
    }
    
    # should never reach this point.
    &die( "fork failed" );
}

# fork a process to run sandstorm; return its PID.
# redirect STDOUT and STDERR to the named log.
# automatically adds the PID to @child_pids and the log filename to @logs.
#
# Usage: &fork_sandstorm( <config_filename>, <log_filename> );

sub fork_sandstorm {
    my ($cfg, $log) = @_;

    my $return_val;
    if( $disable_jit ) {
	$return_val = &fork_java( $log, "-nojit", $sandstorm_main, $cfg );
    } else {
	$return_val = &fork_java( $log, $sandstorm_main, $cfg );
    }	
    return $return_val;
}

# fork a process to run java; return its PID.
# redirect STDOUT and STDERR to the named log.
# automatically adds the PID to @child_pids and the log filename to @logs.
#
# Usage: &fork_java( <log_filename>, <mainclass>, [<arg1>, ..., <argN>] );

sub fork_java {
    my $log = shift( @_ );

    my $run_java = $ostore_dirs{"run_dir"} . "/run-java";
    
    return &fork_process( $log, $run_java, @_ );
}

# Greps the specified log to find the sandstorm ready signal
#
# Usage: &wait_for_ready( <log_filename> )
sub wait_for_ready {
    my $log = shift @_;

    &wait_for( $log, "^Sandstorm: Ready" );
}

# Greps the specified log to find the Tapestry ready signal
#
# Usage: &wait_for_ready( <log_filename> )
sub wait_for_tapestry_ready {
    my $log = shift @_;

    &wait_for( $log, "Tapestry: ready" );
}

# Greps the specified log to find the Network ready signal
#
# Usage: &wait_for_network_ready( <log_filename> )
sub wait_for_network_ready {
    my $log = shift @_;

    &wait_for( $log, "Network.*now listening on port.*" );
}

# Greps a log for a string
#
# Usage: &wait_for( <log_filename>, <string>, <timeout> )
# TODO: grep through the log as it is written, instead of repeatedly
#  scanning the whole file.  Give up after <timeout> seconds.
sub wait_for {
    my ($log, $string, $timeout) = @_;
    if( ! defined $timeout ) {
	$timeout = 600;
    }
    $timeout += time;

    while( ! -r $log ) {
	if( time > $timeout ) {
	    warn "Stopping - Timed out\n";
	    kill "TERM", $$;
	}
	sleep 1;
    }
    while( `grep -c \'$string\' $log` == 0 ) {
	if( time > $timeout ) {
	    warn "Stopping - Timed out\n";
	    kill "TERM", $$;
	} elsif( ! &is_clean( $log ) ) {
	    # catch exceptions and kill self
	    # Send a signal (rather than just calling &die) to
	    #  allow main scripts to override &die in interesting ways.
	    kill "TERM", $$;
	} else {
	    # don't hog the processor
	    sleep 2;
	}
    }
}

# Waits for child processes to exit and checks their return values.
# Returns 1 (true) if all returned 0, or 0 (false) if any child returned a
# nonzero return value.
#
# Usage: &returned_zero( ["--quiet"] (<pid1>,..., <pidN>) );
sub returned_zero {
    my $quiet = shift;
    if( $quiet ne "--quiet" ) {
	unshift @_, $quiet;
	$quiet = 0;
    }
    my $child;
    foreach $child (@_) {
	waitpid $child, 0; #no flags
	if( $? > 0 ) {
	    if( not $quiet ) {
		warn( "Child $child exited with value " . ($? >> 8) .
		      ", signal " . ($? & 255) . "\n" );
	    }
	    return 0;
	}
    }
    return 1;
}

# Waits for child processes to exit and checks their return values.
# Returns 1 (true) if all returned 0, or 0 (false) if any child returned a
# nonzero return value.  Times out after a given number of seconds.  Returns
# -1 on a timeout.  If the given timeout is non-positive, it's set to 1.
#
# Usage: &returned_zero_or_timeout( ["--quiet"] timeout (<pid1>,..., <pidN>) );
sub returned_zero_or_timeout {
    my $quiet = shift;
    if( $quiet ne "--quiet" ) {
	unshift @_, $quiet;
	$quiet = 0;
    }
    my $timeout = shift;
    if( $timeout <= 0 ) {
	$timeout = 1;
    }
    my $timedone = time + $timeout;
    my $child;
RZOT_FOR:
    foreach $child (@_) {
	while( time < $timedone ) {
	    if( waitpid( $child, WNOHANG ) == 0 ) {
		# the child hasn't exited yet, so sleep a while and try again
		sleep 1;
		next;
	    } 
	    if( $? > 0 ) {
		if( not $quiet ) {
		    warn( "Child $child exited with value " . ($? >> 8) .
			  ", signal " . ($? & 255) . "\n" );
		}
		return 0;
	    }
	    next RZOT_FOR;
	}
	if( time >= $timedone ) {	    
	    return -1;
	}
    }
    return 1;
}

# Waits for child process to exit.
# Return values are ignored; method always succeeds.
# Usage: &wait_for_return( (<pid1>,..., <pidN>) );
sub wait_for_return {
    foreach my $child (@_) {
	waitpid $child, 0; #no flags
    }
    return 1;
}

# greps files for any word in @bad_words
# Returns 1 (true) if clean, 0 (false) if dirty.
#
# Usage: &is_clean( (<filename1>,..., <filenameN>) );
sub is_clean {
    my @bad_words = ("ERROR", "WARNING", "JAVA.*EXCEPTION",
		     "ASSERTIONVIOLATED", "ILLEGAL", "\"DROPPING IT\"");
    my ($file, $word);
    foreach $file (@_) {
	if( -r $file ) {
	    foreach $word (@bad_words) {
		if( `grep -ic $word $file` > 0 ) {
		    warn( "Found the bad word:\t\'$word\'\n" );
		    warn( `grep -i $word $file` );
		    return 0;
		}
	    }
	}
    }
    return 1;
}


# Creates and returns a node object.
# Rewrites a SandStorm configuration file, filling in variable fields.
# Automatically adds the new file to @configs.
#
# USAGE:
# &create_node( <tag>, <template>, <node id>, <tapestry mode>, <mortal?>,
#		<gateway id>, <number_static>, <tag-to-id map>,
#		<template variable map>,
#		<overrides>, <components to add>, <stages to remove>);
sub create_node {
    my( $tag, $cfg_template, $node_id, $tap_mode, $mortal,
	$gateway_id, $number_static, $id_map_ref, $value_map_ref,
	$overrides_ref,
	$add_list_ref, $remove_list_ref) = @_;

    ####################
    # Step 0: Initialize most of the node object.
    my %node;
    $node{"tag"} = $tag;

    $node{"node_id"} = $node_id;
    $node_id =~ /^(.*):(.*)$/;
    if( ! defined $2 ) {
	&die( "invalid node id: $node_id\n" );
    }
    my $hostname = $1;
    
    my $cfg_output = "$tmp_dir/experiment-$$-$tag.cfg";
    $node{"config"} = $cfg_output;

    $node{"mortal"} = $mortal;
    $node{"tap_mode"} = $tap_mode;

    # new filenames to remember:
    my @tmp_files;
    my @remote_tmp_files;
    my $remote_tmp = &remote_tmp_dir( $hostname );

    ####################
    # Step 1: assemble config template.

    # Optionally swap in Tapas stages.
    if( $use_tapas == 1 ) {
        push @{@$add_list_ref}, @tapas_stage_includes;
        push @{@$remove_list_ref}, @tapestry_stages;
    }
    
    # Add any last-minute components:
    if( defined $add_list_ref ) {
	my @add_list = @$add_list_ref;
	print "adding components to $tag: @add_list \n";	
	my $new_cfg_template = &add_components( $cfg_template, @add_list );
	push @configs, $new_cfg_template;
	$cfg_template = $new_cfg_template;
    }
    
    # If the configuration file links to other files, splice them in:
    if( ! -r $cfg_template ) {
	&die( "unreadable configuration template: $cfg_template\n" );
    }
    my $includes_others = `grep -c '^[^#]*include' $cfg_template`;
#    print "Tools::create_node: $cfg_template: $includes_others\n";
    if( $includes_others > 0 ) {
	my $new_cfg_template = &preprocess( $cfg_template );
	push @configs, $new_cfg_template;
	$cfg_template = $new_cfg_template;
    }
    # Now override specified parameters BEFORE interpolating variables.
    if( defined $overrides_ref ) {
	my %overrides = %{$overrides_ref};	
	foreach my $stage (keys %overrides) {
#	    print "overriding: $stage => " .
#		join( ' ', keys %{$overrides{$stage}} ) . "\n";
	    &insert( $cfg_template, $stage, %{$overrides{$stage}} );
	}
    }

    # Remove a few stages:
    if( defined $remove_list_ref ) {
	my @remove_list = @$remove_list_ref;
	print "removing stages from $tag: @remove_list \n";
	my $new_cfg_template = &remove_stages( $cfg_template, @remove_list );
	push @configs, $new_cfg_template;
	$cfg_template = $new_cfg_template;
    }

    ####################
    # Step 2: interpolate template variables.
    # Construct a map of variables that the config file can use.
    my %cfg_value_map;

    # NodeIDs
    $cfg_value_map{"NodeID"} = $node_id;
    $cfg_value_map{"GatewayID"} = $gateway_id;
    # Now slurp in the "*ID" variable for all nodes:
    my %ids_by_tags = %$id_map_ref;
    foreach my $nexttag ( keys %ids_by_tags ) {
	$cfg_value_map{ $nexttag . "ID" } = $ids_by_tags{$nexttag};
    }

    # Next add "*PublicKey" and "*PrivateKey" variables for all nodes:
    # Assume that identical keys exist on all remote machines
    my $remote_key_dir = &remote_run_dir( $hostname ) . "/keys";
    foreach my $nexttag ( keys %keys_by_tags ) {
	my $key_prefix = $keys_by_tags{$nexttag};
	next unless $key_prefix;	# skip undefined keys.
	
	# Substitue filename on remote machine.
	my $key_dir = $ostore_dirs{"key_dir"};
	$key_prefix =~ s/^$key_dir/$remote_key_dir/;

	# Now create the new key variables
	$cfg_value_map{ $nexttag . "PublicKey" } = "$key_prefix.pub";
	$cfg_value_map{ $nexttag . "PrivateKey" } = "$key_prefix.priv";
	if( $nexttag eq $tag ) {
	    # This tag is for this config file
	    $cfg_value_map{ "NodePublicKey" } = "$key_prefix.pub";
	    $cfg_value_map{ "NodePrivateKey" } = "$key_prefix.priv";
	    # Backwards compatibility (yuck!)
	    my $deprecated = `grep -c '^[^#]*\{NewPublicKey\}' $cfg_template`;
	    if( $deprecated > 0 ) {
		print STDERR "NewPublicKey is deprecated. Use NodePublicKey\n";
		$cfg_value_map{ "NewPublicKey" } = "$key_prefix.pub";
		$cfg_value_map{ "NewPrivateKey" } = "$key_prefix.priv";
	    }
	}
    }
#    print "variables: " . join( ' ', keys %cfg_value_map) . "\n";
    
    # Now add data-level crypto keys (IR Threshold keys, Principal keys)
    #  Unlike the tagged keys above, these keys are assigned statically.
    $cfg_value_map{"IRPublicKey"} = "$remote_key_dir/$keyshare_prefix.pub";
    my @ir_keyshares;
    for( 0 .. ($ring_size-1)) {
	push @ir_keyshares, "$remote_key_dir/$keyshare_prefix-$_.priv";
    }
    $cfg_value_map{"IRPrivateKeyShares"} = "@ir_keyshares";
    $cfg_value_map{"PrincipalPublicKey"} = "$remote_key_dir/principal.pub";
    $cfg_value_map{"PrincipalPrivateKey"} = "$remote_key_dir/principal.priv";
    
    # Now add a few well-known aliases to key lists.
    # PERL FU!
    $cfg_value_map{"MemberPublicKeys"} =
	join ' ', map "\$\{$_\}", sort
	    grep /^Ring(\d+)PublicKey/, keys %cfg_value_map;
    $cfg_value_map{"ReplicaPublicKeys"} =
	join ' ', map "\$\{$_\}", sort
	    grep /^Replica(\d+)PublicKey/, keys %cfg_value_map;
    $cfg_value_map{"StoragePublicKeys"} =
	join ' ', map "\$\{$_\}", sort
	    grep /^StorageServer(\d+)PublicKey/, keys %cfg_value_map;
#    print "MemberPublicKeys: " . $cfg_value_map{"MemberPublicKeys"}. "\n";
#    print "ReplicaPublicKeys: " . $cfg_value_map{"ReplicaPublicKeys"}. "\n";
#    print "StoragePublicKeys: " . $cfg_value_map{"StoragePublicKeys"}. "\n";

    # Miscellany
    if( $tag =~ /\D+(\d+)$/ ) {
	# Tag ends in a number.  Provide that number as a variable:
	$cfg_value_map{"Index"} = "$1";
    }
    
    my $needs_slow_link = `grep -c '^[^#]*SlowLink' $cfg_template`;
    if( $needs_slow_link > 0 ) {
	$cfg_value_map{"SlowLink"} = &random_slow_link;
    }
    $cfg_value_map{"RandomSeed"} = $seed_counter;
    # Compute a new seed.  Just incrementing seed is not enough.
    srand ($seed_counter + 7);
    $seed_counter = int rand 0xffffffff;	# random 32-bit int.
    
    $cfg_value_map{"RingSize"} = $ring_size;
    $cfg_value_map{"TapestryMode"} = $tap_mode;
    $cfg_value_map{"NumberStatic"} = $number_static;
    
    $cfg_value_map{"allPairsPingsFile"} = &remote_run_dir( $hostname ) . 
	"/x-pl.app";
    
    $cfg_value_map{"InternalLog"} =
	"$remote_tmp/experiment-$$-$tag-internal.log";
    $cfg_value_map{"SSProfileFile"} =
	"$remote_tmp/experiment-$$-$tag-profile.txt";
    $cfg_value_map{"SSGraphFile"} =
	"$remote_tmp/experiment-$$-$tag-graph.txt";
    $cfg_value_map{"CacheDir"} =
	"$remote_tmp/experiment-$$-$tag-blocks/";

    # dump ostore_dirs into variable mappings
    foreach my $dir_name (keys(%ostore_dirs))
    {
        $cfg_value_map{$dir_name} = $ostore_dirs{$dir_name};
    }
    
    push @dirs, $cfg_value_map{"CacheDir"};
    # remember to copy some of those back later.
    # Skip CacheDir - just create/delete it remotely.
    push @remote_tmp_files, $cfg_value_map{"InternalLog"} . "*";
    push @remote_tmp_files, $cfg_value_map{"SSGraphFile"};
    push @remote_tmp_files, $cfg_value_map{"SSProfileFile"};
#    push @remote_tmp_files, "$remote_tmp/experiment-TPS*";
    
    # uncomment the following lines to automatically delete them, too.
#      my $local_copy;
#      ($local_copy = $cfg_value_map{"InternalLog"}) =~
#    		s/^$remote_tmp/$tmp_dir/;
#      push @profiles, $local_copy;
#      ($local_copy = $cfg_value_map{"SSGraphFile"}) =~
#    		s/^$remote_tmp/$tmp_dir/;
#      push @profiles, $local_copy;
#      ($local_copy = $cfg_value_map{"SSProfileFile"}) =~
#    		s/^$remote_tmp/$tmp_dir/;
#      push @profiles, $local_copy;

    # Fill in all those variables
    push @configs, $cfg_output;
    my %caller_specified_map = %{$value_map_ref};
    foreach my $var (keys %caller_specified_map) {
	# interpolate these variables also. They take precendence over
	# other mappings in %cfg_value_map.
	$cfg_value_map{$var} = $caller_specified_map{$var};
    }
    &interpolate( $cfg_template, $cfg_output, \%cfg_value_map );

    ####################
    # Step 3: finish initializing node object.
    
    $node{"tmp_files"} = \@tmp_files;
    # add on the hostname
    @remote_tmp_files = map { "$hostname:$_" } @remote_tmp_files;
    $node{"remote_tmp_files"} = \@remote_tmp_files;
    return %node;
}

# Assigns a public/private keypair to each virtual node.
# Makes new variables "${<Tag>PublicKey}" and "${<Tag>PrivateKey}" available
#  to config files for all tags in list.
# Makes new variables "${NodePublicKey}" and "${NodePrivateKey}" available
#  to each config file its own keys.
# Only call this method once per experiment.
# Generally tries to assign keys in the "expected" order:
#  RP, then Ring*, then Client, then others.
#
# USAGE: &assign_keys( <tag list> )
sub assign_keys {
    my @tag_list = @_;

    # First build the list of available keys.  
    my @available_keys;
    push @available_keys, glob( $ostore_dirs{"key_dir"}."/fed*.pub" );
    push @available_keys, glob( $ostore_dirs{"key_dir"}."/rp*.pub" );
    push @available_keys, glob( $ostore_dirs{"key_dir"}."/pr-*.pub" );
    push @available_keys, glob( $ostore_dirs{"key_dir"}."/client*.pub" );
    push @available_keys, glob( $ostore_dirs{"key_dir"}."/replica*.pub" );
    push @available_keys, glob( $ostore_dirs{"key_dir"}."/storage*.pub" );
#    print "assign_keys: available keys: @available_keys\n";

    # Store just the prefix.
    @available_keys = map { m/(.*)\.pub$/; $1 } @available_keys;
#    print "assign_keys: available keys: @available_keys\n";

    # Now just map each tag to next key.
    my @no_keys;
    foreach my $tag (@tag_list) {
	if( ! @available_keys ) {
	    push @no_keys, $tag;
	} else {
	    my $key_prefix = shift @available_keys;
	    $keys_by_tags{$tag} = $key_prefix;
#	print "assign_keys: $tag => $key_prefix\n";
	}
    }
    if( @no_keys ) {
	print "No keys for: @no_keys\n";
	print "Hopefully their Routers are using fake_keys\n";
    }
}

# Splices together config files that link to component files using the
#  "<!include>" directive
#
# USAGE: <new_filename> = &preprocess( <template> );
sub preprocess {
    my $cfg = shift;

    # Strip out base filename, add proper ".tmp" suffix:
#    print "Tools::preprocess: $cfg\n";
    $cfg =~ m%/?(experiment-\d+-)*([^/]+?)(\.tmp)?$%;
#    print "Tools::preprocess: \"$1\" $2 $3\n";    
    my $new_cfg = "$tmp_dir/experiment-$$-$2.tmp";
    if( $3 ) {
	$new_cfg .= ".1";
    }
#    print "Tools::preprocess: $new_cfg\n";

    my @lines = &read_file( $cfg );

    open( OUT, ">$new_cfg" ) or &die( "can't open temp file $new_cfg" );
    print OUT @lines;
    close( OUT ) or &die( "can't close temp file $new_cfg" );

    return $new_cfg;
}

# Recursively reads configuration files with "<!include *>" directives.
#
# USAGE: <line_list> = &read_file( <filename> );
sub read_file {
    my( $in ) = @_;

    open( IN, "<$in" ) or &die( "can't open $in" );
    my @lines = <IN>;
    close( IN ) or &die( "can't close $in" );
    my @output;
    
    my $line_no = 0;
    foreach my $line (@lines) {
	$line_no++;
	if( $line =~ /^\s*<!include\s+([\w\d.-]+)>\s*/ ) {
	    # insert the included file content.
	    my $included = $1;
#	    print "Found include: $included\n";

	    if( $included !~ m%^/% ) {	# relative pathname
		my $dir = $in;
		$dir =~ s%/(./)?[^/]*$%%;	# find directory of $in
		if( -r "$dir/$included" ) {
		    $included = "$dir/$included";
		} elsif( -r $ostore_dirs{"cfg_dir"} . "/$included" ) {
		    $included = $ostore_dirs{"cfg_dir"} . "/$included";
		} elsif( -r $ostore_dirs{"run_dir"} . "/cfg/$included" ) {
		    $included = $ostore_dirs{"run_dir"} . "/cfg/$included";
		} else {
		    &die( "can't find included file: $included\n" );
		}
	    }
#	    print "Found include: $included\n";	    
	    push @output, &read_file( $included );

	} else {
	    # Just copy the line
	    push @output, $line;
	}
    }
    return @output;
}

# Creates and returns a host object.
#  All useful information from node objects is copied into the host object.
#  Creates and stores a unified configuration file for a remote host.
#  Adds artificial topology information to a configuration file.
#
# USAGE: &add_topology( <topology file>, <topology mapping>,
#			<remote host name>, <node object refs> );
# To run without an artificial topology, set <topology file> to "NONE".

sub create_host {
    my( $topology_file, $mapping_file, $hostname, @nodes ) = @_;
    my %host;
    $host{"name"} = $hostname;
    print "$hostname:";

    my $remote_tmp = &remote_tmp_dir( $hostname );
    
    # Collect files from each node
    my @tmp_files;
    my @remote_tmp_files;
    my @node_cfgs;
    foreach (@nodes) {
	my %node = %{$_};
	print " ", $node{"tag"};
	push @node_cfgs, $node{"config"};
	if( $node{"mortal"} eq "mortal" ) {
	    $host{"mortal"} = "mortal";
	}
	push @tmp_files, @{$node{"tmp_files"}};	# Warning - Perl-Fu
	push @remote_tmp_files, @{$node{"remote_tmp_files"}};
	# take highest mode: fed > static > dynamic 
	my $tap_mode = $node{"tap_mode"};
	my $host_tap_mode = $host{"tap_mode"};
	if( (!defined $host_tap_mode) || ($tap_mode eq "fed") ||
	    (($tap_mode eq "static") && ($host_tap_mode ne "fed")) ) {
	    $host{"tap_mode"} = $tap_mode;
	}
    }
    print ".\n";	# end line started above.

    # Create one large config file for entire jvm
    my $cfg_num = 0;
    my $unified_cfg = "$tmp_dir/experiment-$$-$hostname-unified.cfg";
    while( -f $unified_cfg ) {
	$cfg_num++;
	$unified_cfg = "$tmp_dir/experiment-$$-$hostname-$cfg_num-unified.cfg";
    }
    push @configs, $unified_cfg;
    system( "@concat_cmd @node_cfgs > $unified_cfg" );
    $host{"config"} = $unified_cfg;

    # Now add the artificial topology if necessary.
    if( $topology_file ne "NONE" ) {
	my %topo_values;
	my $remote_topology_file;
#	if( $topology_file =~ /^$ostore_dirs{"graph_dir"}\/([\w.-]+)$/ ) {
#	    # topology file is already on remote host
#	    $remote_topology_file =
#		&remote_run_dir( $hostname ) . "/graphs/$1";
#	} else {
	    # copy file over to remote /tmp.
	    $topology_file =~ /([\w.-]+)$/;
	    $remote_topology_file = "$remote_tmp/$1";
	    push @tmp_files, $topology_file;
#	}
	# copy over mapping file.
	push @tmp_files, $mapping_file;
	my $remote_mapping_file = $mapping_file;
	$remote_mapping_file =~ s/^$tmp_dir/$remote_tmp/;
#	print "remote map: $remote_mapping_file\n";
	
	$topo_values{"topology_graph"} = $remote_topology_file;
	$topo_values{"topology_mapping"} = $remote_mapping_file;
	&insert( $unified_cfg, "global", %topo_values );
    }
    # finish initializing the host.
    $host{"mortal"} ||= "daemon";
    $host{"tmp_files"} = \@tmp_files;
    $host{"remote_tmp_files"} = \@remote_tmp_files;
    return %host;
}

# Starts SandStorm on the local machine.
# USAGE: &start_local( <host ref> );
sub start_local {
    my $host_ref = shift;
    my $hostname = $$host_ref{"name"};

    if( ! &is_localhost( $hostname ) ) {
	&die( "start_local: Not localhost: $hostname\n" );
    }
    
    print "Starting local SandStorm...";
    my $log_num = 0;
    my $log = "$tmp_dir/experiment-$$-localhost.log";    
    while( -f $log ) {
	$log_num++;
	$log = "$tmp_dir/experiment-$$-localhost-$log_num.log";    
    }
    $$host_ref{"local_pid"} = &fork_sandstorm( $$host_ref{"config"}, $log );
    
    print "  started.\n";
    return $log;
}

# Copies configuration files to a remote host asynchronously.
# Returns a list containing the log and PID of the forked thread.
#
# USAGE: &async_copy_to_remote( <host>, <status_string> );
sub async_copy_to_remote {
    my $host_ref = shift;
    my $status_string = shift;
    
    my $hostname = $$host_ref{"name"};
    my $remote_tmp = &remote_tmp_dir( $hostname );
    my @ssh_args = &ssh_args( $hostname );
    
    push @running_hosts, $host_ref;

    print $status_string if $status_string;
    print "Copying files to $hostname.\n";
    my $log_num = 0;
    my $log = "$tmp_dir/scp_setup.$hostname.log";
    while( -f $log ) {
	$log_num++;
	$log = "$tmp_dir/scp_setup.$hostname-$log_num.log";	
    }
    my $child_pid =
	&fork_process( $log, @scp_cmd, @ssh_args, $$host_ref{"config"}, 
		       @{$$host_ref{"tmp_files"}}, "$hostname:$remote_tmp" );
    return ($log, $child_pid);
}

# Copies configuration files to a remote host synchronously.
# Returns true (1) iff the test succeeds.
#
# USAGE: &copy_to_remote( <host> );
sub copy_to_remote {
    my $host_ref = shift;
    
    my $hostname = $$host_ref{"name"};
    my $remote_tmp = &remote_tmp_dir( $hostname );
    my @ssh_args = &ssh_args( $hostname );

    my $ret_val = system( @scp_cmd, @ssh_args, $$host_ref{"config"}, 
			  @{$$host_ref{"tmp_files"}}, "$hostname:$remote_tmp" );
    return (! $ret_val );
}

# Starts SandStorm on a remote machine asynchronously.
# Returns a list containing the log and PID of the forked thread.
#
# USAGE: &async_start_remote( <host>, <status_string> );
sub async_start_remote {
    my $host_ref = shift;
    my $status_string = shift;
    
    my $hostname = $$host_ref{"name"};
    my $remote_tmp = &remote_tmp_dir( $hostname );
    my @ssh_args = &ssh_args( $hostname );

    print $status_string if $status_string;    
    print "Starting SandStorm on $hostname.\n";
    
    my $remote_unified_cfg = $$host_ref{"config"};
    $remote_unified_cfg =~ s/^$tmp_dir/$remote_tmp/;
    my @remote_sandstorm_cmd =
	((&remote_run_dir( $hostname ) . "/run-java"), $sandstorm_main);
    if( $disable_jit ) {
	splice @remote_sandstorm_cmd, 1, 0, "-nojit";
    }
    my $log_num = 0;
    my $log = "$tmp_dir/experiment-$$-$hostname.log";
    while( -f $log ) {
	$log_num++;
	$log = "$tmp_dir/experiment-$$-$hostname-$log_num.log";
    }
    my $child_pid;
    if( $use_nohup ) {
	# change the log to use the remote guy's temp dir
	my $local_log = $log;
	$log =~ s/^$tmp_dir/$remote_tmp/;
	# for some reason, passing in the arguments separately doesn't
	# work.  So here we pass them in as a big string.
	$child_pid = &fork_process( $local_log, "$ssh_cmd @ssh_args " .
				    "$hostname \"$nohup_cmd " . 
				    "@remote_sandstorm_cmd " . 
				    "$remote_unified_cfg >& $log &\"" );    
    } else {
	$child_pid = &fork_process( $log, $ssh_cmd, @ssh_args,
				    $hostname, @remote_sandstorm_cmd,
				    $remote_unified_cfg );    
    }
    $$host_ref{"local_pid"} = $child_pid;
    return ($log, $child_pid);
}    

# Calls a remote subrouting in Tools.pm on a remote machine.  Returns the
# output a an array of lines.
#
# USAGE: &call_remote_sub( <hostname>, <subroutine_name>, [@params] )
sub call_remote_sub {

    my $hostname = shift;
    my $sub_name = shift;
    my @params = @_;
    my @ssh_args = &ssh_args( $hostname );
    my $remote_run = &remote_run_dir( $hostname );
    my $tmp_log = "$tmp_dir/$sub_name.$$.tmp";
    # for whatever reason it doesn't work to pass these in as separate
    # parameters, so we'll pass it in as a big string
    my $rv = system( "$ssh_cmd @ssh_args $hostname \"@perl_exp_cmd " . 
	    "\'use lib \\\"$remote_run\\\"; use Tools; use strict; " . 
	    "print &$sub_name( @params );\'\" >& $tmp_log") 
	and die( "Couldn't execute ssh in call_remote_sub" );
    open( TMP, "<$tmp_log" ) or die( "Couldn't open $tmp_log" );
    my @output = <TMP>;
    close( TMP );
    unlink( $tmp_log );
    return @output;
}

# Parses a logfile from a remote SandStorm (presumably started using
# &async_start_remote() ) and returns the PID of the remote process.
#
# USAGE: &extract_remote_pid( <log> )
sub extract_remote_pid {
    my $log = shift;
    &wait_for( $log, "PID" );
    my $pid_string = `grep PID $log`;
    $pid_string =~ /PID:\s*(\S*)/;
    $pid_string = $1;
    return $pid_string;
}

# Call &extract_remote_pid on a remote machine, and return only the pid.
#
# USAGE: &extract_remote_pid( <log>, <hostname> )
sub remote_extract_remote_pid {
    my $log = shift;
    my $hostname = shift;
    my @output = call_remote_sub( $hostname, "extract_remote_pid", 
				  "\\\"$log\\\"" );
    # the PID should be the last line, right?
    return $output[$#output];
}

# Like &remote_extract_remote_pid, but doesn't use &call_remote_sub.
# Also, allows remote log to contain multiple "PID: " lines.
# Returns a list containing all remote PIDs.
#
# USAGE: &remote_extract_all_pids( <log>, <hostname> )
sub remote_extract_all_pids {
    my ($remote_log, $hostname) = @_;
    my @ssh_args = &ssh_args( $hostname );
    my $remote_test = "perl -w -e \' my \@pids; " .
	"until\( \@pids = \\`grep -a --no-messages PID $remote_log\\` \) " .
	    "{ sleep 1; } print \"\@pids\";\'";
    my @pid_strings;
    if( &is_localhost( $hostname ) ) {
	push @pid_strings, &extract_remote_pid( $remote_log );
    } else {
	@pid_strings = `$ssh_cmd @ssh_args $hostname \"$remote_test\"`;
    }
    my @pids;
    foreach (@pid_strings) {
	/PID:\s*(\S*)/;
	push @pids, $1;
    }
    return @pids;
}

# Copies configuration files to a remote host, then invokes SandStorm
#  on that host.
# Adds PID of local ssh client, remote sandstorm to appropriate lists.
#
# USAGE: &start_remote( <host> );
sub start_remote {
    my $host_ref = shift;
    my $hostname = $$host_ref{"name"};

    if( &is_localhost( $hostname ) ) {
	# Copying files to/from local host breaks scp.  Just use
	#  &start_local instead - it's faster anyway.
	return &start_local( $host_ref );
    }

    # Block on the asynchronous copy.  :/
    my ($log, $child_pid) = &async_copy_to_remote( $host_ref );
    if( &returned_zero( $child_pid ) ) {
	unlink $log if( -f $log );
    } else {
	&die( "Can't copy files to $hostname.\n" );
    }

    # Start the remote helper
    ($log, $child_pid) = &async_start_remote( $host_ref );

    my $remote_pid;
    if( $use_nohup  ) {
	$remote_pid = &remote_extract_remote_pid( $log, $hostname );
    } else {
	$remote_pid = &extract_remote_pid( $log );
    }
    $$host_ref{"remote_pid"} = $remote_pid;
    
    print "$hostname running.  Remote PID: $remote_pid\n";
    return $log;
}

# A simpler version of &start_remote; &fork_remote starts any program that
# prints out its PID but does no setup and knows nothing about SandStorm.
# Returns the host object created to track the remote process.
#
# WARNING: This method requires the remote process to print out its PID in
#  the format expected by &extract_remote_pid().
#
# USAGE: &fork_remote( <log>, <hostname>, <command>, [<arg1> ... <argN>] );
sub fork_remote {
    my( $log, $hostname, @rest ) = @_;

    # Fake out enough of the remote host pseudo-classes
    #   so that &exit_all() cleans up properly. 
    my %host_data;
    my @empty = ();
    $host_data{"name"} = $hostname;
    $host_data{"mortal"} = "daemon";
    $host_data{"tmp_files"} = \@empty;
    @empty = ();    
    $host_data{"remote_tmp_files"} = \@empty;

    # Make sure the remote process on this machine is killed later.
    push @running_hosts, \%host_data;

    my @ssh_args = &ssh_args( $hostname );
    my $remote_tmp = &remote_tmp_dir( $hostname );
    # Now start up the remote process and read its PID.
    if( $use_nohup ) {
	# Do not stay connected.  Reading remote PID is tricky.
	my $remote_log = $log;
	$remote_log =~ s/^$tmp_dir/$remote_tmp/;
	$remote_log .= ".remote";
	$host_data{"local_pid"} =
	    &fork_process( $log, $ssh_cmd, @ssh_args, $hostname,
			   "nohup @rest >& $remote_log &" );
	push @{$host_data{"remote_tmp_files"}}, "$hostname:$remote_log";

	unless( $ignore_remote_pids ) {
	    my @pids = &remote_extract_all_pids( $remote_log, $hostname );
	    $host_data{"remote_pid"} = $pids[0];
	}
	
    } else {
	$host_data{"local_pid"} =  &fork_process( $log, $ssh_cmd, @ssh_args,
						  $hostname, @rest );
	$host_data{"remote_pid"} = &extract_remote_pid( $log );
    }
	
    return \%host_data;
}

# Starts several remote SandStorm processes in parallel.
# Returns a list containing the log filenames of each process.
#
# USAGE: &start_multiple_remote( (<host1> ... <hostn>) );
sub start_multiple_remote {
    my @hosts = @_;
    my @sandstorm_logs;	# The logs to return.
    my @setup_logs;
    my @setup_pids;

    # First copy over the required files (asynchronously).
    my $total = scalar( @hosts );
    for( my $i=0; $i < $total; $i++ ) {
	my $status = ("[" . ($i+1) . "/$total] ");
	my $host_ref = $hosts[$i];
	my $hostname = $$host_ref{"name"};    

	if( &is_localhost( $hostname ) ) {
	    print $status;
	    push @sandstorm_logs, &start_local( $host_ref );
	} else {
	    my ($log, $child_pid) = &async_copy_to_remote( $host_ref, $status );
	    push @setup_logs, $log;
	    push @setup_pids, $child_pid;
	}

	# check for self-DoS:
	if( $num_parallel_calls <= @setup_pids ) {
	    # too many in flight.  Wait for one to exit.
	    my $child_pid = shift @setup_pids;
	    my $child_log = shift @setup_logs;

	    if( ! &returned_zero( "--quiet", $child_pid )
		or ! &is_clean( $child_log ) ) {
		&die( "Can't copy config files.  See log $child_log\n" );
	    } else {
		unlink $child_log if( -f $child_log);
	    }
	}
    }
    # Barrier: wait for all copies to finish.
    for( my $i=0; $i <= $#setup_pids; $i++ ) {
	my $child_pid = $setup_pids[$i];
	my $child_log = $setup_logs[$i];
	if( ! &returned_zero( "--quiet", $child_pid )
	    or ! &is_clean( $child_log ) ) {
	    &die( "Can't copy config files.  See log $child_log\n" );
	} else {
	    unlink $child_log if( -f $child_log);
	}
    }

    my %logs_to_hosts;
    my @ordered_logs;	# For more-readable output
    my @log_window;	# For rate-limiting
    # Now start the remote processes (asynchronously).
    for( my $i=0; $i < $total; $i++ ) {
	my $status = ("[" . ($i+1) . "/$total] ");
	my $host_ref = $hosts[$i];
	my $hostname = $$host_ref{"name"};    
	
	if( &is_localhost( $hostname ) ) {
	    print $status . "Localhost already started\n";
	    push @ordered_logs, "localhost";
	    # Nothing - localhost is handled monolithically above.
	} else {
	    my ($log, $child_pid) = &async_start_remote( $host_ref, $status );
	    $logs_to_hosts{$log} = $host_ref;
	    push @ordered_logs, $log;
	    push @log_window, $log;
	    push @sandstorm_logs, $log;
	}

	# check for self-DoS:
	if( $num_parallel_calls <= @log_window ) {
	    # Wait for one to finish starting up.
	    my $child_log = shift @log_window;
	    my $remote_pid;
	    if( $use_nohup  ) {
		$remote_pid = &remote_extract_remote_pid($child_log, 
							 $hostname );
	    } else {
		$remote_pid = &extract_remote_pid( $child_log );
	    }
	}
    }
    # Finally, store the remote PIDs.
    for( my $i=0; $i < $total; $i++ ) {
	my $status = ("[" . ($i+1) . "/$total] ");
	my $log = $ordered_logs[$i];
	if( $log eq "localhost" ) {
	    print $status . "Localhost already started\n";
	} else {
	    my $host_ref = $logs_to_hosts{$log};
	    my $hostname = $$host_ref{"name"};
	    my $remote_pid;
	    if( $use_nohup  ) {
		$remote_pid = &remote_extract_remote_pid( $log, 
							  $hostname );
	    } else {
		$remote_pid = &extract_remote_pid( $log );
	    }
	    $$host_ref{"remote_pid"} = $remote_pid;
	    
	    print $status . "$hostname running.  Remote PID: $remote_pid\n";
	}
    }

    return @sandstorm_logs;
}

# Chooses a line from "./slow_links.txt" pseudorandomly.
# Each line should be a valid <source>:<sink>:<time> tuple.
# Set the environment variable "OSTORE_SEED" to choose a different tuple.
sub random_slow_link {
    my $link_file = $ostore_dirs{"run_dir"} . "/slow_links.txt";
    open LINKS, "<$link_file" or &die( "Can't open $link_file\n" );
    my @links = <LINKS>;
    close LINKS;
    srand $seed_counter;
    my $choice = int rand scalar( @links );
    my $link = $links[$choice];
    chomp $link;
    return $link;
}

# Returns true (1) if the named machine is on one of the local
# OceanStore clusters.  Otherwise returns false (0).
#
# Usage: &in_local_cluster( <hostname> );
sub in_local_cluster {
    my $remote_host = shift;
    my $answer = 0;	# false

    if( ($remote_host =~ /^x\d/) or
	($remote_host =~ /^169\.229\.50\.2../) ) {
	$answer = 1;	# true - X cluster
    } elsif( ($remote_host =~ /^ibm\d/) or
	     ($remote_host =~ /^169\.229\.59\.21./) ) {
	$answer = 1;	# true - IBM cluster
    }
    return $answer;
}

# Choose arguments to ssh and scp based on remote hostname.
#
# usage: my @args = &ssh_args( <hostname> );
sub ssh_args {
    my $remote_host = shift;
    my @args;

    if( defined( $ENV{"OSTORE_SMART_TOOLS"} )
	and &in_local_cluster( $remote_host ) ) {
	# Probably an experiment with nodes in and out of the cluster.
	# Ignore REMOTE_OSTORE_USER.
	@args = ();
    } elsif( $ENV{"REMOTE_OSTORE_USER"} ) {
	@args = ("-o", "User=".$ENV{"REMOTE_OSTORE_USER"} );
    } else {
	# nothing special
	@args = ();
    }
    return @args;
}


# Find "run/" on the remote machines.  Fall back on "REMOTE_OSTOREHOME".
#
# usage:  &remote_run_dir( <hostname> );
sub remote_run_dir {
    my $remote_host = shift;
    my $remote_run_dir;

    if( &is_localhost( $remote_host ) ) {
	$remote_run_dir = $ostore_dirs{"run_dir"};
    } elsif( defined( $ENV{"OSTORE_SMART_TOOLS"} )
	     and &in_local_cluster( $remote_host ) ) {
	# Probably an experiment with nodes in and out of the cluster.
	# Use the package in /stripe.
	$remote_run_dir = "/stripe/oceanstore/run";	
    } elsif( $ENV{"REMOTE_OSTORE_RUN_DIR"} ) {
	$remote_run_dir = $ENV{"REMOTE_OSTORE_RUN_DIR"};
    } elsif( $ENV{"REMOTE_OSTOREHOME"} ) {
	warn( "REMOTE_OSTOREHOME is deprecated.  " .
	      "Use REMOTE_OSTORE_RUN_DIR instead.\n" );
	$remote_run_dir = $ENV{"REMOTE_OSTOREHOME"} . "/pond/run";
    } else {
	# Assume same as local.
	$remote_run_dir = $ostore_dirs{"run_dir"};
    }
    return $remote_run_dir;
}


# Kills all local and reomote processes, removes all files
# If the argument (error code) is nonzero, print the logs to STDOUT.
sub exit_all {
    my $error_code = shift;

    # First handle output
    if( $error_code == 0 ) {
	print "*** Success! ***\n";	
    } else {
	print "*** Failure! Logs left in $tmp_dir ***\n";
    }

    # Now clean up any remaining remote sites.
    my $hostname;
    my @cleanup_pids;
    my @cleanup_logs;
    my %combined_hosts;
    foreach (@running_hosts) {
	my %host = %{$_};
	my $hostname = $host{"name"};
#	print "checking $hostname\n";
	my $other_ref = $combined_hosts{$hostname};
	if( defined $other_ref ) {
#	    print "combining\n";
	    # combine hosts.  Only certain fields are important
	    if( ! $$other_ref{"mortal"} eq $host{"mortal"} ) {
		# one must be non-mortal
		$$other_ref{"mortal"} = "daemon";
	    }
	    if( defined $host{"remote_pid"} ) {
		if( defined $$other_ref{"remote_pid"} ) {
		    $$other_ref{"remote_pid"} = ($$other_ref{"remote_pid"} . " "
						 . $host{"remote_pid"});
		} else {
		    $$other_ref{"remote_pid"} = $host{"remote_pid"};
		}
	    }
	    push @{$$other_ref{"remote_tmp_files"}},
			@{$host{"remote_tmp_files"}};
	    
	} else {
	    # save
	    $combined_hosts{$hostname} = \%host;
	}
#  	print "mortal => " . $combined_hosts{$hostname}{"mortal"} . "\n";
#  	print "remote_pid => " .
#  	    $combined_hosts{$hostname}{"remote_pid"} . "\n";
#  	print "remote_tmp_files => " .
#  	    (join ' ', @{$combined_hosts{$hostname}{"remote_tmp_files"}}) .
#  		"\n";
    }
    my @sorted_running_hosts =
	sort { $$a{"name"} cmp $$b{"name"} } (values %combined_hosts);
    my $total = scalar( @sorted_running_hosts );
    # Kill the remote clients:
    for( my $i=0; $i < $total; $i++ ) {
	my %host = %{$sorted_running_hosts[$i]};
	$hostname = $host{"name"};
	if( &is_localhost( $hostname ) ){
	    print( "WARNING: Looks like localhost: $hostname\n" );
 	}
	my @ssh_args = &ssh_args( $hostname );
	
	# kill remote SandStorm.
	if( ($error_code == 0) and ($host{"mortal"} eq "mortal") ) {
	    # must already have exited cleanly.  Skip the remote kill,
	    #  which would only produce a meaningless error.
	    print "[" . ($i+1) . "/$total] Skipping $hostname\n";
	} else {
	    if( defined $host{"remote_pid"} ) {
		my $log_num = 0;
		my $log = "$tmp_dir/ssh_kill.$hostname.log";
		while( -f $log ) {
		    $log_num++;
		    $log = "$tmp_dir/ssh_kill.$hostname-$log_num.log";	
		}
		print("[" . ($i+1) . "/$total] Killing " . $host{"remote_pid"}
		      . " on $hostname\n");
		
		push @cleanup_pids,
		&fork_process( $log, $ssh_cmd, @ssh_args, $hostname,
			       @remote_kill_cmd, $host{"remote_pid"} );
		push @cleanup_logs, $log;

		# check for self-DoS:
		if( $num_parallel_calls <= @cleanup_pids ) {
		    # too many in flight.  Wait for one to exit.
		    &wait_for_return( shift @cleanup_pids );
		}
	    }
	}
    }
    
    # Barrier: Wait for all ssh processes to exit.
    foreach( @cleanup_pids ) { &wait_for_return( $_ ); }
#    print "kill done";
    @cleanup_pids = ();

    if( (defined $disable_copyback) and
	(($disable_copyback > 0) or ($error_code != 0)) ) {
	print "Copyback disabled\n";
    } else {
	print "Copying logs\n" if( $total > 0 );
	# Copy back over interesting output (main log is already here)
	for( my $i=0; $i < $total; $i++ ) {
	    my %host = %{$sorted_running_hosts[$i]};
	    $hostname = $host{"name"};
	    if( &is_localhost( $hostname ) ){
		print "[" . ($i+1) . "/$total] Skipping $hostname. " .
		    " Looks like localhost\n";
		next;
	    }
	    my @ssh_args = &ssh_args( $hostname );	
	    my $log_num = 0;
	    my $log = "$tmp_dir/scp_cleanup.$hostname.log";
	    while( -f $log ) {
		$log_num++;
		$log = "$tmp_dir/scp_cleanup.$hostname-$log_num.log";	
	    }
	    my @remote_tmp_files = @{$host{"remote_tmp_files"}};
	    if( @remote_tmp_files ) {
		print "[" . ($i+1) . "/$total] Copying logs from $hostname\n";
		push @cleanup_pids, &fork_process( $log, @scp_cmd, @ssh_args,
						   @remote_tmp_files, $tmp_dir );
		push @cleanup_logs, $log;

		# check for self-DoS:
		if( $num_parallel_calls <= @cleanup_pids ) {
		    &wait_for_return( shift @cleanup_pids );
		}
	    } else {
		print "[" . ($i+1) . "/$total] Skipping $hostname\n";
	    }
	}
	# Barrier: Wait for all ssh processes to exit.
	foreach( @cleanup_pids ) { &wait_for_return( $_ ); }
#    print "Copy done";
	@cleanup_pids = ();
    }
    
    # Remove all other output.
    print "Cleaning remote /tmp\n" if( $total > 0 );
    for( my $i=0; $i < $total; $i++ ) {
	my %host = %{$sorted_running_hosts[$i]};
	$hostname = $host{"name"};
	if( &is_localhost( $hostname ) ){
	    print "[" . ($i+1) . "/$total] Skipping $hostname. " .
		" Looks like localhost\n";
	    next;
	}
	my @ssh_args = &ssh_args( $hostname );	
	my $remote_tmp = &remote_tmp_dir( $hostname );
	my $log_num = 0;
	my $log = "$tmp_dir/ssh_rm.$hostname.log";
	while( -f $log ) {
	    $log_num++;
	    $log = "$tmp_dir/ssh_rm.$hostname-$log_num.log";	
	}
	print "[" . ($i+1) . "/$total] Cleaning $hostname\n";	

	# Strip of machine name:
	my @remote_tmp_files = @{$host{"remote_tmp_files"}};
	map { s/.*:(.*)/$1/ } @remote_tmp_files;
	
	push @cleanup_pids,
	&fork_process( $log, $ssh_cmd, @ssh_args, $hostname,
		       @remote_rm_cmd, "$remote_tmp/experiment-$$*",
		       @remote_tmp_files );
	push @cleanup_logs, $log;	
	# check for self-DoS:
	if( $num_parallel_calls <= @cleanup_pids ) {
	    &wait_for_return( shift @cleanup_pids );
	}
    }
    # Barrier: Wait for all ssh processes to exit.
    foreach( @cleanup_pids ) { &wait_for_return( $_ ); }
#    print "Copy done";
    foreach my $file (@cleanup_logs) {
	unlink $file if( -f $file );
    }

    # clean up all other state (local ssh processes, temp files, etc.)
    &clean_up( $error_code );

    print "*** Goodbye ***\n";
    exit $error_code;
}

# Kills child processes
# If exit code is zero, removes logs and config files.
#
# Usage: &clean_up( <exit_code> )
sub clean_up {
    my $error_code = shift;
    my ($gonner, $file, $dir);
    
    foreach $gonner (@child_pids) {
	kill 9, $gonner;
    }

    if( $disable_cleanup ) {
	print "Local cleanup disabled\n";
    } elsif( $error_code == 0 ) {
	foreach $file (@logs, @configs, @profiles) {
	    if( -f $file ) {
		unlink $file;
	    }
	}
    }

    foreach $dir (@dirs) {
	if( -d $dir ) {
	    rmtree( $dir );
	}
    }
}

# Exits cleanly
sub die {
    warn "@_\n";
    &exit_all( 1 );
}

# Exits cleanly
sub catch_sig {
    # Disable signals to avoid duplicates, then die.
    $SIG{INT} = \&ignore_sig;
    $SIG{QUIT} = \&ignore_sig;
    $SIG{HUP} = \&ignore_sig;
    $SIG{PIPE} = \&ignore_sig;
    $SIG{TERM} = \&ignore_sig;
    &die(  "*** Caught SIG@_ ***\n" );
}

# Ignores signals
sub ignore_sig {
    warn( "Dying already.  Please wait.\n" );
}

# Exits cleanly (no error)
sub user_exit {
    warn "\nExiting...\n";
    &exit_all( 0 );
}

####################
# Set up default signal handlers
$SIG{INT} = \&catch_sig;
$SIG{QUIT} = \&catch_sig;
$SIG{HUP} = \&catch_sig;
$SIG{PIPE} = \&catch_sig;
$SIG{TERM} = \&catch_sig;


############################################################
# PSEUDO-CLASSES (really just hashtables)

# OceanStore Nodes - all the information for a single virtual machine
#   fields:
#
# tag - the unique name of this node.  Special names:
#		"Federation", "RP", "RingX", "StorageServerX" (X is an index).
# node_id - the hostname:port of the machine on which to run this node
# config - the SandStorm config file
# mortal - 0 for daemon nodes, 1 for mortal (usually client) nodes
# tap_mode - tapestry type: "fed", "static", or "dynamic"
# tmp_files - an array reference storing names of temp files to copy TO the
#		remote host. 
# remote_tmp_files - an array reference storing names of temp files to copy
#		FROM the remote host.


# Remote Hosts - all the information for a single remote host
#   fields:
#
# name - the hostname
# config - the unified SandStorm config file.
# tap_mode - tapestry type: "fed", "static", or "dynamic"
# mortal - 0 if hosting only daemon nodes, 1 if any mortal.
# tmp_files - an array reference storing names of temp files to copy TO the
#		remote host. 
# remote_tmp_files - an array reference storing names of temp files to copy
#		FROM the remote host.
# local_pid - the PID for the local ssh connection or sandstorm.
# remote_pid - the remote PID for SandStorm on the host.

1;	# All modules should return 1.
