#!/usr/bin/perl -w
#
# author: Jeremy Stribling
# $Id: all_pairs_pings_times.pl,v 1.11 2003/04/28 18:28:57 strib Exp $
#
# calculates the all-pairs ping times for the given hosts
# usage: all_pairs_ping_times.pl <host1> ... <hostN>
# if no hosts are specified, defaults to $OSTORE_SERVERS

BEGIN {
    # Find the directory with Tools.pm.  Should this be easier?
    my $run_dir = "$0";
    if( $run_dir !~ m%^/% ) {	# relative pathname
	$run_dir = $ENV{"PWD"} . "/$run_dir";
    } 
    $run_dir =~ s%/(./)?[^/]*$%%;	# strip off "run-experiment"
    $ENV{"OSTORE_RUN_DIR"} = $run_dir;
}

use lib $ENV{"OSTORE_RUN_DIR"};
use Net::hostent;
use Socket;
use strict;
use Tools;
use Traceroute;

my $PINGS_PER_SERVER = 10;
my $TMP_FILE = "all_pairs_pings.$$.tmp";
# in secs (1/2 hour):
my $TOTAL_TIMEOUT = 3600;


my @servers = @ARGV;
if( $#servers == -1 ) {
    @servers = split(/\s+/, $ENV{"OSTORE_SERVERS"}) 
	if( $ENV{"OSTORE_SERVERS"} );
}
# if it's still empty, die
if( $#servers == -1 ) {
    die( "No host names provided, and \$OSTORE_SERVERS not set." );
}

# let the pings begin
my $my_tmp = "$tmp_dir/$TMP_FILE";
my $server;
my @pingtimes;
my @pids;
my $last_ssh_started_time;
my $kill_pings = "killall -9 -q ping_times.pl";
foreach $server (@servers) {

    # make the call
    my @ssh_args = &ssh_args( $server );
    my $command = &remote_run_dir( $server ) . "/ping_times.pl " . 
	"$PINGS_PER_SERVER";
    # killall previous ping_times
    $command = "$kill_pings; $command";
    my $this_tmp = $my_tmp . ".$server";
    my $pid = &fork_process_err( $this_tmp, $my_tmp, 
				 "ssh @ssh_args $server " . 
				 "\"$command @servers\"" );
    push @pids, $pid;
    sleep 1;

}
$last_ssh_started_time = time;

my $pid;
my $index = 0;
foreach $pid ( @pids ) {

    my $server = $servers[$index];
    my $time = time-$last_ssh_started_time;
    my $this_tmp = $my_tmp . ".$server";
    my $rzot_rt = &returned_zero_or_timeout( $TOTAL_TIMEOUT - $time, $pid );
    if( $rzot_rt == 0 ) {
	print STDERR "Ssh command to $server failed\n";
	# if we failed, do a traceroute and try to see where it went wrong
	my $tr = Net::Traceroute->new(host=> $server, 
				      max_ttl=> 30,
				      queries=> 1,
				      query_timeout=> 5);
	if(defined($tr->found) and $tr->stat == $tr->TRACEROUTE_OK and 
	   $tr->hop_query_host($tr->hops, 0) eq $server) {
	    # somehow we found it, even though we couldn't ping it
	    push @pingtimes, "*** ssh ping_times.pl failed on $server, " . 
		"but traceroute was successful ***\n";
	} else {
	    # return the last hop
	    my $hop = 1;
	    while( defined $tr->hop_query_host( $hop, 0 ) ) {
		$hop++;
	    }
	    my $lasthop = $tr->hop_query_host($hop-1, 0);
	    push @pingtimes, "*** ssh ping_times.pl failed on $server, " . 
		"last hop in traceroute was $lasthop ***\n";
	}

	$index++;
	unlink( $this_tmp );
	next;
    } elsif( $rzot_rt == -1 ) {
	# timeout occured
	kill $pid;
	my @ssh_args = &ssh_args( $server );
	$pid = &fork_process_err( $this_tmp, $my_tmp, 
				  "ssh @ssh_args $server $kill_pings" );
	# try to kill it for a minute at most, then give up
	if( &returned_zero_or_timeout( 60, $pid ) == -1 ) {
	    kill $pid;
	}
	push @pingtimes, "*** ssh ping_times.pl timed out after " . 
	    "$TOTAL_TIMEOUT seconds on $server ***\n";
	$index++;
	unlink( $this_tmp );
	next;
    }

    # now parse the results
    open( TMP, "<$this_tmp" ) or &clean_death( "Could not open $this_tmp" );

    my $line;
    foreach $line (<TMP>) {
	my @splitline = split( /\s+/, $line );
	if( $#splitline == $#servers ) {
	    push @pingtimes, $line;
	}
    }

    my $old_fd = select( STDERR );
    print "done with server $server\n";
    select( $old_fd );
    close( TMP );
    unlink( $this_tmp );
    $index++;
}
unlink( $my_tmp );

my($sec, $min, $hour, $monthday, $month, $year, $weekday)=localtime(time);

$year += 1900;
$month++;

if( $monthday < 10 ) {
    $monthday = "0" . $monthday;
}
if( $sec < 10 ) {
    $sec = "0" . $sec;
}
if( $min < 10 ) {
    $min = "0" . $min;
}

# print date and time to file
print "$hour:$min:$sec $month/$monthday/$year\n";

# print the number of pings per server
print "$PINGS_PER_SERVER\n";

# print out the servers
foreach $server (@servers) {
    my $ip;
    unless( $ip = gethost($server) ) {
	print $server . " ";
    } else {
	print inet_ntoa($ip->addr_list->[0]) . " ";
    }
}
print "\n";
# now print out the times and we're done
my $time;
foreach $time (@pingtimes) {
    print "$time";
}

################

sub clean_death {
    my $death_msg = shift;
    my $server;
    foreach $server (@servers) {
	unlink( $my_tmp . ".$server" );
    }
    unlink( $my_tmp );
    print $death_msg . "\n";
    exit( 1 );
}

################

sub fork_process_err {
    my $log = shift( @_ );
    my $err = shift( @_ );
    push @logs, $log;

    my $pid = fork();
    if( $pid > 0 ) {
	#parent;  return PID for child.
	push @child_pids, $pid;
	return $pid;
	
    } elsif( defined $pid ) {
	# child; run the named process..
	
	open STDOUT, ">$log" or &die( "can't redirect STDOUT");
	open STDERR, ">>$my_tmp" or &die( "can't redirect STDERR");

	# now exec the process as specified.
	exec (@_);
    }
    
    # should never reach this point.
    &die( "fork failed" );
}


