#!/usr/bin/perl -w

use strict;
use Getopt::Std;
use File::Copy;
use Tools;
use vars qw($opt_a $opt_c $opt_i $opt_u $opt_k $opt_l $opt_t $opt_x $opt_w $opt_h $opt_v $opt_s $opt_m $opt_f $opt_n $opt_z $opt_e $opt_d $opt_r);

sub usage
{
    select(STDERR);
    print <<'EOUsage';

run-lat-test.pl [options]
  Options:
    -h           - Display this help message.
    -r           - Perform a read latency test instead of a write latency test.
                   Valid arguments are 'a' for Archive (-d has to be true as 
                   well), 't' for Tapestry, and 'd' for Dissemination Tree.
    -a           - Enable the archive.  By default, the inner ring does
                   not initiate the inner ring.
    -i           - Inline the archive.  This option has no effect if the
                   archive is disabled.  By default, when the archive is
                   enabled, it runs asynchronously with the updates.
    -e           - Use a single Serializer node instead of a set of 
                   Inner Ring nodes.  Inherently the archive is on 
                   and inline.
    -d           - Disseminate Fragments.  false by default.
    -n SERS      - Use SERS Serializers.  Distribute the clients evenly
                   over the Serializers.  Only meaningful with -s.  Default
		   value is 1.
    -u UPDT_SIZE - Sweep the size of the update submitted by test stages
                   across the range describe by UPDT_SIZE.  The format
                   for notating the range is described below.  Note, the
                   size of the data object is set to twice the length of
                   the update size.
    -w WARM_CFG  - Warm the configuration by performing a number of
                   updates at an interval as specified by the WARM_CFG
                   phase descriptor.  The format for notating the test
                   stage is described below.
    -m MEAS_CFG  - Measure the configuration by performing a number of
                   updates at an interval as specified by the MEAS_CFG
                   phase descriptor.  The format for notating the test
                   stage is described below.
    -s SETTLE    - Pause SETTLE ms between the initialization and test
                   phases.
    -x ROC       - Relocate the test in the X-cluster to start using
                   machine xROC.millennium.berkeley.edu.
		   NOTE: if no -x option is specified, the test will use
		   the machines listed in the OSTORE_SERVERS environment
		   variable.
    -v           - Verbose results.  Dump the individual stats to a file.
    -f INNER_CFG - Use the file INNER_CFG as the configuration file
                   for members of the inner ring.
    -k KEY_SIZE  - Use keys of of length KEY_SIZE bits.  Valid values
                   are 512 and 1024.  The default value is 512.
    -z		 - Include the Visdemo stage for each node.

  Ranges can be specified in a very flexible manner.  Most generally,
  each range has a 'start' value, an 'end' value, a 'skip operator',
  and a 'skip value'.  The skip operator is one of the four basic
  arithmetic operators (+, -, *, and /).  During execution, when the
  range operator is '+' or '*', the range leads to execution as
    for (x = $start; x <= $end; x = x $skip_op $skip_value) { ; }
  When the operator is '-' or '/', the range leads to execution as
    for (x = $start; x >= $end; x = x $skip_op $skip_value) { ; }

  A range is notated as '$start-$end/$skip_op$skip_value'.  Several
  shortened forms are also accepted.  These forms include '$start' which
  executes only one point, '$start-$end' which uses the '+' operator
  with a skip value of 1, and '$start-$end/$skip_value' which defaults
  to use the '+' operator.  Examples of valid ranges include

    o  28
    o  28-56
    o  2-10/2
    o  128-1024/*2
    o  2048-128//2

  The different stages of the test are described by phase descriptor.
  A phase descriptor specifies the number of updates to perform during
  that stage and the amount of time to pause, in milliseconds, before
  submitting each update.  The general form of the phase descriptor is
  '$number_updates/$pause_time'.  The shortened for '$number_updates'
  is also accepted; the pause time defaults to 0.  Examples of valid
  phase descriptors include

    o  1000
    o  1000/5
EOUsage

    exit(1);
}

# Set a nice signal handler.
$SIG{INT} = \&slow_death;
$SIG{QUIT} = \&slow_death;
$SIG{HUP} = \&slow_death;
$SIG{PIPE} = \&slow_death;
$SIG{TERM} = \&slow_death;

my $LAT_RESULT_LOG;
my $LOG_MODIFIER;

#  Test parameters.
my $SKIP_ARCHIVE = 'true';
my $INLINE_ARCHIVE = 'false';
my @SERVERS = split /\s+/, $ENV{"OSTORE_SERVERS"} if ($ENV{"OSTORE_SERVERS"});;
my $KEY_SIZE = 512;
my $SETTLE = 5000;
my $VERBOSE_RESULTS = 0;
my $INNER_CFG = "./inner.cfg";
my $NUM_SERIALIZERS = 1;
my $INNER_RING = 'true';

# archive parameters
my $DISSEMINATE = 'false';
my $NUM_STORAGE_NODES = 32;
my $NUM_STORAGE_MACHINES = 14;
my $NUM_FRAGS = 32;		# Must align with REQUEST_CARDINALITY
my $REQUEST_CARDINALITY = 24;	# Must align with NUM_FRAGS
my $ARCHIVER_CFG = "./archiver.cfg";
my $STORAGE_CFG = "./storage.cfg";
my $SET_CREATOR_CFG = "./creator.cfg";
my $RP_CFG = "./rp.cfg";
my $VERBOSE_ARCHIVE = 'false';

my $PUBLISH = 'true';

# read parameters
my $READ = 'false';
my $USE_DTREE = 'false';

# Visdemo stage
my $VISDEMO = 'false';

my @WARM_PHASE_DESC;
my $WARM_UPDATES      = 1000;
my $WARM_PAUSE        = 5;

my @MEAS_PHASE_DESC;
my $MEAS_UPDATES      = 1000;
my $MEAS_PAUSE        = 100;

my @UPDATE_SIZE_SPEC;
my $UPDATE_SIZE_START = 1024;
my $UPDATE_SIZE_END   = 1024;
my $UPDATE_SIZE_OP    = '+';
my $UPDATE_SIZE_SKIP  = 1;

my $DELAY = 10000;

my $run_experiment_pid;


#  Get the user-defined parameters.
my $opts_ok = getopts('haivdec:n:u:k:r:m:f:w:s:t:x:z');
if ($opt_h) { &usage; }
if ($opt_z) { $VISDEMO = 'true'; }
if ($opt_a) { $SKIP_ARCHIVE = 'false'; }
if ($opt_i) { $INLINE_ARCHIVE = 'true'; $SKIP_ARCHIVE = 'false'; }
if ($opt_d) { $DISSEMINATE = 'true'; $PUBLISH = 'false'; }
if ($opt_x) { @SERVERS = map "x$_.millennium.berkeley.edu", ($opt_x .. 42); }
if ($opt_e) { $INNER_RING = 'false'; }
if ($opt_n) { $NUM_SERIALIZERS = $opt_n; }
if ($opt_k) { $KEY_SIZE = $opt_k; }
if ($opt_v) { $VERBOSE_RESULTS = 1; }
if ($opt_s) { $SETTLE = $opt_s; }
if ($opt_f) { $INNER_CFG = $opt_f; }
if ($opt_r)
{
  $READ = 'true'; 
  $INLINE_ARCHIVE = 'true';
  $SKIP_ARCHIVE = 'false';
  if($opt_r eq 'a') {
    $PUBLISH = 'false';
    $USE_DTREE = 'false';
    if($DISSEMINATE eq 'false') {
      &usage;
    }
  } elsif($opt_r eq 't') {
    $PUBLISH = 'true';
    $USE_DTREE = 'false';
  } elsif($opt_r eq 'd') {
    $PUBLISH = 'false';
    $USE_DTREE = 'true';
  } else {
      print "DANG!!!! opt_r=$opt_r \n\n\n\n";
    &usage;
  }
}
if ($opt_w)
{
    @WARM_PHASE_DESC = &parse_phase_desc($opt_w);
    my @desc = @WARM_PHASE_DESC;
    $WARM_UPDATES = shift(@desc);
    $WARM_PAUSE   = shift(@desc);
}
if ($opt_m)
{
    @MEAS_PHASE_DESC = &parse_phase_desc($opt_m);
    my @desc = @MEAS_PHASE_DESC;
    $MEAS_UPDATES = shift(@desc);
    $MEAS_PAUSE   = shift(@desc);
}
if ($opt_u)
{
    @UPDATE_SIZE_SPEC = &parse_range_spec($opt_u);
    my @spec = @UPDATE_SIZE_SPEC;
    $UPDATE_SIZE_START = shift(@spec);
    $UPDATE_SIZE_END   = shift(@spec);
    $UPDATE_SIZE_OP    = shift(@spec);
    $UPDATE_SIZE_SKIP  = shift(@spec);
}

#  The command line is valid if getopts returned without error and
#  there are no arguments left over.  $#ARGV returns the number of
#  remaining arguments minus one.
if ((! $opts_ok) || ($#ARGV != -1)) { &usage; }

#  Check that the set of input parameters is valid.
if ($KEY_SIZE != 512 && $KEY_SIZE != 1024)
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Invalid key size.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}
if ( ! @SERVERS )
{
    my $old_fd = select(STDERR);
    print<<EOErrMsg;
ERROR:  Undefined cluster machine.
EOErrMsg
    select($old_fd);
    &usage();
    exit(1);
}

#  Set the environment variable for the keys.
my $key_cmd;
if ($KEY_SIZE == 512)
{
    $ENV{KEYSHARE_PREFIX} = "ir";
    $key_cmd = "KEYSHARE_PREFIX=ir";
}
elsif ($KEY_SIZE == 1024)
{
    $ENV{KEYSHARE_PREFIX} = "ir-1024-4-2";
    $key_cmd = "KEYSHARE_PREFIX=ir-1024-4-2";
}
else
{
    print(STDERR "Script error.");
}

#  Compute the result log name.
$LAT_RESULT_LOG = "/tmp/lat_result_log.noar.$$.log";
$LOG_MODIFIER = 'noar';
if (($SKIP_ARCHIVE eq 'false') && ($INLINE_ARCHIVE eq 'true'))
{
    $LAT_RESULT_LOG = "/tmp/lat_result_log.inar.$$.log";
    $LOG_MODIFIER = 'inar';
}
elsif ($SKIP_ARCHIVE eq 'false')
{
    $LAT_RESULT_LOG = "/tmp/lat_result_log.dear.$$.log";
    $LOG_MODIFIER = 'dear';
}

if($INNER_RING eq 'true')
{
  #  Create the inner ring config files with the right archive settings.
  copy($INNER_CFG, "/tmp/inner-$$.cfg");

  if($DISSEMINATE eq 'true') {
    &add_stage("/tmp/inner-$$.cfg", "Disseminator.stg");
    if($VISDEMO eq 'true') {
       &add_stage("/tmp/inner-$$.cfg", "Visdemo.stg");
    }
  } else {
    if($VISDEMO eq 'true') {
       &add_stage("/tmp/inner-$$.cfg", "Visdemo.stg");
    }
  }
  push @configs, "/tmp/inner-$$.cfg";
}
else
{
  #  Create the archiver config files with the right settings.
  for( my $i = 0; $i < $NUM_SERIALIZERS; $i++ ) {
    copy($ARCHIVER_CFG, "/tmp/archiver-$i-$$.cfg");

    if($DISSEMINATE eq 'true') {
      &add_stage("/tmp/archiver-$i-$$.cfg", "Disseminator.stg");
      if($VISDEMO eq 'true') {
	&add_stage("/tmp/archiver-$i-$$.cfg", "Visdemo.stg");
      }
    } else {
      if($VISDEMO eq 'true') {
	&add_stage("/tmp/archiver-$i-$$.cfg", "Visdemo.stg");
      }
    }
    push @configs, "/tmp/archiver-$i-$$.cfg";
  }
}

# write rp
copy($RP_CFG, "/tmp/rp-$$.cfg");
if($VISDEMO eq 'true') {
  &add_stage("/tmp/rp-$$.cfg", "Visdemo.stg");
}
push @configs, "/tmp/rp-$$.cfg";


# write disseminate cfg files
if($DISSEMINATE eq 'true')
{
  # write set create
  copy($SET_CREATOR_CFG, "/tmp/creator-$$.cfg");

  if($VISDEMO eq 'true') {
    &add_stage("/tmp/creator-$$.cfg", "Visdemo.stg");
  }
  push @configs, "/tmp/creator-$$.cfg";


  # write set storage server
  copy($STORAGE_CFG, "/tmp/storage-$$.cfg");

  if($VISDEMO eq 'true') {
    &add_stage("/tmp/storage-$$.cfg", "Visdemo.stg");
  }
  push @configs, "/tmp/storage-$$.cfg";
}

#  Run experiments to collect data.
my $update_size = $UPDATE_SIZE_START;
my $update_size_comparator = '<=';
if ($UPDATE_SIZE_OP eq '-' || $UPDATE_SIZE_OP eq '/')
{
    $update_size_comparator = '>=';
}
while (eval("$update_size $update_size_comparator $UPDATE_SIZE_END"))
{
    my $dobj_size = $update_size * 2;

    #  Create configuration files.
    my $tester_cfg   = &write_tester_config_file();
    #my $exp_cfg      = &write_experiment_file();
    my $exp_cfg     = "";
    if($INNER_RING eq 'true')
    {
      $exp_cfg = &write_inner_experiment_file($dobj_size, $update_size,
					      $WARM_UPDATES, $WARM_PAUSE,
					      $MEAS_UPDATES, $MEAS_PAUSE,
					      $SETTLE);
    }
    else
    {
      $exp_cfg = &write_serializer_experiment_file($NUM_SERIALIZERS,
						   $dobj_size,
						   $update_size,
						   $WARM_UPDATES,
						   $WARM_PAUSE,
						   $MEAS_UPDATES,
						   $MEAS_PAUSE,
						   $SETTLE);
    }

    my $lat_fork_log = "/tmp/lat_fork_log.$$.log";

    #  Run the experiment.
    print("Running lat test: $key_cmd ./run-experiment $exp_cfg\n" .
          "  WarmPhase    = $WARM_UPDATES/$WARM_PAUSE\n" .
          "  MeasurePhase = $MEAS_UPDATES/$MEAS_PAUSE\n" .
          "  Settle       = $SETTLE\n" .
          "  KeySize      = $KEY_SIZE\n" .
          "  DobjSize     = $dobj_size\n" .
          "  UpdateSize   = $update_size\n");

    $run_experiment_pid = &fork_process($lat_fork_log,
                            "./run-experiment", $exp_cfg);
    if(! &returned_zero($run_experiment_pid))
    {
        die("Experiment returned non-zero.\n");
    }

    #  Determine the result log.
    my @internal_logs = glob("/tmp/*-${run_experiment_pid}*-Tester-internal.log");
    (scalar(@internal_logs) == 1) ||
      die("Cannot determine result log name.\n");
    my $internal_log = shift(@internal_logs);

    #  Read the result log.
    my @stats = &analyze_result_log($internal_log);
    print("  Median       = $stats[3]\n" .
          "  5th          = $stats[5]\n" .
          "  95th         = $stats[6]\n");

    #  Dump the raw data.
    if ($VERBOSE_RESULTS)
    {
        my $verbose_log =
          "/tmp/lat_result_log.$$.$LOG_MODIFIER.$update_size.raw";
        print("  Writing raw data to $verbose_log...\n");
        &write_raw_result_log($internal_log, $verbose_log);
    }

    #  Write to log.
    &write_result_to_log($SKIP_ARCHIVE, $INLINE_ARCHIVE, $KEY_SIZE,
                         $WARM_UPDATES, $WARM_PAUSE,
                         $MEAS_UPDATES, $MEAS_UPDATES, $SETTLE,
                         $dobj_size, $update_size, @stats);

    #  Remove the files we created.
    unlink($tester_cfg, $exp_cfg, $internal_log, $lat_fork_log);

    $update_size = eval("$update_size $UPDATE_SIZE_OP $UPDATE_SIZE_SKIP");
}

#  Remove the inner ring config file used for this test.
unlink(@configs);
exit(0);

sub parse_phase_desc
{
    my $desc_in = shift(@_);

    my $updates = 0;
    my $pause   = 0;

    if ($desc_in =~ /(\d+)(\/(\d+))?/ )
    {
        $updates = $1;
        if (defined($3)) { $pause = $3; }
    }

    my @desc;
    push(@desc, $updates);
    push(@desc, $pause);

    return @desc;
}

sub parse_range_spec
{
    my $spec = shift(@_);

    my $start = 0;
    my $end   = 0;
    my $op    = '+';
    my $skip  = 1;

    #  (\d+)                  The start of the range...
    #  (-(\d+))?              ...optionally a '-' and the end of the range...
    #  (\/([-\+\*\/])?(\d+))? ...optionally a '/', an optional 
    #                         operation (-, +, *, or /) and the skip value.
    if ($spec =~ /(\d+)(-(\d+))?(\/([-\+\*\/])?(\d+))?/ )
    {
        $start = $1;
        $end   = $start;
        if (defined($3)) { $end = $3; }
        if (defined($5)) { $op = $5; }
        if (defined($6)) { $skip = $6; }
    }

    #print("start=$start \n");
    #print("end=$end   \n");
    #print("op=$op    \n");
    #print("skip=$skip  \n");

    if (($op eq '+') || ($op eq '*'))
    {
        if ($start > $end)
        {
          die("Invalid range specification.\n" .
              "  start=$start, end=$end, operation=$op\n");
        }
    }
    else
    {
        if ($end > $start)
        {
          die("Invalid range specification.\n" .
              "  start=$start, end=$end, operation=$op\n");
        }
    }

    my @range;
    push(@range, $start);
    push(@range, $end);
    push(@range, $op);
    push(@range, $skip);

    return @range;
}

sub write_result_to_log
{
    my $skip         = shift(@_);
    my $inline       = shift(@_);
    my $key_size     = shift(@_);
    my $warm_updates = shift(@_);
    my $warm_pause   = shift(@_);
    my $meas_updates = shift(@_);
    my $meas_pause   = shift(@_);
    my $settle       = shift(@_);
    my $dobj_size    = shift(@_);
    my $update_size  = shift(@_);
    my $min          = shift(@_);
    my $max          = shift(@_);
    my $mean         = shift(@_);
    my $median       = shift(@_);
    my $stddev       = shift(@_);
    my $fiveperc     = shift(@_);
    my $ninefiveperc = shift(@_);

    my $log =
      "SkipArchive=${skip} " .
      "InlineArchive=${inline} " .
      "KeySize=${key_size} " .
      "WarmPhase=${warm_updates}/${warm_pause} " .
      "MeasPhase=${meas_updates}/${meas_pause} " .
      "Settle=${settle} " .
      "DobjSize=${dobj_size} " .
      "UpdateSize=${update_size} " .
      "Min=${min} " .
      "Max=${max} " .
      "Mean=${mean} " .
      "Median=${median} " .
      "StdDev=${stddev} " .
      "5Percentile=${fiveperc} " .
      "95Percentile=${ninefiveperc}\n";

    open(RESULT_LOG, ">> $LAT_RESULT_LOG") ||
      die("Could not open result log ($LAT_RESULT_LOG).");
    print(RESULT_LOG "$log");
    close(RESULT_LOG);

    return;
}

sub write_raw_result_log
{
    my $internal_log = shift(@_);
    my $raw_log      = shift(@_);

    copy("$internal_log", "$raw_log");

    return;
}

sub analyze_result_log
{
    my $result_log = shift(@_);

    open(RESULT, "./stats.pl -f $result_log 4 | ") ||
        die("Could not open result log.\n");

    my $min = 0;
    my $max = 0;
    my $mean = 0;
    my $median = 0;
    my $stddev = 0;
    my $fiveperc = 0;
    my $ninefiveperc = 0;
    while (<RESULT>)
    {
        if (/^min: (\d*)/)
        {
            $min = $1;
        }
        elsif (/^max: (\d*)/)
        {
            $max = $1;
        }
        elsif (/^mean: ([\d\.]*)/)
        {
            $mean = $1;
        }
        elsif (/^median: (\d*)/)
        {
            $median = $1;
        }
        elsif (/^stddev: ([\d\.]*)/)
        {
            $stddev = $1;
        }
        elsif (/^5percentile: (\d*)/)
        {
            $fiveperc = $1;
        }
        elsif (/^95percentile: (\d*)/)
        {
            $ninefiveperc = $1;
        }
    }

    close(RESULT);

    #print("$min $max $mean $median $fiveperc $ninefiveperc\n");
    my @stats;
    push(@stats, $min);
    push(@stats, $max);
    push(@stats, $mean);
    push(@stats, $median);
    push(@stats, $stddev);
    push(@stats, $fiveperc);
    push(@stats, $ninefiveperc);

    return @stats;
}

sub write_inner_experiment_file
{
  my $dobj_size    = shift(@_);
  my $update_size  = shift(@_);
  my $warm_updates = shift(@_);
  my $warm_pause   = shift(@_);
  my $meas_updates = shift(@_);
  my $meas_pause   = shift(@_);
  my $settle       = shift(@_);

  my $exp_cfg = "/tmp/lat.$$.exp";
  open(EXP, ">$exp_cfg") ||
    die("Could not create experiment configuration file ($exp_cfg).\n");
  my $old_fd = select(EXP);
  my $fed = 0;
  my $ir0 = 1;
  my $ir1 = 2;
  my $ir2 = 3;
  my $ir3 = 4;
  my $tester = 5;

  print <<EOExp;
hosts      @SERVERS
topology   NONE
Federation \${run_dir}/federation.cfg fed     daemon $fed
RP         /tmp/rp-$$.cfg         static  daemon $fed
Ring0      /tmp/inner-$$.cfg             static  daemon $ir0
Ring1      /tmp/inner-$$.cfg             static  daemon $ir1
Ring2      /tmp/inner-$$.cfg             static  daemon $ir2
Ring3      /tmp/inner-$$.cfg             static  daemon $ir3
Tester     /tmp/lat-tester.$$.cfg     static  mortal $tester
EOExp

  if($DISSEMINATE eq 'true') {
    &addStorage(($tester + 1),
		$NUM_STORAGE_NODES,
		$NUM_STORAGE_MACHINES,
		$fed);
  }

  # write global parameters
  print <<EOExp;
\# Parameter overrides:
set	*.global.PingDelay			$DELAY
set	*.RequestorStage.SendCachePin		true
set	*.RecoalesceBlkStage.SendCachePin	true

set	Ring*.Inner.num_frags			$NUM_FRAGS
set	Ring*.Inner.skip_archive	       	$SKIP_ARCHIVE
set	Ring*.Inner.inline_archive	       	$INLINE_ARCHIVE
set	Ring*.CacheManager.Publish		$PUBLISH

set	Ring*.GenerateFragsStage.Verify		false

set	Ring*.RequestorStage.RequestCardinality	$REQUEST_CARDINALITY

set	Tester.RequestorStage.UseDtreeRegardless	$USE_DTREE
set	Tester.RequestorStage.RequestCardinality	$REQUEST_CARDINALITY

set	Tester.ReadStage.UseDisk		false

set	Tester.DTreeNode.lease_length  		0
EOExp

  if($READ eq 'false') {
  print <<EOExp;
set	Tester.ReplaceMicroBenchmark.InitLength		$warm_updates
set	Tester.ReplaceMicroBenchmark.SettleTime		$settle
set	Tester.ReplaceMicroBenchmark.TestLength		$meas_updates
set	Tester.ReplaceMicroBenchmark.TestPauseTime	$meas_pause
set	Tester.ReplaceMicroBenchmark.DobjSize		$dobj_size
set	Tester.ReplaceMicroBenchmark.ReplaceSize	$update_size
EOExp
  } else {
  print <<EOExp;
set	Tester.ReadLatencyMicroBenchmark.InitLength    	$warm_updates
set	Tester.ReadLatencyMicroBenchmark.SettleTime    	$settle
set	Tester.ReadLatencyMicroBenchmark.TestLength    	$meas_updates
set	Tester.ReadLatencyMicroBenchmark.TestPauseTime	$meas_pause
set	Tester.ReadLatencyMicroBenchmark.DobjSize      	$dobj_size
set	Tester.ReadLatencyMicroBenchmark.ReplaceSize	$update_size
EOExp
  }

  # write parameters dependant on the fact of dissemination.
  if($DISSEMINATE eq 'true') {
    print <<EOExp;
set	Creator.SetCreatorStage.SetSize      		$NUM_FRAGS

set	Ring*.RequestorStage.AlarmTimeout      		180000
set	Ring*.RequestorStage.AlarmTimeout      		180000
set	Ring*.GenerateChkptStage.AlarmTimeout          	180000
set	Ring*.GenerateFragsStage.AlarmTimeout          	180000
set	Ring*.RecoalesceBlkStage.AlarmTimeout          	180000
set	Ring*.RequestorStage.AlarmTimeout      		180000
set	Ring*.DisseminatorStage.AlarmTimeout      	180000
set	Ring*.ArchivalStorageStage.AlarmTimeout        	180000
set	Ring*.ArchivalStorageStage.StorageServer	false
set	Ring*.MemoryCache.Publish			false

set	StorageServer*.ArchivalStorageStage.AlarmTimeout	60000
EOExp
  }

  close(EXP);
  select($old_fd);

  return $exp_cfg;
}

#for archiver test
sub write_serializer_experiment_file
{
  my $num_serializers = shift(@_);
  my $dobj_size    = shift(@_);
  my $update_size  = shift(@_);
  my $warm_updates = shift(@_);
  my $warm_pause   = shift(@_);
  my $meas_updates = shift(@_);
  my $meas_pause   = shift(@_);
  my $settle       = shift(@_);

  my $exp_cfg = "/tmp/lat.$$.exp";
  open(EXP, ">$exp_cfg") ||
    die("Could not create experiment configuration file ($exp_cfg).\n");
  my $old_fd = select(EXP);
  my $fed = 0;
  my $archiver = 1; # 2 .. 4 are not used right now
  my $tester = 2;

  print <<EOExp;
hosts      @SERVERS
topology   NONE
Federation \${run_dir}/federation.cfg  fed     daemon $fed
Tester     /tmp/lat-tester.$$.cfg     static  mortal $tester
EOExp

  for (my $i = 0; $i < $num_serializers; $i++)
  {
    my $archiver_cfg = "/tmp/archiver-$i-$$.cfg";
    my $machine_num = $archiver + $i;
    my $archiver_name = "Archiver$i";

    print("$archiver_name $archiver_cfg static  daemon $machine_num\n");
  }

  if($DISSEMINATE eq 'true') {
    &addStorage(($tester + $num_serializers),
		$NUM_STORAGE_NODES,
		$NUM_STORAGE_MACHINES,
		$fed);
  }

  # write global parameters
  print <<EOExp;
\# Parameter overrides:
set	*.global.PingDelay			$DELAY
set	*.RequestorStage.SendCachePin		true
set	*.RecoalesceBlkStage.SendCachePin	true

set	Archiver*.GenerateFragsStage.Verify    	false

set	Tester.ReadStage.UseDisk		false
EOExp

  if($READ eq 'false') {
  print <<EOExp;
set	Tester.ReplaceMicroBenchmark.InitLength		$warm_updates
set	Tester.ReplaceMicroBenchmark.SettleTime		$settle
set	Tester.ReplaceMicroBenchmark.TestLength		$meas_updates
set	Tester.ReplaceMicroBenchmark.TestPauseTime	$meas_pause
set	Tester.ReplaceMicroBenchmark.DobjSize		$dobj_size
set	Tester.ReplaceMicroBenchmark.ReplaceSize	$update_size
EOExp
  } else {
  print <<EOExp;
set	Tester.ReadLatencyMicroBenchmark.InitLength    	$warm_updates
set	Tester.ReadLatencyMicroBenchmark.SettleTime    	$settle
set	Tester.ReadLatencyMicroBenchmark.TestLength    	$meas_updates
set	Tester.ReadLatencyMicroBenchmark.TestPauseTime	$meas_pause
set	Tester.ReadLatencyMicroBenchmark.DobjSize      	$dobj_size
set	Tester.ReadLatencyMicroBenchmark.ReplaceSize	$update_size
EOExp
  }

  # write serializer parameters
  for( my $i = 0; $i < $num_serializers; $i++ ) {
    #print("set\tArchiver$i.*.LogEvents\t\t\t$VERBOSE_ARCHIVE\n");
    print "set\tArchiver$i.ArchiverStage.archiver_number\t$i\n";
    print "set\tTester.FakeClient.rp_pkeyfile\t\${Archiver" . $i . "PublicKey}\n";
    print "set\tTester.ReplaceMicroBenchmark.rp_pkeyfile\t\${Archiver" . $i . "PublicKey}\n";
  }

  # write parameters dependant on the fact of dissemination.
  if($DISSEMINATE eq 'true') {

#    for( my $i = 0; $i < $num_serializers; $i++ ) {
#      #print("set\tArchiver$i.*.LogEvents\t\t\t$VERBOSE_ARCHIVE\n");
#      print("set\tArchiver$i.*.AlarmTimeout\t\t180000\n");
#    }
    print "set Creator.SetCreatorStage.SetSize $NUM_FRAGS\n";
    print "set\tStorageServer*.ArchivalStorageStage.AlarmTimeout\t60000\n";
  }

  # write tester parameters
  print "set\tTester.Router.fake_keys\t\ttrue\n";
  print "set\tTester.RequestorStage.RequestCardinality\t$REQUEST_CARDINALITY\n";
  print "set\tTester.FakeClient.AlarmTimeout\t60000\n";

  close(EXP);
  select($old_fd);

  return $exp_cfg;
}

sub write_tester_config_file
{
    my $dobj_size    = shift(@_);
    my $update_size  = shift(@_);
    my $warm_updates = shift(@_);
    my $warm_pause   = shift(@_);
    my $meas_updates = shift(@_);
    my $meas_pause   = shift(@_);
    my $settle       = shift(@_);

    my $tester_cfg = "/tmp/lat-tester.$$.cfg";
    open(TESTER, ">$tester_cfg") ||
        die("Could not create tester configuration file ($tester_cfg).\n");

    my $old_fd = select(TESTER);

    print <<'EOHeader';
# Tester config file
<sandstorm>
    <!include Generic.hdr>
    <stages>
    </stages>
</sandstorm>
EOHeader

    close(TESTER);
    select($old_fd);

    &write_tester_header();
    &write_tester_stage();
#    &write_tester_stage($dobj_size, $update_size,
#                        $warm_updates, $warm_pause,
#                        $meas_updates, $meas_pause, $settle);
    if($VISDEMO eq 'true') {
      &add_stage("/tmp/lat-tester.$$.cfg", "Visdemo.stg");
    }

    return $tester_cfg;
}


sub write_tester_header
{
  &write_preamble_tester_header();
  if($INNER_RING eq 'true')
  {
    &write_inner_tester_header();
  }
  else
  {
    &write_serializer_tester_header();
  }
}

sub write_preamble_tester_header
{
  &add_stage("/tmp/lat-tester.$$.cfg", "Network.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "Router.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "StaticTClient.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "DynamicTClient.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "Patchwork.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "Read.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "Requestor.stg");
#  &add_stage("/tmp/lat-tester.$$.cfg", "MemoryCache_private.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "CacheManager.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "BlockBufferCache.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "StorageManager.stg");

  if($DISSEMINATE eq 'true') {
    &add_stage("/tmp/lat-tester.$$.cfg", "RecoalesceBlk.stg");
  }
  return;
}

sub write_inner_tester_header
{
  &add_stage("/tmp/lat-tester.$$.cfg", "Client.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "Replica.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "DTreeNode.stg");
}

sub write_serializer_tester_header
{
  &add_stage("/tmp/lat-tester.$$.cfg", "FakeClient.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "UpdateVerifiableBlocks.stg");
  &add_stage("/tmp/lat-tester.$$.cfg", "Update.stg");
}

sub write_tester_stage
{
  if($READ eq 'false') {
    &add_stage("/tmp/lat-tester.$$.cfg", "ReplaceMicroBenchmark.stg");
  } else {
    &add_stage("/tmp/lat-tester.$$.cfg", "ReadLatencyMicroBenchmark.stg");
  }
}

# Adds a new stage to a config file.
#
# This functions appends the suffix ".stg" to the end of the stagename.
# NOTE: The stagename should be an existing stage file in the
# pond/run/cfg directory.
#
# Usage: &add_stag( <cfg_filename>, <stagename> )
sub add_stage
{
  my ($cfg_file, $stage) = @_;

  my $tmp_name = "$cfg_file.$$.tmp";
  my $key;

  open IN, "<$cfg_file" or &die( "add_stage: can't open IN=$cfg_file, $stage");
  open OUT, ">$tmp_name" or &die( "add_stage: can't open OUT=$tmp_name");

  # find right before the </stages> 
  while(<IN>) {
    last if (/^\s*<\/stages>/);
    print OUT $_;
  }

  print OUT "\n\t<!include $stage>\n";

  print OUT $_; #print </stages>

  # copy the rest
  while(<IN>) {
    print OUT $_;
  }

  close(IN);
  close(OUT);

  # now replace the old copy with the new one.
  rename($tmp_name, $cfg_file);
}

# Removes a stage from a config file
#
# This functions appends the suffix ".stg" to the end of the stagename.
# and removes the entry matching <stagename>.stg.
#
# Usage: &add_stag( <cfg_filename>, <stagename> )
sub remove_stage
{
  my ($cfg_file, $stage) = @_;

  my $tmp_name = "$cfg_file.$$.tmp";
  my $key;

  open IN, "<$cfg_file" or &die( "add_stage: can't open IN=$cfg_file");
  open OUT, ">$tmp_name" or &die( "add_stage: can't open IN=$tmp_name");

  # find right before the </stages> 
  while(<IN>) {
    next if (/^\s*<\!include\s+$stage\.stg.*/);
    print OUT $_;
  }

  print OUT "\n\t<!include $stage>\n";

  print OUT $_; #print </stages>

  # copy the rest
  while(<IN>) {
    print OUT $_;
  }

  close(IN);
  close(OUT);

  # now replace the old copy with the new one.
  rename($tmp_name, $cfg_file);
}

# add storage servers to experiment file
sub addStorage
{
    my $storage_base         = shift(@_);
    my $num_storage_nodes    = shift(@_);
    my $num_storage_machines = shift(@_);
    my $fed                  = shift(@_);

    for(my $i = 0; $i < $num_storage_nodes; $i++) {
	print("StorageServer$i\t/tmp/storage-$$.cfg\tstatic  daemon " . 
	      ($storage_base + int($i)%int($num_storage_machines)) . "\n"); 
    }
    print("Creator\t/tmp/creator-$$.cfg\tstatic  daemon $fed\n");
}

# try to kill run-experiment nicely.
sub slow_death {
    print(  "*** Caught SIG@_ ***\n" );
    if( defined $run_experiment_pid ) {
	print "  Killing run-experiment...\n";
	kill 2, $run_experiment_pid;
	# Now wait for it to die.
	&returned_zero( $run_experiment_pid );
    }
    &exit_all( 1 );
}
