#!/usr/bin/perl -w
#
# Copyright (c) 2001 Regents of the University of California.
# All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#  3. Neither the name of the University nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
#  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
#  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#  SUCH DAMAGE.

#
# author: Dennis Geels and Jeremy Stribling
# $Id: extract_route_times.pl,v 1.5 2003/04/26 00:23:58 strib Exp $
#
# Extract results from LocalOptimizationTest (simulation) logs.

use strict;
use Statistics::Descriptive;
$| = 1;

my $bucket_width = 15000;
my @begin_times;
my @find_buckets;
my @fail_buckets;
my @logs = @ARGV;
my $max_offset = 0;
my @destroyed_objs = ();
my $min_owner_count = 1;	# need enough votes for consistency.
my @completed_buckets;
my @bw_buckets;
my %owners;	# maps keys to a list of owners found by any lookup.
my %good_owner;	# maps keys to the owner found by most lookups.
my %ignore_owner; # set of keys to the owner found by too few lookups.
my $num_nodes_up = 0;
my @num_node_buckets;
my $num_nodes_started = 0;
my @num_node_started_buckets;

####################
# Extract info from logs
# First pass: choose owners of each key (majority wins)
foreach my $log (@logs) {
#    print "Checking $log\n";
    open( IN, "<$log" ) or die;
    while (<IN>) {
	if( /FOUND: .* obj (.+) on node (.+) with guid (.+) at time (\d+) in time (\d+) .*/ ) {
	    my ($start_time, $key, $owner, $find_time) = ($4, $1, $3, $5);
	    push @{$owners{$key}}, $owner;
	}
    }
    close( IN ) or die;
}
# Choose owner
foreach my $key (keys %owners) {
    my @owners = sort @{$owners{$key}};
    my $owner_count = scalar(@owners);
    $good_owner{$key} = "NOBODY";
    if( $owner_count < $min_owner_count ) {
	warn( "Only $owner_count owners found $key\n" );
	$ignore_owner{$key} = $key;
    } else {
	my %votes;
	my $owner;
	foreach $owner (@owners) {
	    $votes{$owner}++;
	}
	foreach $owner (keys %votes) {
	    if( $votes{$owner} > ($owner_count/2) ) {
		$good_owner{$key} = $owner;
		if( $votes{$owner} < $owner_count ) {
		    warn "$owner gets $votes{$owner}/$owner_count " .
			"votes for $key\n";
		}
	    }
	}
    }
}

foreach my $log (@logs) {
#    print "Checking $log\n";

    my $begin_time = 0;
    my $bw_mode = 0;
    my $bw_bucket = 0;
    my $curr_head = -1;
    my $curr_tail = -1;
    my $curr_total = -1;
    
    open( IN, "<$log" ) or die;
    while (<IN>) {

	if( $bw_mode ) {
	    # looking for bandwidth number, only care about stub edges
	    if( /(\d+) (\d+) (\d+) .* true/ ) {
		# aggregate bw over all message per link
		if( $curr_head != $1 and $curr_tail != $2 ) {
		    # new link . . .
		    if( $curr_total != -1 ) {
			push @{$bw_buckets[$bw_bucket]}, ($curr_total*1000
							  /$bucket_width);
			#print "Pushed $curr_total onto bucket $bw_bucket " . 
			#    "for link $curr_head->$curr_tail\n";
		    }
		    $curr_head = $1;
		    $curr_tail = $2;
		    $curr_total = 0;
		}
		$curr_total += $3;
	    } elsif( /BANDWIDTH STOP/ ) {
		$bw_mode = 0;
		$curr_head = -1;
		$curr_tail = -1;
		$curr_total = -1;
	    }
	    next;
	}


	if( /FOUND: .* obj (.+) on node (.+) with guid (.+) at time (\d+) in time (\d+) .*/ ) {
	    my ($start_time, $key, $owner, $find_time) = ($4, $1, $3, $5);
	    my $start_time_offset = $start_time - $begin_time;
	    $max_offset = $start_time_offset if( $start_time_offset > $max_offset );
#	    print "FIND: $start_time_offset, ($start_time): $find_time\n";
	    my $bucket_index = ($start_time_offset / $bucket_width);
	    $completed_buckets[$bucket_index]++;

	    if( $good_owner{$key} eq $owner ) {
		# Consistent lookup.
		push @{$find_buckets[$bucket_index]}, $find_time;
	    } elsif (defined $ignore_owner{$key}) {
		warn( "ignoring $key.  Too few lookups.\n" );
	    } else {
		# Consider it a failure.
	    }
	    $num_node_buckets[$bucket_index] = $num_nodes_up;
	    $num_node_started_buckets[$bucket_index] = $num_nodes_started;
	    
	} elsif( /FAILED: .* obj (.*) on .* at time (\d+) .*/ ){
	    # don't count it if it's been destroyed
	    # assumption: there is one unified log in chronological order
	    # (like the simulator)
	    # TODO: make general (in two scans maybe? But it should only
	    #       count as detroyed AFTER it has actually been destroyed
	    if( grep( /$1/, @destroyed_objs ) ) {
		#print "Skipping destroyed object $1\n";
		next;
	    }
	    my $start_time = $2;
	    my $start_time_offset = $start_time - $begin_time;
	    $max_offset = $start_time_offset if( $start_time_offset > $max_offset );
#	    print "FAIL: $start_time_offset, ($start_time)\n";

	    my $bucket_index = ($start_time_offset / $bucket_width);
	    $fail_buckets[$bucket_index]++;
#	    print "\t $fail_buckets[$bucket_index] $1 failed\n";
	} elsif( /DESTROYED (.*)/ ){
	    #print "Destroyed object $1\n";
	    push @destroyed_objs, $1;
	} elsif( /BANDWIDTH START (\d+)/ ) {
	    $bw_mode = 1;
	    my $start_time_offset = $1 - $begin_time;
	    $bw_bucket = ($start_time_offset / $bucket_width);
#	} elsif( /.* Got LocateMsg .*/ ) {
	} elsif( /.*Joined the network.*/ ) {
	    $num_nodes_up++;
	} elsif( /RepairMonitor.* finished detachment.*/ ) {
	    $num_nodes_up--;
	} elsif( /.*event .*StagesInitialized.*/ ) {
	    $num_nodes_started++;
	} elsif( /.*Dispatching failure.*/ ) {
	    $num_nodes_started--;
	}
    }
}
print "#bucket_low bucket_high count cons succ_frac median min mean max stddev bw nodes_up nodes_started\n";


my $max_index = ($max_offset / $bucket_width);
for( my $i=0; $i<=$max_index; $i++ ) {
    my $stat = Statistics::Descriptive::Full->new();
    my $fail_count = ($fail_buckets[$i] or 0);    
    my @find_times;
    my $bw_stat = Statistics::Descriptive::Full->new();
    my $bw = 0;
    if( defined $bw_buckets[$i] ) {
	my @bws = @{$bw_buckets[$i]};
	$bw_stat->add_data(@bws);
	$bw = $bw_stat->mean();
    }
    if( defined $find_buckets[$i] ) {
	@find_times = @{$find_buckets[$i]};
	$stat->add_data(@find_times);

	my $find_count = $stat->count();
	my $fail_count = ($fail_buckets[$i] or 0);
	my $complete_count = ($completed_buckets[$i] or 0);	# complete
	my $cons_frac = (($complete_count==0) ? 1 : 
			 ($find_count/$complete_count));
	my $req_count = ($fail_count + $complete_count);	# total
	my $succ_frac = (($req_count==0) ? 1 : 
			 ($complete_count/$req_count));
	my $num_nodes = $num_node_buckets[$i];
	my $num_nodes_started = $num_node_started_buckets[$i];
	print_stats( ($i*$bucket_width), (($i+1)*$bucket_width),
		     ($find_count+$fail_count),
		     $cons_frac,
		     $succ_frac,
		     $stat->median(),
		     $stat->min(),
		     $stat->mean(),
		     $stat->max(),
		     (($find_count > 1) ? $stat->standard_deviation() : 0),
		     $bw,
		     $num_nodes,
		     $num_nodes_started);
#  	print "Bucket #$i: [" . ($i*$bucket_width) . " - "
#  	    . (($i+1)*$bucket_width) . "]: Found: $find_frac ($find_count/"
#  		. ($find_count+$fail_count) . ")\n";
#  	print "Bucket #$i: [" . ($i*$bucket_width) . " - "
#  	    . (($i+1)*$bucket_width) . "]: Med: " . $stat->median() . " ";
#  	print "Min: " . $stat->min() . " ";
#  	print "Mean: " . $stat->mean() . " ";
#  	print "Max: " . $stat->max() . " ";
#  	if( $find_count > 1 ) {
#  	    print "Stddev: " . $stat->standard_deviation() . "\n";
#  	} else {
#  	    print "\n";
#  	}
    } else {
#  	print "Bucket #$i: [" . ($i*$bucket_width) . " - "
#  	    . (($i+1)*$bucket_width) . "]: Found: 0 (0/$fail_count)\n";
	print_stats( ($i*$bucket_width), (($i+1)*$bucket_width),
		      $fail_count, 0, 0, 0, 0, 0, 0, 0, $bw, 0, 0);
    }

}
    
#  my $stat = Statistics::Descriptive::Full->new();	    
#  $stat->add_data( @begin_times );	
#  print "BEGIN times:  Med: " . $stat->median() . " ";
#  print "Min: " . $stat->min() . " ";
#  print "Max: " . $stat->max() . " ";
#  print "Diff: " . ($stat->max()-$stat->min()) . " ";
#  print "Stddev: " . $stat->standard_deviation() . "\n";

exit( 0 );

sub print_stats {
    my( $bucket_low, $bucket_high, $count, $cons, $frac, $median, $min, $mean,
	$max, $stddev, $bw, $num_nodes, $num_nodes_started ) = @_;

    print "$bucket_low $bucket_high $count $cons $frac $median $min $mean $max $stddev $bw $num_nodes $num_nodes_started\n";
}
