# Tools.pm
#
# Copyright (c) 2001 Regents of the University of California.
# All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions
#  are met:
#  1. Redistributions of source code must retain the above copyright
#     notice, this list of conditions and the following disclaimer.
#  2. Redistributions in binary form must reproduce the above copyright
#     notice, this list of conditions and the following disclaimer in the
#     documentation and/or other materials provided with the distribution.
#  3. Neither the name of the University nor the names of its contributors
#     may be used to endorse or promote products derived from this software
#     without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
#  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
#  OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
#  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
#  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
#  OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#  SUCH DAMAGE.

#
# $Id: stats.pm,v 1.1 2002/11/13 21:22:36 hweather Exp $
#
# Useful common statistical subroutines.
# original author of stats.pl Matt Welsh <mdw@cs.berkeley.edu>
# modularized by Hakim Weatherspoon <hweather@cs.berkeley.edu>
# 
# Reads in a histogram of <val>, <count> values and prints some
# basic statistics about it. Also generates histograms.

use Exporter();
use strict;
use POSIX;
use vars qw( @ISA @EXPORT );

@ISA = ("Exporter");
@EXPORT = qw( &calc_stats &calc_histogram &graph_histogram );

$| = 1;

############################################################
# GLOBAL VARIABLES

# print to stdout
my $PRINT = 1;


############################################################
# SUBROUTINES

# Given an array of data calculate the mean, median, 
# min, max, stddev, 5th percentile, and 95th percentile.
#
# Pass in an array of data,
#	(e.g. @data=[2,6,2,7,3,...]).
# Usage: &calc_stats( <data> )
sub calc_stats
{
  my (@data) = @_;

  my @stats;
  my $num = $#data + 1;
  my $min = -1;
  my $max = -1;
  my $mean = 0.0;
  my $median = -1;
  my $stddev = 0.0;
  my $fiveperc = -1;
  my $ninefiveperc = -1;

  if($num <= 0) {
    push(@stats, $num);
    push(@stats, $min);
    push(@stats, $max);
    push(@stats, $mean);
    push(@stats, $median);
    push(@stats, $stddev);
    push(@stats, $fiveperc);
    push(@stats, $ninefiveperc);

    return @stats;
  }

  # Sort data
  my @sorted = sort { $a <=> $b} @data;
  if($num != $#sorted+1) {
    die("num=$num != sorted+1=" . ($#sorted+1));
  }

  # Calculate median
  if ($num % 2) {
    $median = ($sorted[$num/2] + $sorted[($num/2)+1]) / 2;
  } else {
    $median = $sorted[$num/2];
  }

  # Calculate min and max
  $min = $sorted[0];
  $max = $sorted[$#sorted];

  # Calculate the 5th and 95th percentile
  $fiveperc = $sorted[floor($num * 0.05)];

  my $ninefiveindex = ceil($num * 0.95);
  $ninefiveperc = $sorted[($ninefiveindex==$num?$num-1:$ninefiveindex)];

  # Calculate mean
  my $total = 0;
  my $i = -1;
  for ($i = 0; $i <= $#sorted; $i++) {
    $total += $sorted[$i];
  }
  $mean = $total / $num;

  # Calculate stddev
  $total = 0;
  for ($i = 0; $i <= $#sorted; $i++) {
    $total += ($sorted[$i] - $mean) * ($sorted[$i] - $mean)
  }
  $stddev = sqrt($total / ($num - 1));

  push(@stats, $num);
  push(@stats, $min);
  push(@stats, $max);
  push(@stats, $mean);
  push(@stats, $median);
  push(@stats, $stddev);
  push(@stats, $fiveperc);
  push(@stats, $ninefiveperc);

  if($PRINT)
  {
    print "$num data points\n";
    print "min: $min\n";
    print "max: $max\n";
    print "mean: $mean\n";
    print "stddev: $stddev\n";
    print "median: $median\n";
    print "5percentile: $fiveperc\n";
    print "95percentile: $ninefiveperc\n";
  }
  return @stats;
}

# Given a bucket size and array of data calculate 
# histogram.
#
# Pass in a bucket size and array of data,
#	(e.g. @data=[2,6,2,7,3,...]).
# Usage: &calc_stats( <bucket_size> <data> )
sub calc_histogram
{
  my ($bucket_size, @data) = @_;

  my @sorted = sort { $a <=> $b} @data;
  my @histogram;
  
  my $num_buckets = &ceiling($sorted[$#sorted]/$bucket_size);
  my $i = -1;
  
  for($i = 0; $i < $num_buckets; $i++)
  {
    push(@histogram, 0);
  }
  
  die("histogram length=$#histogram != num_buckets=" . ($num_buckets-1)) if($num_buckets != $#histogram+1);
  
  for($i = 0; $i <= $#sorted; $i++) {
    my $bucket = int($sorted[$i] / $bucket_size);
    $histogram[$bucket]++;
  }
  return @histogram;
}

# Given a bucket size and histogram present a graphical
# representation of the histogram.
#
# Pass in title, filename, graph (i.e. {1, 0}), 
#   save (i.e. save file or not {1,0}), print (i.e. {1,0}),
#   bucket_size, and histogram.
#
# Usage: &calc_histogram( <title> <filename> <graph> <save> <print>
#                         <bucket_size> <histogram> )
sub graph_histogram
{
  my ($title, $filename, $graph, $save, $print, $bucket_size, @histogram) = @_;

  my $dat_filename = $filename . ".dat";
  my $graph_filename = $filename . ".gnuplot";

  if($print)
  {
    print "\nHistogram follows:\n\n";
    for (my $i = 0; $i <= $#histogram; $i++) {
      if ($histogram[$i] != 0) {
	my $bucket = $i * $bucket_size;
	print "BUCKET $bucket COUNT $histogram[$i]\n";
      }
    }
  }

  open(OUT, ">$dat_filename") || die "Can't open $dat_filename\n";
  for (my $i = 0; $i <= $#histogram; $i++) {
    if ($histogram[$i] != 0) {
      my $bucket = $i * $bucket_size;
      print OUT "BUCKET $bucket COUNT $histogram[$i]\n";
    }
  }
  close OUT;
  open(OUT, ">$graph_filename") || die "Can't open $graph_filename\n";
  print OUT "set title '$title'\n";
  print OUT "set xlabel 'Bucket number (bucketsize $bucket_size)'\n";
  print OUT "set ylabel 'Count'\n";
  print OUT "set grid noxtics ytics\n";
  print OUT "plot [][0:] \"$dat_filename\" using 2:4 with i lw 5\n";
  print OUT "pause -1 'Press return to exit...'\n";
  close (OUT);

  print "graph filename=$graph_filename, dat filename=$dat_filename\n";
  `gnuplot $graph_filename`;
  if(0) {
    print "deleting everything";
    `rm -f $dat_filename $graph_filename`;
  }
}


# Given a bucket size and histogram present a graphical
# representation of the histogram.
#
# Pass in title, filename, save (i.e. save file or not {1,0}),
# bucket_size, and histogram.
#
# Usage: &calc_stats( <title> <filename> <save> <bucket_size> <histogram>)
sub ceiling
{
  my $value = shift(@_);
  my $int_value = int $value;
  my $rv = $int_value;
  $rv = $int_value + 1 if($int_value / $value > 0.0);

  return $rv;
}

1;
