/**
 * FilterAnalyzer.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package Sonar.PacketAnalyzers;

import Sonar.*;
import Sonar.util.*;
import Sonar.event.*;
import Sonar.Dialog.*;

import java.util.*;
import java.io.*;
import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.accessibility.*;
import java.util.zip.*;

import ostore.util.NodeId;
import ostore.tapestry.impl.RouteMsg;
import ostore.tapestry.impl.RouteObjRootMsg;
import ostore.tapestry.impl.RouteObjDestMsg;

/**
 * This analyzer receives a class name as a string input in its sonar menu. 
 * Then, the analyzer will look out for this specific message, and display
 * the appropriate edge with the desired color. Currently, it only knows how to
 * handle 4 different types of messages: 1) those transmitted by the network
 * stage itself 2) as the user data of a RouteMsg 3) inside a 
 * RouteObjRootMsg 3) or in the user data of a RouteObjDestMsg.
 * 
 * @author	Jason Lee <jlee81@cs.berkeley.edu>
 */

public class FilterAnalyzer extends PacketAnalyzer implements Runnable {
    protected TwoKeyHashtable proprietary_edge_info;
    protected long expiration_time = MainClass.edgeEventLongevity;
    protected Thread del_edge_thread;
    protected String class_input;
    protected MenuManager menu_manager;

    protected int msg_type = 1;
    protected final int AS_IS = 0;
    protected final int IN_ROUTE_MSG = 1;
    protected final int IN_ROUTE_OBJ_ROOT = 2;
    protected final int IN_ROUTE_OBJ_DEST = 3;
    
    public FilterAnalyzer() {
	proprietary_edge_info = new TwoKeyHashtable();
	del_edge_thread = new Thread(this);
	del_edge_thread.start();

	class_input = "";
	MainClass.displayManager.custom_color = Color.green;

	menu_manager = new MenuManager();
        MainClass.addOptionsPanelContributor(this);
    }

    public void run() {
	EdgeInfo tempEdge = null;
	while (del_edge_thread == Thread.currentThread()) {
	    long currentTime = Calendar.getInstance().getTime().getTime();
	    for(Enumeration edges = proprietary_edge_info.elements();
		    edges.hasMoreElements();) {
		tempEdge = (EdgeInfo)edges.nextElement(); 
		if(currentTime - expiration_time > 
			tempEdge.getTime()) {
		    if(tempEdge.getInstances() > 1) {
		        tempEdge.instances--;
			MainClass.displayManager.modifyEdgeWidth(tempEdge);
		    } else {
			MainClass.objectMaintainer.triggerEdgeDeletedEvent(
				new EdgeEvent(this, tempEdge, Calendar.getInstance().getTime()));
			proprietary_edge_info.remove(tempEdge.getSource(), 
				tempEdge.getDest());
		    }	
		}
	    }
	}
    }
    
    public void packetReceived(PacketEvent e) {
	NodeId sender = null;
	NodeId recp = null;

	if(e.hasPayload()) {
	    try {
		Object obj = null;

		switch(msg_type) {
		    case IN_ROUTE_MSG:
			if(e.getPacket() instanceof RouteMsg)
			    obj = ((RouteMsg) e.getPacket()).user_data();
			break;
		    case IN_ROUTE_OBJ_ROOT:
			if(e.getPacket() instanceof RouteObjRootMsg) 
			    obj = ((RouteObjRootMsg) e.getPacket()).user_data();
			break;
		    case IN_ROUTE_OBJ_DEST:
			if(e.getPacket() instanceof RouteObjDestMsg)
			    obj = ((RouteObjDestMsg) e.getPacket()).user_data();
			break;
		    case AS_IS:
			obj = e.getPacket();
			//break;
		}

		//Placed this early return for making the code more legible (instead
		//of an enormous if/else case)
		if(obj == null || !obj.getClass().getName().equals(class_input)) 
		    return;
		//MainClass.out("received a "+obj.getClass().getName());

		if(MainClass.analyzer_menu.show_along_tapestry) {
		    sender = e.getSender();
		    recp = e.getRecipient();
		} else {
		    switch(msg_type) {
			case IN_ROUTE_MSG:
			    sender = MainClass.objectMaintainer.guid_to_node(((RouteMsg) e.getPacket()).src);
			    recp = MainClass.objectMaintainer.guid_to_node(((RouteMsg) e.getPacket()).dest);
			    break;
			case IN_ROUTE_OBJ_DEST:
			    sender = MainClass.objectMaintainer.guid_to_node(((RouteObjDestMsg) e.getPacket()).src);
			    recp = MainClass.objectMaintainer.guid_to_node(((RouteObjDestMsg) e.getPacket()).dest);
			    //break;
		    }
		}
	    } catch (Exception exc) { exc.printStackTrace(); }

	} else {
	    if(!e.getMsgType().equals(class_input))
		return;

	    if(MainClass.analyzer_menu.show_along_tapestry) {
		sender = e.getSender();
		recp = e.getRecipient();
	    } else {
		sender = MainClass.objectMaintainer.guid_to_node(e.getSourceGuid());
		recp = MainClass.objectMaintainer.guid_to_node(e.getDestGuid());
	    }
	}

	//if a routing table containing their node'ids have not been recvd
	if(sender == null || recp == null) return;

	if(sender != recp) {
	    EdgeInfo edgeInfo = null;
	    if((edgeInfo = (EdgeInfo) proprietary_edge_info.get(sender,
			    recp)) != null) {
		edgeInfo.instances++;
		edgeInfo.modifyTime( 
			Calendar.getInstance().getTime().getTime());
		MainClass.displayManager.modifyEdgeWidth(edgeInfo); 
	    } else {
		edgeInfo = new EdgeInfo(sender, recp, "FilterAnalyzerMsg", Calendar.getInstance().getTime().getTime());
		proprietary_edge_info.put(sender, recp, edgeInfo); 

		MainClass.objectMaintainer.triggerEdgeCreatedEvent(new 
			EdgeEvent(this, edgeInfo, Calendar.getInstance().getTime()));
	    }

	}
    }

    public void showOptionsDialog() {
	StandardDialog newDialog = new StandardDialog(new OptionsPanel());
	newDialog.show(); 
    } 

    public ActivePanel getOptionsPanel() { 
	return new OptionsPanel(); 
    }

    public class OptionsPanel extends Sonar.Dialog.ActivePanel 
    {		 
	public OptionsPanel() 
	{ 
	    tabTitle = "Filter";//this will be the title of the tab 
	    setLayout(null); 
	    setSize(307,168); 

	    //input string
	    j_label1.setText("currently:"+class_input);
	    j_label1.setBounds(20,34, 300, 24);
	    add(j_label1);
	    
	    j_label2.setText("class:");
	    j_label2.setBounds(20,68, 80, 24);
	    add(j_label2);

	    j_textfield1.setToolTipText("type the class name of the msg you would like to display");
	    add(j_textfield1);
	    j_textfield1.setBounds(100, 68, 200, 24);

	    //choose a color
	    j_label3.setText("color:");
	    j_label3.setBounds(20, 92, 80, 24);
	    add(j_label3);
	    
	    j_combobox1.addItem(new String("black"));
	    j_combobox1.addItem(new String("red"));
	    j_combobox1.addItem(new String("blue"));
	    j_combobox1.addItem(new String("yellow"));
	    j_combobox1.addItem(new String("green"));
	    j_combobox1.addItem(new String("orange"));
	    add(j_combobox1);
	    j_combobox1.setBounds(100, 92, 200, 24);

	    //mark how its sent
	    j_combobox2.addItem(new String("as is"));
	    j_combobox2.addItem(new String("in a RouteMsg"));
	    j_combobox2.addItem(new String("in a RouteObjRootMsg"));
	    j_combobox2.addItem(new String("in a RouteObjDestMsg"));
	    add(j_combobox2);
	    j_combobox2.setBounds(90, 116, 210, 24);
	} 

	javax.swing.JTextField j_textfield1 = new javax.swing.JTextField(); 
	javax.swing.JComboBox j_combobox1 = new javax.swing.JComboBox(); 
	javax.swing.JComboBox j_combobox2 = new javax.swing.JComboBox();

	javax.swing.JLabel j_label1 = new javax.swing.JLabel();
	javax.swing.JLabel j_label2 = new javax.swing.JLabel();
	javax.swing.JLabel j_label3 = new javax.swing.JLabel();


    
	public void applyChanges() { 
	    class_input = j_textfield1.getAccessibleContext().getAccessibleText().getAtIndex(AccessibleText.SENTENCE, 0);
	    //MainClass.out(class_input);
	    
	    if(j_combobox1.getSelectedItem().equals("black")) {
		MainClass.displayManager.custom_color = Color.black;
	    } else if(j_combobox1.getSelectedItem().equals("red")) {
		MainClass.displayManager.custom_color = Color.red;
	    } else if(j_combobox1.getSelectedItem().equals("blue")) {
		MainClass.displayManager.custom_color = Color.blue;
	    } else if(j_combobox1.getSelectedItem().equals("yellow")) {
		MainClass.displayManager.custom_color = Color.yellow;
	    } else if(j_combobox1.getSelectedItem().equals("green")) {
		MainClass.displayManager.custom_color = Color.green;
	    } else if(j_combobox1.getSelectedItem().equals("orange")) {
		MainClass.displayManager.custom_color = Color.orange;
	    }

	    if(j_combobox2.getSelectedItem().equals("as is")) {
		msg_type = AS_IS;
	    } else if(j_combobox2.getSelectedItem().equals("in a RouteMsg")) {
		msg_type = IN_ROUTE_MSG;
	    } else if(j_combobox2.getSelectedItem().equals("in a RouteObjRootMsg")) {
		msg_type = IN_ROUTE_OBJ_ROOT;
	    } else if(j_combobox2.getSelectedItem().equals("in a RouteObjDestMsg")) {
		msg_type = IN_ROUTE_OBJ_DEST;
	    }
	} 
    
	//this function will be called when the panel is first shown 
	public void initializeDisplayValues()
	{
	} 

    }	           
 
    protected class MenuManager implements Serializable, ActionListener, ItemListener 
    { 
	JMenu mainMenu = new JMenu(); 
	JCheckBoxMenuItem packetProcessingCheckBox = new JCheckBoxMenuItem(); 
	JCheckBoxMenuItem backgroundProcessingCheckBox = new JCheckBoxMenuItem(); 
	JSeparator separator1 = new JSeparator(); 
	JMenuItem propertiesItem = new JMenuItem(); 
	JSeparator separator2 = new JSeparator(); 
	JMenu serializeMenu = new JMenu(); 
	JMenuItem saveNodesItem = new JMenuItem(); 
	JMenuItem loadNodesItem = new JMenuItem(); 
	JMenuItem saveEdgesItem = new JMenuItem(); 
	JMenuItem loadEdgesItem = new JMenuItem(); 
	JSeparator separator3 = new JSeparator(); 
	JMenu paintMenu = new JMenu(); 
	JCheckBoxMenuItem paintNodesItem = new JCheckBoxMenuItem(); 
	JCheckBoxMenuItem paintEdgesItem = new JCheckBoxMenuItem(); 
	JCheckBoxMenuItem paintScreenItem = new JCheckBoxMenuItem(); 

	public MenuManager() 
	{ 
	    mainMenu.setText("Filter Analyzer"); 
	    mainMenu.setActionCommand("Filter Analyzer"); 

	    packetProcessingCheckBox.setSelected(true); 
	    packetProcessingCheckBox.setText("Packet Processing"); 
	    packetProcessingCheckBox.setActionCommand("Packet Processing"); 
	    mainMenu.add(packetProcessingCheckBox); 

	    mainMenu.add(separator1); 
	    propertiesItem.setText("Options"); 
	    propertiesItem.setActionCommand("Options"); 
	    mainMenu.add(propertiesItem); 

	    MainClass.mainFrame.PacketAnalyzersMenu.add(mainMenu);

	    packetProcessingCheckBox.addItemListener(this); 
	    backgroundProcessingCheckBox.addItemListener(this); 
	    propertiesItem.addActionListener(this); 
	    saveNodesItem.addActionListener(this); 
	    loadNodesItem.addActionListener(this); 
	    saveEdgesItem.addActionListener(this); 
	    loadEdgesItem.addActionListener(this); 
	    paintNodesItem.addItemListener(this); 
	    paintEdgesItem.addItemListener(this); 
	    paintScreenItem.addItemListener(this); 
	} 

	public void actionPerformed(ActionEvent e) 
	{ 
	    Object object = e.getSource(); 
	    if (object == propertiesItem) 
		showOptionsDialog(); 
	} 

	public void itemStateChanged(ItemEvent e) 
	{ 
	    Object object = e.getSource(); 
	    if (object == packetProcessingCheckBox) 
		togglePacketProcessing(); 
	}		 

	public void togglePacketProcessing() { 
	    if(packetProcessingCheckBox.isSelected()) {
		MainClass.addPacketEventListener(FilterAnalyzer.this);
	    } else { 
		MainClass.removePacketEventListener(FilterAnalyzer.this);
	    }	
	}

    }
}
