/**
 * ObjectMaintainerOStore.java
 *
 * Copyright (c) 2001 Regents of the University of California.  All
 * rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above
 *    copyright notice, this list of conditions and the following
 *    disclaimer in the documentation and/or other materials provided
 *    with the distribution.
 * 3. Neither the name of the University nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS
 * IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE
 * REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */ 
package Sonar.PacketAnalyzers; 
 
import Sonar.*; 
import Sonar.event.*; 
import Sonar.util.*; 
import java.util.*; 
import java.lang.*; 
import java.io.*; 
import java.util.zip.*; 
import java.awt.*; 
import java.awt.event.*; 
import javax.swing.*; 
import Sonar.Dialog.*; 

import ostore.util.NodeId;
import ostore.util.SecureHash;

/**
 * Object Maintainer only keeps track of packets received from the node, 
 * whereas the RoutingTableAnalyzer determines the layout of the given
 * network itself. Thus, "mute" nodes are not stored in this class's
 * proprietyNodeInfo
 *
 * @author	Jason Lee <jlee81@cs.berkeley.edu>
 * @author	Kamin Whitehouse <kamin@cs.berkeley.edu>
 */
 
public class ObjectMaintainerOStore extends PacketAnalyzer {

  protected static Vector nodeListeners = new Vector();
  protected static Vector edgeListeners = new Vector();
  protected Hashtable proprietaryNodeInfo = new Hashtable();
  protected Map nodeToHashMap = new HashMap();
  protected Map hashToNodeMap = new HashMap();

  public ObjectMaintainerOStore() {
    MainClass.objectMaintainer.addNodeEventListener(this);
  }

  public void packetReceived(PacketEvent e) { }
 
  public synchronized SecureHash node_to_guid(NodeId node) {
    return (SecureHash) nodeToHashMap.get(node);
  }
  public synchronized NodeId guid_to_node(SecureHash guid) {
    return (NodeId) hashToNodeMap.get(guid);
  }
  
  /*
   * Attribute a node with a SHA1Hash and vice versa
   */
  public synchronized boolean associate_to_guid(NodeId node, SecureHash element) {
      NodeId oldNode = null;
      SecureHash oldSHash = null;
     
      /*
       * If both the node id and the SHA1Hash have never been seen before
       */
      if((oldSHash = (SecureHash) nodeToHashMap.get(node)) == null && 
	      (oldNode = (NodeId) hashToNodeMap.get(element)) == null) {
	  nodeToHashMap.put(node, element);
	  hashToNodeMap.put(element, node);
	  //MainClass.out("assigned "+node+" to "+element);
	  return true;
      } else if(oldNode == null) {
	  oldNode = (NodeId) hashToNodeMap.get(oldSHash);
	  //If we have an assoc node id with the given SHA1Hash
	  if(oldNode != null) { 
	      if(oldSHash.equals(element)) { //If the SHA1Hash's are the same
	      	  MainClass.out(oldNode+" for (same) old: "+
		  oldSHash+ " vs. new: "+element);
	      } else { //If the SHA1Hash's are different
	      	  MainClass.out(oldNode+" for (diff) old: "+
		  oldSHash+ " vs. new: "+element +" and old nodeid reset");
		  //Remove the old mapping
	  	  nodeToHashMap.remove(node); 
	  	  nodeToHashMap.put(node, element);
	      }
	  }
	  //hashToNodeMap.remove(element); //Keep both elements
	  hashToNodeMap.put(element, node);
      } else if(oldSHash == null) {
	  MainClass.out("null hash, but node's there");
	  //nodeToHashMap.remove(node);
      }
      MainClass.out("Attempting to assign existing nodeid "+oldNode+" ,"
		  + oldSHash + " to:"+node+" ,"+element);
      return false;
  }

  /*
   * Replaces the current mapping of node id to SHA1Hash to a new hash value
   */
  public synchronized void replace_with_new_guid (NodeId node, SecureHash element) {
    SecureHash oldHash = (SecureHash) nodeToHashMap.get(node);

    nodeToHashMap.remove(node);
    nodeToHashMap.put(node, element);

    hashToNodeMap.remove(oldHash);
    hashToNodeMap.put(element, node);
    MainClass.out("Replaced ==("+oldHash.equals(element)+") "+oldHash+" with "+element+ " for "+node.port());
  }
	
  public synchronized void nodeCreated(NodeEvent e) {
    NodeId node = e.getNodeId();
    if(!proprietaryNodeInfo.containsKey(node)) {
	proprietaryNodeInfo.put(node, e.getNodeInfo());
    }    
  }

  public synchronized void nodeDeleted(NodeEvent e) {
    //Keep the mapping between NodeId and Hash 
  }

  public synchronized void identifyNode(NodeId node, int type) {
      MainClass.displayManager.identifyNode(node, type);
      MainClass.locationAnalyzer.identifyNode(node, type);
  }
  
  public static void addNodeEventListener(NodeEventListener pListener) 
  { 
    nodeListeners.add(pListener); 
  } 
  
  public static void removeNodeEventListener(NodeEventListener pListener) 
  { 
    nodeListeners.remove(pListener); 
  } 
  
  public static void addEdgeEventListener(EdgeEventListener pListener) 
  { 
    edgeListeners.add(pListener); 
  }

  public static void removeEdgeEventListener(EdgeEventListener pListener) 
  { 
    edgeListeners.remove(pListener); 
  } 

  protected synchronized void triggerNodeCreatedEvent(NodeEvent e) 
  { 
    //for each listener 
    NodeEventListener currentListener; 
    for(Enumeration list = nodeListeners.elements(); list.hasMoreElements();) 
      { 
	currentListener = (NodeEventListener)list.nextElement(); 
	currentListener.nodeCreated(e);//send the listener an event 
      }			 
  }
  
  protected synchronized void triggerNodeDeletedEvent(NodeEvent e) 
  { 
    //for each listener 
    NodeEventListener currentListener; 
    for(Enumeration list = nodeListeners.elements(); list.hasMoreElements();) 
      { 
	currentListener = (NodeEventListener)list.nextElement(); 
	currentListener.nodeDeleted(e);//send the listener an event 
      }			 
  }
  
  protected synchronized void triggerEdgeCreatedEvent(EdgeEvent e) 
  { 
    //for each listener 
    EdgeEventListener currentListener; 
    for(Enumeration list = edgeListeners.elements(); list.hasMoreElements();) 
      { 
	currentListener = (EdgeEventListener)list.nextElement(); 
	currentListener.edgeCreated(e);//send the listener an event 
      }			 
  }
  protected synchronized void triggerEdgeDeletedEvent(EdgeEvent e) 
  { 
    //for each listener 
    EdgeEventListener currentListener; 
    for(Enumeration list = edgeListeners.elements(); list.hasMoreElements();) 
      { 
	currentListener = (EdgeEventListener)list.nextElement(); 
	currentListener.edgeDeleted(e);//send the listener an event 
      }			 
  }      

}	
